/*
 * Copyright (C) 2023 Associated Universities, Inc. Washington DC, USA.
 *
 * This file is part of NRAO Workspaces.
 *
 * Workspaces is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Workspaces is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
 */

// Package properties This file contains methods to retrieve properties from workflows in the
// archive database
package properties

import (
	"database/sql"
	"encoding/json"
	"fmt"
	"github.com/PaesslerAg/jsonpath"
	"strconv"
)

// Intermediary object to store values from the workflow_requests entry so they
// can be marshalled to JSON
type workflowRequest struct {
	ResultsDir string `json:"results_dir"`
}

// Property The type which stores the JSONPath queries for each property
// If we want to be able to get a new property from the database, we simply
// 1. Add a command line flag to retrieve it
// 2. Add an enum value for it
// 3. Add the JSONPath query in the String method
type Property int

const (
	ResultsDir Property = 1
)

func (p Property) String() string {
	switch p {
	case ResultsDir:
		return "$.results_dir"
	default:
		return ""
	}
}

// GetArgument Query the workflow arguments for a property with JSONPath
//
// @param prop a Property with the JSONPath query
// @param wfId an int with the Archive or Workspaces workflow ID
// @param db a *sql.DB with a connection to the archive database
// @return A string with the property value or an error
func GetArgument(prop Property, wfId int, db *sql.DB) (string, error) {
	var value string

	arguments, err := getWfArgs(wfId, db)
	if err != nil {
		return value, err
	}

	var valInterface interface{}
	valInterface, err = jsonpath.Get(prop.String(), arguments)
	if err != nil {
		return value, err
	}

	value = valInterface.(string)

	return value, err
}

// getWfArgs Get the workflow request arguments as JSON in an interface so they can be
// queried
//
// @param wfId an int with the workflow ID
// @param db a *sql.DB with a connection to the archive database
// @return the workflow arguments as an interface or an error
func getWfArgs(wfId int, db *sql.DB) (interface{}, error) {
	var arguments map[string]interface{}

	// Allow querying for either workspaces ID or archive ID
	queryString := fmt.Sprintf(
		`SELECT results_dir FROM workflow_requests WHERE workflow_request_id=$1 OR argument->>'rh_id'=CAST($1 AS VARCHAR)`,
	)
	query, err := db.Prepare(queryString)
	if err != nil {
		return arguments, err
	}

	row := query.QueryRow(strconv.Itoa(wfId))

	var reqArgs workflowRequest
	err = row.Scan(
		&reqArgs.ResultsDir,
	)
	if err != nil {
		return arguments, err
	}

	// Convert the workflowRequest struct to an interface for querying
	var reqJson []byte
	reqJson, err = json.Marshal(reqArgs)
	if err != nil {
		return arguments, err
	}

	err = json.Unmarshal(reqJson, &arguments)
	if err != nil {
		return arguments, err
	}

	return arguments, err
}
