#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
"""
Auditor has one task: inspection of metadata.json and PPR.xml for required properties
"""
import logging
import os.path
import shutil
from copy import copy
from pathlib import Path
from typing import Dict, Optional

from casa_envoy.enums import ProductType
from casa_envoy.interfaces import AuditorIF
from casa_envoy.schema import AbstractTextFile
from casa_envoy.utilities import get_xml_content, get_json_content, get_value_for
from lxml import etree

PPR_FORMATTER = """<xsl:stylesheet version="1.0" xmlns:xsl="http://www.w3.org/1999/XSL/Transform">
    <xsl:output indent="yes"/>
    <xsl:strip-space elements="*"/>

    <!-- IDENTITY TRANSFORM -->
    <xsl:template match="@*|node()">
      <xsl:copy>
        <xsl:apply-templates select="@*|node()"/>
      </xsl:copy>
    </xsl:template>

    <!-- RUN normalize-space() ON ALL TEXT NODES -->
    <xsl:template match="text()">
        <xsl:copy-of select="normalize-space()"/>
    </xsl:template>
</xsl:stylesheet>
"""


def get_fields_for(product_type: str, filename: str) -> list:
    cal_metadata_list = [
        "fileSetIds",
        "workflowName",
        "systemId",
        "creationTime",
        "productLocator",
        "destinationDirectory",
    ]
    img_metadata_list = [
        "fileSetIds",
        "workflowName",
        "systemId",
        "creationTime",
        "productLocator",
        "destinationDirectory",
        "calibrationSourceDirectory",
        "cmsName",
    ]

    restore_metadata_list = [
        "calProductLocator",
    ]

    ppr_list = [
        "RootDirectory",
        "RelativePath",
        "SdmIdentifier",
    ]

    if ".xml" in filename:
        return ppr_list

    if ".json" in filename and "vlass" in product_type:
        return []
    elif ".json" in filename and "cal" in product_type or product_type == "integrated":
        return cal_metadata_list
    elif ".json" in filename and "img" in product_type:
        return img_metadata_list
    elif ".json" in filename and product_type == "restore":
        return cal_metadata_list + restore_metadata_list


class AuditFiles(AuditorIF):
    def __init__(self, ppr: Optional[AbstractTextFile], other_files: list[AbstractTextFile], settings: Dict[str, str]):
        self.logger = logging.getLogger("casa_envoy")
        self.ppr = ppr
        self.files = copy(other_files)
        if ppr is not None:
            self.files.append(ppr)
        self.settings = settings
        self.product_type = settings.get("product_type")

    @staticmethod
    def check_required_fields(file: AbstractTextFile, fields: list) -> bool:
        missing = []
        if ".xml" in file.filename:
            ppr_content = get_xml_content(file)
            for tag in fields:
                try:
                    # ensure tag is not empty
                    get_value_for(ppr_content, tag)
                except AttributeError:
                    missing.append(tag)

        if ".json" in file.filename:
            json_content = get_json_content(file)
            for tag in fields:
                if tag not in json_content:
                    missing.append(tag)
        if len(missing) > 0:
            print(f"Missing fields: {missing}")
            return False
        else:
            return True

    @staticmethod
    def format_relative_path(relative_path: str) -> str:
        """
        Occasionally (very very very occasionally) the relative path might be numerical only.
        CASA hates this and falls over (see WS-2116). If the relative path is numerical enclose it in quotes

        :param relative_path: the relative path to check
        :return: relative path string, corrected if necessary
        """
        return relative_path if not relative_path.isdigit() else "'" + str(relative_path) + "'"

    def correct_for_condor(self, ppr: AbstractTextFile) -> AbstractTextFile:
        """
        Correct PPR.xml for HTCondor directory and path

        :param ppr: content of the waiting PPR
        :return: updated content of PPR written to working area
        """
        shutil.copy(ppr.filename, "./working")
        os.chdir("./working")

        parsed_xml = get_xml_content(ppr)
        parsed_xml.find("RootDirectory").string = self.settings["rootDirectory"]
        parsed_xml.find("RelativePath").string = self.format_relative_path(self.settings["processingDirectory"])
        strip_declaration = str(parsed_xml).split("\n")[1:]
        fixed_declaration = "".join(strip_declaration)

        corrected_xml = etree.fromstring(fixed_declaration)
        formatter = etree.XSLT(etree.fromstring(PPR_FORMATTER))

        corrected_content = etree.tostring(
            formatter(corrected_xml),
            pretty_print=True,
            xml_declaration=True,
            encoding="UTF-8",
            standalone=True,
        )

        with open(ppr.filename, "w") as file:
            file.write(corrected_content.decode())

        # make sure we also update the currently used content. needed for field validation
        ppr.content = corrected_content.decode()
        os.chdir("../")
        return ppr

    def audit(self) -> bool:
        if self.ppr is None and not (
            self.settings["product_type"] == ProductType.RESTORE.value
            and self.settings.get("piperestorescript_path") is not None
        ):
            self.logger.error(
                "INVALID CONFIGURATION: Expected either a PPR.xml or a restore with a piperestorescript, found neither"
            )
            return False
        invalid_files = []
        for file in self.files:
            self.logger.info(f"Auditing file {file.filename}...")

            if "PPR" in file.filename and ".xml" in file.filename:
                self.logger.info(f"Correcting {file.filename} for condor processing...")
                file = self.correct_for_condor(file)

            valid = self.check_required_fields(file=file, fields=get_fields_for(self.product_type, file.filename))
            if not valid:
                invalid_files.append(file.filename)

        if len(invalid_files) != 0:
            self.logger.error(f"INVALID FILE FOUND: {invalid_files}")
            return False
        else:
            return True


class AuditDirectories(AuditorIF):
    def __init__(self, metadata: AbstractTextFile, settings: Dict[str, str]):
        self.logger = logging.getLogger("casa_envoy")
        self.metadata_content = get_json_content(metadata)
        self.parameters = settings
        self.rootDirectory = settings["rootDirectory"]
        self.relative_path = settings["processingDirectory"]

    def audit(self) -> bool:
        current = os.getcwd()
        needed = self.rootDirectory + "/" + self.relative_path
        if needed != current:
            self.logger.error("DIRECTORY ERROR: not in correct directory for processing.")
            return False
        else:
            working = Path(current + "/working").is_dir()
            products = Path(current + "/products").is_dir()
            rawdata = Path(current + "/rawdata").is_dir()

            if working and products and rawdata:
                self.logger.info("Checking for rawdata...")
                data = os.listdir(Path(current + "/rawdata/"))
                if len(data) > 0:
                    self.logger.info("Data is available. Proceeding...")

                    if self.parameters["product_type"] == ProductType.RESTORE.value:
                        self.logger.info("Checking products/ for calibration tables...")
                        cal_data = os.listdir(Path(current + "/products/"))
                        if len(cal_data) > 0:
                            self.logger.info("Calibration data is available. Proceeding...")
                        else:
                            self.logger.error("FAILURE: calibration data not found in products/")
                            return False
                    # if (
                    #     self.parameters["product_type"] == ProductType.VLASS_IMG.value
                    #     and self.metadata_content["workflowName"] == "vlass_coarse"
                    # ):
                    #     self.logger.info("Checking working/ for reimaging resources and parameters.list...")
                    #     working_path = current + "/working/"
                    #     resources = len([os.path.isdir(d) for d in os.listdir(Path(working_path))])
                    #     self.logger.info(f"resources count = {resources}")
                    #     param_list = [f for f in os.listdir(working_path) if f.endswith(".list")]
                    #     self.logger.info(f"param list count = {len(param_list)}")
                    #     # param_list should contain both SEIP and CCIP parameter.list files
                    #     # reimaging_resources.tgz contains multiple directories, check if they're available
                    #     if resources > 0 and len(param_list) == 2:
                    #         self.logger.info("Reimaging resources and parameter.list are available. Proceeding...")
                    #     else:
                    #         self.logger.error("FAILURE: reimaging resources not found in working/")
                    #         return False

                    return True
                else:
                    self.logger.error("FAILURE: data not found in rawdata/")
                    return False
            else:
                self.logger.error("DIRECTORY ERROR: A directory is missing from the processing root directory.")
