#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
import logging
import os

from casa_envoy.interfaces import RunnerABC


class CasaConfigure:
    """
    Class for unifying environment setup. Required by all CASA Runner types
    """

    def __init__(self, parameters: dict, content: dict):
        self.logger = logging.getLogger("casa_envoy")
        self.parameters = parameters
        self.metadata_content = content

    def setup_environment(self):
        """
        Set required environment variables

        No need to audit them afterwards for None's since the keys and values in `os.environ` must all be strings.
        Source: https://docs.python.org/3.10/library/os.html#os.environ

        :return:
        """
        os.environ["SCIPIPE_ROOTDIR"] = self.parameters["rootDirectory"]

        # If user specified CASA Version, use it.
        if self.metadata_content.get("casaHome") is not None:
            os.environ["CASA_HOME"] = self.metadata_content["casaHome"]
        elif (
            self.metadata_content.get("main") is not None and self.metadata_content["main"].get("CASA_HOME") is not None
        ):
            # Catch case of VLASS nesting
            os.environ["CASA_HOME"] = self.metadata_content["main"]["CASA_HOME"]
        else:
            # If VLASS, use the current VLASS CASA release
            if "vlass" in self.parameters["product_type"]:
                os.environ["CASA_HOME"] = self.parameters["vlassHome"]
            else:
                os.environ["CASA_HOME"] = self.parameters["homeForReprocessing"]

        os.environ["LANG"] = "en_US.UTF-8"
        # Restores oftentimes use a piperestorescript instead of a PPR.xml
        # AuditFiles would catch an erroneously absent PPR
        if self.parameters.get("ppr") is not None:
            os.environ["PPR_FILENAME"] = self.parameters["ppr"]

        self.log_processing_env()

    def log_processing_env(self):
        """
        Log out environment variables relevant to CASA

        :return:
        """
        env_list = ["SCIPIPE_ROOTDIR", "CASA_HOME", "PPR_FILENAME", "LANG"]

        for var in env_list:
            self.logger.info(f"{var}: {os.environ.get(var)}")


class StandardRunner(RunnerABC):
    """
    Class for launch of standard (non-parallel) CASA
    """

    @property
    def build_command(self):
        return (
            "PYTHONPATH='' xvfb-run -e ${PWD}/xvfb-run.err.txt -d -s \"-screen 0 800x600x16\" "
            "${CASA_HOME}/bin/casa --pipeline --nogui --nologger -c "
            "${CASA_HOME}/pipeline/pipeline/" + self.pipeline_script + " ${PPR_FILENAME} || true"
        )


class PiperestorescriptRunner(RunnerABC):
    """
    Class for launch of non-parallel CASA, driven by a piperestorescript.py
    Only intended for CMS restores
    """

    @property
    def build_command(self):
        if not self.parameters.get("piperestorescript_path"):
            raise ValueError(f"Exected path to a casa_piperestorescript.py to be specified, found none")
        return (
            "PYTHONPATH='' xvfb-run -e ${PWD}/xvfb-run.err.txt -d -s \"-screen 0 800x600x16\" "
            "${CASA_HOME}/bin/casa --pipeline --nogui --nologger -c "
            f"{self.parameters['piperestorescript_path']} || true"
        )


class ParallelRunner(RunnerABC):
    """
    Class for launch of parallel CASA processing using mpicasa
    """

    @property
    def log_message(self):
        return f"RUNNING PARALLEL CASA with command '{self.build_command}'"

    @property
    def build_command(self):
        # CASA wants a supervisor process, in addition to the workers.
        # So increment the parallelization request by one to account for that
        integer_processes = int(self.parameters["requested_parallel"]) + 1

        # The --oversubscribe argument to mpicasa lets the N+1 processes share N CPUs happily
        return (
            "PYTHONPATH='' xvfb-run -e ${PWD}/xvfb-run.err.txt -d -s \"-screen 0 800x600x16\""
            " ${CASA_HOME}/bin/mpicasa --oversubscribe -n "
            + str(integer_processes)
            + " ${CASA_HOME}/bin/casa --pipeline --nogui --nologger -c "
            "${CASA_HOME}/pipeline/pipeline/" + self.pipeline_script + " ${PPR_FILENAME} || true"
        )
