#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
"""
Conveyor is responsible for moving or copying results from standard capability requests to the DA
QA area, called "delivery", and back, called "post QA".
"""
import argparse
import logging
import pathlib
import sys
from typing import Dict

from pycapo import CapoConfig

from .deliver import DeliveryConveyor
from .retrieve import RetrievalConveyor
from .solicitor import Solicitor
from .vlass import VlassConveyor

logger = logging.getLogger("conveyor")
logger.setLevel(logging.INFO)
logger.addHandler(logging.StreamHandler(sys.stdout))


def arg_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        description="Workspaces Standard Capability Postprocessing",
        formatter_class=argparse.RawTextHelpFormatter,
    )
    parser.add_argument(
        "--deliver",
        "--deliver-img",
        nargs=1,
        action="store",
        required=False,
        help="deliver standard calibration and image directories to analyst QA area",
    )
    parser.add_argument(
        "--retrieve",
        "--retrieve-img",
        nargs=1,
        action="store",
        required=False,
        help="return standard calibration and image directories to original parent directory",
    )
    parser.add_argument(
        "--qa-fail",
        nargs=1,
        action="store",
        required=False,
        help="move results of failed QA to original parent directory",
    )
    parser.add_argument(
        "--vlass",
        nargs=1,
        action="store",
        required=False,
        help="cache weblog for VLASS QA",
    )
    return parser


def _get_settings(filename: str) -> Dict[str, str]:
    solicitor = Solicitor(filename=filename)

    delivery_properties = CapoConfig().settings("edu.nrao.workspaces.DeliverySettings")
    processing_properties = CapoConfig().settings("edu.nrao.workspaces.ProcessingSettings")
    workflow_url = CapoConfig().settings("edu.nrao.workspaces.WorkflowSettings").serviceUrl
    external_capability_url = CapoConfig().settings("edu.nrao.workspaces.CapabilitySettings").externalServiceUrl

    qa_delivery_area = delivery_properties.qaDelivery
    weblog_cache_area = delivery_properties.cacheWeblogDirectory
    workspaces_lustre_root_dir = processing_properties.rootDirectory

    current_root_directory = str(pathlib.Path.cwd().parent)
    current_subdirectory = str(pathlib.Path.cwd().name)
    destination_dir = solicitor.solicit_delivery_location()
    destination_subdir = pathlib.Path(destination_dir).name
    request_params = solicitor.solicit_request_params()

    return {
        "qa_delivery_area": qa_delivery_area,
        "weblog_cache_area": weblog_cache_area,
        "workspaces_lustre_root_dir": workspaces_lustre_root_dir,
        "current_root_directory": current_root_directory,
        "current_subdirectory": current_subdirectory,
        "destination_dir": destination_dir,
        "destination_subdir": destination_subdir,
        "request_id": request_params["system_id"],
        "sdm_ids": request_params["sdm_ids"],
        "workflow_url": workflow_url,
        "capability_url": external_capability_url,
        "workflow_type": request_params["workflow"],
    }


def _get_vlass_settings(filename: str) -> Dict[str, str]:
    solicitor = Solicitor(filename=filename)

    vlass_cache = CapoConfig().settings("edu.nrao.workspaces.VlassDeliverySettings").cache
    vlass_weblog_cache_area = vlass_cache + "/weblog"

    parameters = solicitor.solicit_vlass_parameters()

    merge_params = {
        **parameters,
        "vlass_weblog_cache_area": vlass_weblog_cache_area,
    }

    return merge_params


def _determine_fail_action(settings: Dict):
    if "calibration" in settings.get("workflow_type"):
        action = "QA Fail Calibration"
    else:
        action = "QA Fail Image"

    return action


def main():
    args = arg_parser().parse_args()
    action = None
    conveyor = None

    if args.deliver is not None:
        action = "Calibration and/or Imaging Delivery"
        settings = _get_settings(args.deliver[0])
        conveyor = DeliveryConveyor(settings, action)
    elif args.retrieve is not None:
        action = "Calibration and/or Imaging Retrieval"
        settings = _get_settings(args.retrieve[0])
        conveyor = RetrievalConveyor(settings, action)
    elif args.qa_fail is not None:
        settings = _get_settings(args.qa_fail[0])
        action = _determine_fail_action(settings)
        conveyor = RetrievalConveyor(settings, action)
    elif args.vlass is not None:
        settings = _get_vlass_settings(args.vlass[0])
        action = "VLASS Weblog"
        conveyor = VlassConveyor(settings, action)

    conveyor.convey()
    if args.vlass is not None:
        logger.info(f"{action} caching is complete!")
    else:
        logger.info(f"Standard {action} is complete!")
