#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
"""
Tests for conveyor.deliver module
"""
from unittest.mock import MagicMock, mock_open, patch

from conveyor.deliver import DeliveryConveyor

settings = {
    "qa_delivery_area": "/lustre/aoc/cluster/pipeline/docker/workspaces/qa",
    "weblog_cache_area": "/lustre/aoc/cluster/pipeline/docker/workspaces/cache/weblog",
    "workspaces_lustre_root_dir": "/lustre/aoc/cluster/pipeline/docker/workspaces/spool",
    "current_root_directory": "current/directory",
    "current_subdirectory": "dir_123",
    "destination_dir": "/lustre/aoc/cluster/pipeline/docker/workspaces/spool/tmpabcd1234",
    "destination_subdir": "tmpabcd1234",
}

action = "Calibration Delivery"

conveyor = DeliveryConveyor(settings=settings, action=action)


class TestDeliveryConveyor:
    def test_delivery(self):
        with patch("conveyor.deliver.DeliveryConveyor.cache_weblog") as weblog:
            with patch("conveyor.deliver.DeliveryConveyor.move_subdirectories_to_qa_area") as qa2:
                conveyor.delivery()
                assert weblog.call_count == 1
                assert qa2.call_count == 1

    @patch("glob.glob")
    @patch("sys.exit")
    @patch("shutil.copytree")
    @patch("tarfile.open")
    @patch("os.chmod")
    def test_cache_weblog(self, mock_chmod, mock_tar, mock_shutil, mock_exit, mock_glob):
        with patch("os.makedirs") as make_dirs:
            with patch("conveyor.deliver.DeliveryConveyor._get_pipeline_dir") as pipeline:
                with patch("conveyor.deliver.DeliveryConveyor.create_qa_notes") as qa:
                    conveyor.cache_weblog()
                    assert make_dirs.call_count == 1
                    assert mock_shutil.call_count == 1

    def test_get_pipeline_dir(self):
        with patch("glob.glob") as glob:
            choice = conveyor._get_pipeline_dir()
            assert glob.call_count == 1
            assert choice is True

    @patch("os.chmod")
    @patch("os.chown")
    @patch("os.chdir")
    @patch("glob.glob")
    def test_create_qa_notes(self, mock_glob, mock_chdir, mock_chown, mock_chmod):
        with patch("builtins.open", mock_open()) as o:
            conveyor.create_qa_notes()
            assert mock_glob.call_count == 1
            assert mock_chdir.call_count == 2
            # assert mock_chown.call_count == 1
            assert mock_chmod.call_count == 1
            assert o.call_count == 1

    @patch("pathlib.Path.symlink_to")
    @patch("os.system")
    @patch("shutil.move")
    @patch("os.chmod")
    @patch("os.chown")
    @patch("os.makedirs")
    def test_move_subdirectories_to_qa_area(self, mock_dirs, mock_chown, mock_chmod, mock_move, mock_system, mock_link):
        with patch("glob.glob", MagicMock(return_value=["working/", "products/"])) as contents:
            path = "/lustre/aoc/cluster/pipeline/docker/workspaces/qa2/tmpabcd1234"
            conveyor.move_subdirectories_to_qa_area(path)
            assert mock_dirs.call_count == 1
            assert mock_chown.call_count == 1
            assert mock_chmod.call_count == 1
            assert mock_move.call_count == 2
            assert mock_system.call_count == 2
            assert mock_link.call_count == 2

    @patch("os.symlink")
    def symlink_qa(self, mock_link):
        conveyor.symlink_qa(settings["destination_subdir"], settings["qa_delivery_area"])
        assert mock_link.call_count == 1

    def test_convey(self):
        with patch("conveyor.deliver.DeliveryConveyor.delivery") as delivery:
            conveyor.convey()
            assert delivery.call_count == 1
