#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
"""
Tests for conveyor.retrieve module
"""
from pathlib import Path
from unittest.mock import patch

from conveyor.retrieve import RetrievalConveyor

settings = {
    "qa_delivery_area": "/lustre/aoc/cluster/pipeline/docker/workspaces/qa",
    "weblog_cache_area": "/lustre/aoc/cluster/pipeline/docker/workspaces/cache/weblog",
    "workspaces_lustre_root_dir": "/lustre/aoc/cluster/pipeline/docker/workspaces/spool",
    "current_root_directory": "current/directory",
    "current_subdirectory": "dir_123",
    "destination_dir": "/lustre/aoc/cluster/pipeline/docker/workspaces/spool/tmpabcd1234",
    "destination_subdir": "tmpabcd1234",
}

action = "Calibration Retrieval"

PRODUCTS = "products/"
WORKING = "working/"

conveyor = RetrievalConveyor(settings=settings, action=action)


def fake_determine_qa():
    qa_area = settings["qa_delivery_area"]
    wf_dir = settings["destination_subdir"]
    return Path(qa_area + "/" + wf_dir)


def fake_determine_spool():
    spool_area = settings["workspaces_lustre_root_dir"]
    wf_dir = settings["destination_subdir"]
    return Path(spool_area + "/" + wf_dir)


class TestRetrievalConveyor:
    @patch("pathlib.Path.rmdir")
    @patch("glob.glob")
    @patch("os.chdir")
    def test_retrieval(self, mock_chdir, mock_glob, mock_rm):
        with patch("conveyor.retrieve.RetrievalConveyor.break_symlinks") as symlinks:
            with patch("conveyor.retrieve.RetrievalConveyor.move_subdirectories_to_spool") as move:
                with patch("conveyor.retrieve.RetrievalConveyor.check_spool_contents") as contents:
                    conveyor.retrieval()
                    assert mock_chdir.call_count == 1
                    assert mock_glob.call_count == 1
                    assert symlinks.call_count == 1

    @patch("pathlib.Path.is_symlink", return_value=True)
    @patch("pathlib.Path.unlink")
    def test_break_symlinks(self, mock_unlink, mock_sym):
        path = fake_determine_spool()
        dir_list = [WORKING, PRODUCTS]
        conveyor.break_symlinks(path, dir_list)
        assert mock_sym.call_count == 2
        assert mock_unlink.call_count == 2

    @patch("shutil.move")
    def test_move_subdirectories_to_spool(self, mock_move):
        qa_path = fake_determine_qa()
        spool_path = fake_determine_spool()
        dir_list = [WORKING, PRODUCTS]
        conveyor.move_subdirectories_to_spool(qa_path, spool_path, dir_list)
        assert mock_move.call_count == 2

    def test_determine_qa_directory(self):
        path = conveyor.determine_qa_directory()
        assert isinstance(path, Path)

    def test_determine_spool_directory(self):
        path = conveyor.determine_spool_directory()
        assert isinstance(path, Path)

    @patch("glob.glob", return_value=[WORKING, PRODUCTS])
    @patch("os.chdir")
    def test_check_spool_contents(self, mock_chdir, mock_glob):
        spool = fake_determine_spool()
        expected_list = [WORKING, PRODUCTS]
        check = conveyor.check_spool_contents(spool, expected_list)
        assert check is True

    def test_convey(self):
        with patch("conveyor.retrieve.RetrievalConveyor.retrieval") as retrieval:
            conveyor.convey()
            assert retrieval.call_count == 1
