def retrieve_problem_types(cursor):
    """
    Retrieves all valid problem types from the database.

    :param connection: a MDDBCursor object to interface with the database
    :return: a list of tuples representing problem type descriptions and their database ids
    """
    cursor.execute("SELECT * FROM problem_types")
    rows = cursor.fetchall()
    return rows


def retrieve_problem_severities(cursor):
    """
    Retrieves all valid problem severities from the database.

    :param cursor: a MDDBCursor object to interface with the database
    :return: a list of tuples representing problem severity descriptions and their database ids
    """
    cursor.execute("SELECT * FROM problem_severity_list")
    rows = cursor.fetchall()
    return rows


def retrieve_acknowledgements(cursor):
    """
    Retrieves all valid data acknowledgements from the database.

    :param cursor: a MDDBCursor object to interface with the database
    :return: a list of tuples representing acknowledgements, their institutions, and their database ids
    """
    cursor.execute("SELECT * FROM acknowledgements")
    rows = cursor.fetchall()
    return rows


def get_exec_blocks(project_code, date_range, external_name, band, configuration, cursor):
    """
    Finds all execution blocks associated with a user's query by building and executing a SQL query

    :param project_code: a string corresponding to the project_code field in the database
    :param date_range: a tuple corresponding to the starttime and endtime fields in the database
    :param external_name: a string corresponding to the external name field in the database
    :param band: a string corresponding to the band field in the database
    :param configuration: a string corresponding to the configuration field in the database
    :param cursor: a MDDBCursor object to interface with the database
    :return: a list of spls for all execution blocks, images, and calibrations related to the initial query
    """
    # Format WHERE clauses with provided search pieces
    project_code = f"project_code='{project_code}' AND " if project_code else ""
    date_range = f"starttime>={date_range[0]} AND endtime<={date_range[1]} AND " if date_range else ""
    external_name = f"external_name='{external_name}' AND " if external_name else ""
    configuration = f"configuration='{configuration}' AND " if configuration else ""
    band = f"band='{band}'" if band else ""

    search_command = f"SELECT execution_block_id, execution_blocks.science_product_locator FROM execution_blocks \
        INNER JOIN science_products ON execution_blocks.science_product_locator=science_products.science_product_locator \
        WHERE {project_code}{configuration}{date_range}{external_name}{band}"

    # Remove trailing ANDs, if necessary
    if search_command[-4:] == "AND ":
        search_command = search_command[:-4]

    cursor.execute(search_command)
    return cursor.fetchall()


def get_all_spls_from_exec_blocks(ebs, cursor):
    """
    Finds all science product locators associated with a list of execution blocks

    :param ebs: a list tuples representing execution blocks
    :param cursor: a MDDBCursor object to interface with the database
    :return: a list of spls for all execution blocks, images, and calibrations related to the initial query
    """
    ebs_to_find = []
    spls_to_return = []
    for eb in ebs:
        ebs_to_find.append(str(eb[0]))
        spls_to_return.append(str(eb[1]))

    print("Finding all exec blocks, cals, and images fitting the passed in search terms...")
    spl_query = f"""
        with recursive
            input_outputs as (select sppg2.science_product_locator as input_spl, sppg1.science_product_locator as output_spl
                              from science_products_product_groups sppg1
                                       join product_groups pg on pg.product_group_id = sppg1.product_group_id
                                       join product_groups ppg on ppg.product_group_id = pg.parent_product_group_id
                                       join science_products_product_groups sppg2
                                            on sppg2.product_group_id = ppg.product_group_id),
            all_inputs_outputs as (select *
                                   from input_outputs
                                   union all
                                   select input_outputs.input_spl, all_inputs_outputs.output_spl
                                   from input_outputs
                                            join all_inputs_outputs on input_outputs.output_spl = all_inputs_outputs.input_spl)
        select distinct all_inputs_outputs.output_spl
            from all_inputs_outputs, execution_blocks eb
        where
            eb.science_product_locator = all_inputs_outputs.input_spl AND
            eb.execution_block_id in ({",".join(ebs_to_find)})
    """

    cursor.execute(spl_query)
    rows = cursor.fetchall()
    for row in rows:
        spls_to_return.append(row[0])

    return spls_to_return


def create_comments(ebs, comment, problem_severity, problem_type, cursor):
    """
    Creates a list of data comments and inserts them into the database.

    :param ebs: a list execution block ids
    :param comment: a string representing the comment to create
    :param problem_severity: an integer representing the problem severity id
    :param problem_type: an integer representing the problem type id
    :param cursor: a MDDBCursor object to interface with the database
    """
    print("Creating comments...")
    comments_to_add = []
    for eb in ebs:
        comments_to_add.append((problem_type, eb, problem_severity, comment))

    cursor.executemany("INSERT INTO science_product_comments VALUES(DEFAULT,%s,%s,%s,%s)", comments_to_add)
    print(f"Created comments for {cursor.rowcount} total rows...")


def delete_comments(ebs, comment, problem_severity, problem_type, cursor):
    """
    Finds a list of comments in the database and deletes them.

    :param ebs: a list execution block ids
    :param comment: a string representing the comment to delete
    :param problem_severity: an integer representing the problem severity id
    :param problem_type: an integer representing the problem type id
    :param cursor: a MDDBCursor object to interface with the database
    """
    print("Deleting comments...")
    comments_to_add = []
    for eb in ebs:
        comments_to_add.append((problem_type, eb, problem_severity, comment))

    cursor.executemany(
        f"DELETE FROM science_product_comments WHERE problem_type_id=%s AND science_product_locator=%s AND severity=%s AND comment=%s",
        comments_to_add,
    )
    print(f"Deleted comments for {cursor.rowcount} total rows...")


def modify_comments(ebs, new_comment, old_comment, problem_severity, problem_type, cursor):
    """
    Finds a list of comments in the database and modifies the values.

    :param ebs: a list execution block ids
    :param new_comment: a string representing the comment to be inserted
    :param old_comment: a string representing the comment to be replaced
    :param problem_severity: an integer representing the problem severity id
    :param problem_type: an integer representing the problem type id
    :param cursor: a MDDBCursor object to interface with the database
    """
    print("Modifying comments...")
    comments_to_add = []
    for eb in ebs:
        comments_to_add.append((new_comment, problem_type, eb, problem_severity, old_comment))

    cursor.executemany(
        f"UPDATE science_product_comments SET comment=%s WHERE problem_type_id=%s AND science_product_locator=%s AND severity=%s AND comment=%s",
        comments_to_add,
    )
    print(f"Modified comments for {cursor.rowcount} total rows...")


def create_acknowledgement(institution, acknowledgement_text, cursor):
    """
    Create an acknowledgement in the database.

    :param institution: a string representing the institution associated with the acknowledgement to create
    :param acknowledgement_text: a string representing the desired acknowledgement text
    :param cursor: a MDDBCursor object to interface with the database
    """
    print("Creating acknowledgement...")
    cursor.execute(
        f"INSERT INTO acknowledgements (institution, acknowledgement_text) "
        f"VALUES('{institution}','{acknowledgement_text}')"
    )
    print(f"Created one acknowledgement for {institution}...")


def modify_acknowledgement(acknowledgement_id, acknowledgement_text, cursor):
    """
    Finds an acknowledgement in the database and modifies it.

    :param acknowledgement_id: an integer representing the acknowledgement to modify
    :param acknowledgement_text: a string representing the new acknowledgement text to set
    :param cursor: a MDDBCursor object to interface with the database
    """
    print("Modifying acknowledgement...")
    cursor.execute(
        f"UPDATE acknowledgements SET acknowledgement_text='{acknowledgement_text}' "
        f"WHERE acknowledgement_id='{acknowledgement_id}'"
    )
    print(f"Modified acknowledgement #{acknowledgement_id}...")


def delete_acknowledgement(acknowledgement_id, cursor):
    """
    Finds an acknowledgement in the database and deletes it.

    :param acknowledgement_id: an integer representing the acknowledgement to delete
    :param cursor: a MDDBCursor object to interface with the database
    """
    print("Deleting acknowledgement...")
    cursor.execute(f"DELETE FROM acknowledgements WHERE acknowledgement_id='{acknowledgement_id}'")
    print(f"Deleted acknowledgement #{acknowledgement_id}...")


def assign_acknowledgement_to_project(acknowledgement_id, project_code, cursor):
    """
    Assign an acknowledgement to a project in the database.

    :param acknowledgement_id: an integer representing the acknowledgement to assign to
    :param project_code: a string representing the project code to assign the acknowledgement to
    :param cursor: a MDDBCursor object to interface with the database
    """
    print("Assigning acknowledgement...")
    cursor.execute(f"INSERT INTO projects_acknowledgements VALUES('{acknowledgement_id}','{project_code}')")
    print(f"Assigned acknowledgement #{acknowledgement_id} to project {project_code}...")


def delete_acknowledgement_from_project(acknowledgement_id, project_code, cursor):
    """
    Remove an acknowledgement from a project in the database.

    :param acknowledgement_id: an integer representing the acknowledgement to assign to
    :param project_code: a string representing the project code to assign the acknowledgement to
    :param cursor: a MDDBCursor object to interface with the database
    """
    print("Deleting acknowledgement...")
    cursor.execute(
        f"DELETE FROM projects_acknowledgements "
        f"WHERE acknowledgement_id='{acknowledgement_id}' AND project_code='{project_code}'"
    )
    print(f"Deleted acknowledgement #{acknowledgement_id} from project {project_code}...")
