#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
import pathlib
import shutil
from test import (
    RESTORE_AUXPRODUCTS_NAME,
    TEST_RESTORE_METADATA,
    TEST_VLASS_COARSE_CUBE_METADATA,
    get_expected_files_and_dirs_for_restore,
)
from test.conftest import assert_restore_delivered_only_expected_files
from typing import Dict

from delivery.context import DeliveryContext
from delivery.destinations.interfaces import (
    DeliveryContextIF,
    Destination,
    DestinationTempFile,
)
from delivery.finder import RestoreProductFinder
from delivery.products import ArchiveProduct, Calibration, RestoreProduct


class FakeDestination(Destination):
    def __init__(self, context: DeliveryContextIF):
        super().__init__(context)
        self.files_added: dict[str, pathlib.Path] = {}

    def add_file(self, file: pathlib.Path, relative_path: str):
        if relative_path in self.files_added:
            # Prevent future tests from passing when they try to deliver files with the same name to the same dir
            raise ValueError(
                f"Path {relative_path} already exists at destination with value {self.files_added[relative_path]}"
            )
        self.files_added[relative_path] = file

    def create_file(self, relative_path: str) -> DestinationTempFile:
        raise NotImplementedError

    def results(self) -> Dict:
        return {}


def test_web_calibration(resource_path_root):
    calibration = Calibration(resource_path_root / "calibrations" / "foo")

    # first, let's make sure we parsed the metadata properly
    project_code, pipeline_id = calibration.parse_metadata()
    assert project_code == "21A-409"
    assert pipeline_id == "59311.953874212966"

    # OK now let's make sure that we're delivering properly
    destination = FakeDestination(DeliveryContext(calibration.path))
    calibration.deliver_to(destination)

    expected_files = set(
        [
            "21A-409/calibration_pipeline.59311.953874212966/pipeline-20210409T164741/.keep",
            "21A-409/calibration_pipeline.59311.953874212966/unknown.hifv_cal.pipeline_manifest.xml",
            "21A-409/calibration_pipeline.59311.953874212966/casa_commands.log",
            "21A-409/calibration_pipeline.59311.953874212966/flux.csv",
            "21A-409/calibration_pipeline.59311.953874212966/unknown.hifv_cal.pprequest.xml",
            "21A-409/calibration_pipeline.59311.953874212966/21A-409.sb39530507.eb39564922.59311.953874212966.ms.calapply.txt",
            "21A-409/calibration_pipeline.59311.953874212966/casa_pipescript.py",
            "21A-409/calibration_pipeline.59311.953874212966/casa_piperestorescript.py",
            "21A-409/calibration_pipeline.59311.953874212966/PPR.xml",
            "21A-409/calibration_pipeline.59311.953874212966/pipeline_aquareport.xml",
            "21A-409/calibration_pipeline.59311.953874212966/weblog.tgz",
        ]
    )
    actual_files = destination.files_added.keys()
    assert sorted(actual_files) == sorted(expected_files)


def test_vlass_coarse_cube_flatten(vlass_coarse_cube_with_ancillaries_dir: pathlib.Path):
    cube_product = ArchiveProduct(
        vlass_coarse_cube_with_ancillaries_dir / "a650ba0e-3b81-4f6f-8f39-35982f03a717",
        TEST_VLASS_COARSE_CUBE_METADATA,
        True,
    )

    # OK now let's make sure that we're delivering properly
    destination = FakeDestination(DeliveryContext(cube_product.path))
    cube_product.deliver_to(destination)

    deliver_rel_path_root = f"{TEST_VLASS_COARSE_CUBE_METADATA.project_code}/{TEST_VLASS_COARSE_CUBE_METADATA.pipeline_spec}//"  # Extra slash because ProductMetadata.to_path seems to add one because of `$PRODUCT`
    expected_files = set(
        [
            f"{deliver_rel_path_root}CCIP_parameter.list",
            f"{deliver_rel_path_root}SEIP_parameter.list",
            f"{deliver_rel_path_root}VLASS2.1.cc.T10t02.J010200-013000.06.2048.V_mean_minmaxrej.subim.fits",
            f"{deliver_rel_path_root}VLASS2.1.cc.T10t02.J010200-013000.06.2048.v1.spw10.IQU.iter3.image.pbcor.tt0.subim.fits",
            f"{deliver_rel_path_root}VLASS2.1.cc.T10t02.J010200-013000.06.2048.v1.spw11.IQU.iter3.image.pbcor.tt0.subim.fits",
            f"{deliver_rel_path_root}VLASS2.1.cc.T10t02.J010200-013000.06.2048.v1.spw12.IQU.iter3.image.pbcor.tt0.subim.fits",
            f"{deliver_rel_path_root}VLASS2.1.cc.T10t02.J010200-013000.06.2048.v1.spw13.IQU.iter3.image.pbcor.tt0.subim.fits",
            f"{deliver_rel_path_root}VLASS2.1.cc.T10t02.J010200-013000.06.2048.v1.spw14.IQU.iter3.image.pbcor.tt0.subim.fits",
            f"{deliver_rel_path_root}VLASS2.1.cc.T10t02.J010200-013000.06.2048.v1.spw15.IQU.iter3.image.pbcor.tt0.subim.fits",
            f"{deliver_rel_path_root}VLASS2.1.cc.T10t02.J010200-013000.06.2048.v1.spw2.IQU.iter3.image.pbcor.tt0.subim.fits",
            f"{deliver_rel_path_root}VLASS2.1.cc.T10t02.J010200-013000.06.2048.v1.spw3.IQU.iter3.image.pbcor.tt0.subim.fits",
            f"{deliver_rel_path_root}VLASS2.1.cc.T10t02.J010200-013000.06.2048.v1.spw5.IQU.iter3.image.pbcor.tt0.subim.fits",
            f"{deliver_rel_path_root}VLASS2.1.cc.T10t02.J010200-013000.06.2048.v1.spw6.IQU.iter3.image.pbcor.tt0.subim.fits",
            f"{deliver_rel_path_root}VLASS2.1.cc.T10t02.J010200-013000.06.2048.v1.spw7.IQU.iter3.image.pbcor.tt0.subim.fits",
            f"{deliver_rel_path_root}VLASS2.1.cc.T10t02.J010200-013000.06.2048.v1.spw8.IQU.iter3.image.pbcor.tt0.subim.fits",
            f"{deliver_rel_path_root}VLASS2.1.cc.T10t02.J010200-013000.06.2048.v1.spw9.IQU.iter3.image.pbcor.tt0.subim.fits",
            f"{deliver_rel_path_root}VLASS2.1.cc.T10t02.J010200-013000.06.2048.v1.spw10.IQU.iter3.image.pbcor.tt0.rms.subim.fits",
            f"{deliver_rel_path_root}VLASS2.1.cc.T10t02.J010200-013000.06.2048.v1.spw11.IQU.iter3.image.pbcor.tt0.rms.subim.fits",
            f"{deliver_rel_path_root}VLASS2.1.cc.T10t02.J010200-013000.06.2048.v1.spw12.IQU.iter3.image.pbcor.tt0.rms.subim.fits",
            f"{deliver_rel_path_root}VLASS2.1.cc.T10t02.J010200-013000.06.2048.v1.spw13.IQU.iter3.image.pbcor.tt0.rms.subim.fits",
            f"{deliver_rel_path_root}VLASS2.1.cc.T10t02.J010200-013000.06.2048.v1.spw14.IQU.iter3.image.pbcor.tt0.rms.subim.fits",
            f"{deliver_rel_path_root}VLASS2.1.cc.T10t02.J010200-013000.06.2048.v1.spw15.IQU.iter3.image.pbcor.tt0.rms.subim.fits",
            f"{deliver_rel_path_root}VLASS2.1.cc.T10t02.J010200-013000.06.2048.v1.spw2.IQU.iter3.image.pbcor.tt0.rms.subim.fits",
            f"{deliver_rel_path_root}VLASS2.1.cc.T10t02.J010200-013000.06.2048.v1.spw3.IQU.iter3.image.pbcor.tt0.rms.subim.fits",
            f"{deliver_rel_path_root}VLASS2.1.cc.T10t02.J010200-013000.06.2048.v1.spw5.IQU.iter3.image.pbcor.tt0.rms.subim.fits",
            f"{deliver_rel_path_root}VLASS2.1.cc.T10t02.J010200-013000.06.2048.v1.spw6.IQU.iter3.image.pbcor.tt0.rms.subim.fits",
            f"{deliver_rel_path_root}VLASS2.1.cc.T10t02.J010200-013000.06.2048.v1.spw7.IQU.iter3.image.pbcor.tt0.rms.subim.fits",
            f"{deliver_rel_path_root}VLASS2.1.cc.T10t02.J010200-013000.06.2048.v1.spw8.IQU.iter3.image.pbcor.tt0.rms.subim.fits",
            f"{deliver_rel_path_root}VLASS2.1.cc.T10t02.J010200-013000.06.2048.v1.spw9.IQU.iter3.image.pbcor.tt0.rms.subim.fits",
            f"{deliver_rel_path_root}pipeline_artifacts.tar",
            f"{deliver_rel_path_root}weblog.tgz",
        ]
    )
    actual_files = destination.files_added.keys()
    assert sorted(actual_files) == sorted(expected_files)


def test_restore(restore_directory: pathlib.Path):
    restore = RestoreProduct(restore_directory, TEST_RESTORE_METADATA)

    # OK now let's make sure that we're delivering properly
    destination = FakeDestination(DeliveryContext(restore_directory))
    restore.deliver_to(destination)

    deliver_rel_path_root = f"{TEST_RESTORE_METADATA.project_code}/{TEST_RESTORE_METADATA.pipeline_spec}"
    # Just check the file counts in each dir since they contain a lot of files
    actual_files = list(pathlib.Path(f) for f in destination.files_added.keys())
    files_to_is_file = {pathlib.Path(key): value.is_file() for key, value in destination.files_added.items()}
    expected_files, expected_dirs_to_fcs = get_expected_files_and_dirs_for_restore(deliver_rel_path_root)
    assert_restore_delivered_only_expected_files(actual_files, expected_files, expected_dirs_to_fcs, files_to_is_file)


def test_restore_with_flagtemplate(restore_directory: pathlib.Path, tmp_path: pathlib.Path):
    auxproducts_path = restore_directory.parent / RESTORE_AUXPRODUCTS_NAME
    assert auxproducts_path.is_file()
    # To avoid possibly polluting the repo,
    #     use a tmp dir for the restore since the auxproducts tar needs to be copied into it
    shutil.copytree(restore_directory, tmp_path, dirs_exist_ok=True)
    shutil.copy2(auxproducts_path, tmp_path / "products")

    restore = RestoreProduct(tmp_path, TEST_RESTORE_METADATA)

    # OK now let's make sure that we're delivering properly
    destination = FakeDestination(DeliveryContext(tmp_path))
    restore.deliver_to(destination)

    deliver_rel_path_root = f"{TEST_RESTORE_METADATA.project_code}/{TEST_RESTORE_METADATA.pipeline_spec}"
    # Just check the file counts in each dir since they contain a lot of files
    actual_files = list(pathlib.Path(f) for f in destination.files_added.keys())
    files_to_is_file = {pathlib.Path(key): value.is_file() for key, value in destination.files_added.items()}
    expected_files, expected_dirs_to_fcs = get_expected_files_and_dirs_for_restore(
        deliver_rel_path_root, do_include_flagtemplate=True
    )
    assert_restore_delivered_only_expected_files(actual_files, expected_files, expected_dirs_to_fcs, files_to_is_file)


def test_restore_product_finder_parse_metadata(restore_directory: pathlib.Path):
    expected_metadata = TEST_RESTORE_METADATA
    finder = RestoreProductFinder(restore_directory)
    actual_metadata = finder.parse_metadata()
    assert actual_metadata == expected_metadata
