#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
import logging
import subprocess
import sys
from pathlib import Path
from typing import Union

from ingest_envoy.collectors import (
    CoarseCollector,
    ImageCollector,
    ObservationCollector,
    SECICollector,
    collect_image_metadata,
)
from ingest_envoy.ingestion_manifest import IngestionManifestBuilder
from ingest_envoy.interfaces import LauncherIF
from ingest_envoy.schema import AbstractTextFile
from ingest_envoy.utilities import CuratorType, IngestType, VLASSIngestType


def trigger_ingest(real_ingest: bool, staging_dir: str, bin_dir: str = ".") -> int:
    """
    Run ingest

    :param real_ingest: real ingestion vs. testing flag
    :param staging_dir: staging directory to ingest from
    :param bin_dir: directory containing the ingestion utility
    :return: return code
    """
    if real_ingest:
        ingest_process = subprocess.run(
            [f"{bin_dir}/ingest", "--json", "-p", staging_dir],
            stdout=sys.stdout,
            stderr=sys.stderr,
        )
        return ingest_process.returncode
    else:
        # For local testing
        return 0


def trigger_curator(real_ingest: bool, manifest_dir: Path, bin_dir: str = ".") -> int:
    """
    Run curator

    :param real_ingest: real curation or testing only?
    :param manifest_dir: The directory containing the manifest
    :param bin_dir: directory containing the curator utility
    :return: return code
    """

    if real_ingest:
        curator_process = subprocess.run(
            [f"{bin_dir}/curator", "-m", f"{manifest_dir / 'manifest.json'}"],
            stdout=sys.stdout,
            stderr=sys.stderr,
        )
        return curator_process.returncode
    else:
        # for local testing
        return 0


class IngestCalibrationLauncher(LauncherIF):
    """Setup and Launch Calibration Ingestion"""

    def __init__(self, arg_type: IngestType, parameters: dict):
        self.logger = logging.getLogger("ingest_envoy")
        self.sci_product_type = arg_type.value.__str__()
        self.parameters = parameters
        self.staging_source_dir = self.parameters["staging_area"] + "/" + self.parameters["workflowDir"]

    def launch(self) -> (int, bool):
        """
        Prepare for and run ingest

        :return: (return code from ingest, whether to run the indexer)
        """
        self.logger.info("RUNNING CALIBRATION INGESTION!")
        self.prepare_for_launch()

        self.logger.info("Running ingest!")
        return (
            trigger_ingest(self.parameters["useIngest"], self.staging_source_dir, self.parameters["script_location"]),
            True,
        )

    def prepare_for_launch(self):
        """
        Prepare files for ingestion, creation of collection tars and manifest

        :return:
        """
        self.logger.info("Preparing for ingest...")
        # 1. run collection script to create calibration tarfile
        self.run_collection_script()

        # 2. create ingestion manifest
        self.create_manifest()

    def run_collection_script(self):
        """
        Run calibration collection script, creates collection tars
        ex. caltables, weblogs, pipeline-artifacts, etc

        If this was a multi-version request, also package up the initial version weblog

        :return:
        """
        self.logger.info("Collecting calibration tables for staging...")
        workflow_dir = self.parameters["workflowDir"]
        sbin_dir = self.parameters["script_location"]
        sdm_id = self.parameters["sdmId"]
        cal_processing_datetime = self.parameters["processingStart"]

        # run script
        if self.parameters["multiVersion"] is False:
            collector = subprocess.run(
                [f"{sbin_dir}/calibration-table-collector.sh", workflow_dir, sdm_id, cal_processing_datetime],
                stdout=sys.stdout,
                stderr=sys.stderr,
            )
        else:
            initial_version_dir = self.parameters["initialVersionDir"]

            collector = subprocess.run(
                [
                    f"{sbin_dir}/calibration-table-collector.sh",
                    workflow_dir,
                    sdm_id,
                    cal_processing_datetime,
                    initial_version_dir,
                ],
                stdout=sys.stdout,
                stderr=sys.stderr,
            )
        if collector.returncode != 0:
            self.logger.error("ERROR: Calibration product collection failed!")
            exit(1)

    def create_manifest(self, additional_file=None):
        """
        Create the calibration ingestion manifest

        :param additional_file: Unneeded for calibration ingestion
        :return:
        """
        self.logger.info("Creating ingestion manifest...")
        spl = self.parameters["spl"]
        telescope = self.parameters["telescope"]

        IngestionManifestBuilder(
            manifest_source_dir=Path(self.staging_source_dir),
            sp_type=self.sci_product_type,
            locator=spl,
            telescope=telescope,
        ).build()


class IngestVLASSLauncher(LauncherIF):
    """Setup and Launch VLASS Caching and Ingestion"""

    def __init__(self, arg_type: VLASSIngestType, parameters: dict):
        self.logger = logging.getLogger("ingest_envoy")
        self.sci_product_type = arg_type.value.__str__()
        self.parameters = parameters
        self.staging_source_dir = self.parameters["staging_area"] + "/" + self.parameters["workflowDir"]
        self.qa_passed = parameters["qa_status"] == "PASS"

        match arg_type:
            case VLASSIngestType.COARSE:
                self.collector = CoarseCollector(self.parameters)
            case _:
                raise NotImplementedError(f"Error: {arg_type.value} is not currently supported.")

    def launch(self) -> (int, bool):
        """
        Prepare and run ingestion script

        :return: Return code from ingestion script process
        """
        self.logger.info(f"RUNNING {self.sci_product_type} CACHING AND INGESTION!")
        self.prepare_for_launch()

        if self.qa_passed:
            self.logger.info("Running ingest!")
            return (
                trigger_ingest(
                    self.parameters["useIngest"], self.staging_source_dir, self.parameters["script_location"]
                ),
                True,
            )

        # In the case that we didn't run ingest for a coarse cube due to QA fail
        # Return 0 because this is expected behavior
        self.logger.info("Not an error; ingest not run for a coarse cube due to QA fail.")
        return 0, False

    def prepare_for_launch(self):
        """
        Prepare files for ingestion, creation of collection tars and manifest

        :return:
        """
        self.logger.info("Preparing for ingest...")
        # 1. collect products and auxiliary image metadata
        aux_file = self.run_collector()
        # 2. create manifest
        if self.qa_passed:
            self.create_manifest(additional_file=aux_file)

    def run_collector(self) -> AbstractTextFile:
        """
        Run ImageCollector

        :return: AbstractTextFile containing additional metadata needed for manifest creation
        """
        # 1. collect products to staging area
        self.collector.collect_products()
        # 2. create auxiliary metadata file and return
        return collect_image_metadata()

    def create_manifest(self, additional_file: AbstractTextFile = None):
        """
        Create VLASS cube ingestion manifest

        :param additional_file: image ingestion requires additional metadata
        :return:
        """
        self.logger.info("Creating ingestion manifest...")
        # Input product locator
        precursor_spl = self.parameters["precursorSpl"]
        telescope = self.parameters["telescope"]
        additional_metadata = additional_file

        IngestionManifestBuilder(
            manifest_source_dir=Path(self.staging_source_dir),
            sp_type=self.sci_product_type,
            locator=precursor_spl,
            telescope=telescope,
            additional_metadata=additional_metadata,
        ).build()


class IngestImageLauncher(LauncherIF):
    """Setup and Launch Image Ingestion"""

    def __init__(self, arg_type: Union[IngestType, VLASSIngestType], parameters: dict):
        self.logger = logging.getLogger("ingest_envoy")
        self.sci_product_type = arg_type.value.__str__()
        self.parameters = parameters
        self.staging_source_dir = self.parameters["staging_area"] + "/" + self.parameters["workflowDir"]

        match arg_type:
            case VLASSIngestType.SECI:
                self.collector = SECICollector(self.parameters)
            case IngestType.IMG:
                self.collector = ImageCollector(self.parameters)
            case _:
                raise NotImplementedError(f"Error: {arg_type.value} is not currently supported for an image launcher.")

    def launch(self) -> (int, bool):
        """
        Prepare and run ingestion script

        :return: Return code from ingestion script process
        """
        self.logger.info(f"RUNNING {self.sci_product_type} INGESTION!")
        self.prepare_for_launch()

        self.logger.info("Running ingest!")
        return (
            trigger_ingest(self.parameters["useIngest"], self.staging_source_dir, self.parameters["script_location"]),
            True,
        )

    def prepare_for_launch(self):
        """
        Prepare files for ingestion, creation of collection tars and manifest

        :return:
        """
        self.logger.info("Preparing for ingest...")
        # 1. collect products and auxiliary image metadata
        aux_file = self.run_collector()
        # 2. create manifest
        self.create_manifest(additional_file=aux_file)

    def run_collector(self) -> AbstractTextFile:
        """
        Run ImageCollector

        :return: AbstractTextFile containing additional metadata needed for manifest creation
        """
        # 1. collect products to staging area
        self.collector.collect_products()
        # 2. create auxiliary metadata file and return
        return collect_image_metadata()

    def create_manifest(self, additional_file: AbstractTextFile = None):
        """
        Create image ingestion manifest

        :param additional_file: image ingestion requires additional metadata
        :return:
        """
        self.logger.info("Creating ingestion manifest...")
        # needs to be the CAL locator!!
        spl = self.parameters["calSpl"]
        telescope = self.parameters["telescope"]
        additional_metadata = additional_file

        IngestionManifestBuilder(
            manifest_source_dir=Path(self.staging_source_dir),
            sp_type=self.sci_product_type,
            locator=spl,
            telescope=telescope,
            additional_metadata=additional_metadata,
        ).build()


class IngestObservationLauncher(LauncherIF):
    """Setup and Launch Observation Ingestion"""

    def __init__(self, arg_type: IngestType, parameters: dict):
        self.logger = logging.getLogger("ingest_envoy")
        self.sci_product_type = arg_type.value.__str__()
        self.parameters = parameters
        self.staging_source_dir = self.parameters["staging_area"] + "/" + self.parameters["workflowDir"]
        self.collector = ObservationCollector(self.parameters)

    def launch(self) -> (int, bool):
        """
        Prepare for and run ingest

        :return: (return code from ingest, whether to run the indexer)
        """
        self.logger.info("RUNNING OBSERVATION INGESTION!")
        self.prepare_for_launch()

        self.logger.info("Running ingest!")
        return (
            trigger_ingest(self.parameters["useIngest"], self.staging_source_dir, self.parameters["script_location"]),
            True,
        )

    def prepare_for_launch(self):
        """
        Prepare files for ingestion, creation of collection tars and manifest

        :return:
        """
        self.logger.info("Preparing for ingest...")
        # 1. run collection script to move observation data to the staging area
        self.run_collector()

        # 2. create ingestion manifest
        self.create_manifest()

    def run_collector(self):
        """
        Run ObservationCollector
        """
        # 1. collect products to staging area
        self.collector.collect_products()

    def create_manifest(self, additional_file=None):
        """
        Create the observation ingestion manifest

        :param additional_file: Unneeded for observation ingestion
        :return:
        """
        self.logger.info("Creating ingestion manifest...")
        telescope = self.parameters["telescope"]
        filename = self.parameters["filename"]
        IngestionManifestBuilder(
            manifest_source_dir=Path(self.staging_source_dir),
            sp_type=self.sci_product_type,
            locator=None,
            telescope=telescope,
            filename=filename,
        ).build()


class CuratorLauncher(LauncherIF):
    """Setup and Launch Curator"""

    def __init__(self, arg_type: CuratorType, parameters: dict):
        self.logger = logging.getLogger("ingest_envoy")
        self.curator_type = arg_type
        self.sci_product_type = parameters["product_type"]
        self.curation_source = parameters["curation_source"]
        self.manifest_destination_dir = Path.cwd()
        self.parameters = parameters

    def prepare_for_launch(self):
        """
        Prepare for curation, manifest building

        :return:
        """
        self.logger.info("Preparing for curator...")
        #  no collection needed for curation, manifest building only
        self.create_manifest()

    def create_manifest(self, additional_file: AbstractTextFile = None):
        """
        Create the curation manifest

        :param additional_file: Unneeded for curation
        :return:
        """
        self.logger.info("Creating curation manifest...")
        spl = self.parameters["spl"]
        telescope = self.parameters["telescope"]
        curation_source = self.parameters["curation_source"]
        target_list = self.parameters["target_list"]
        file_list = self.parameters["file_list"]
        product_group = self.parameters["product_group_id"]
        input_group_locator = self.parameters["input_group_locator"]

        IngestionManifestBuilder(
            manifest_source_dir=curation_source,
            sp_type=self.sci_product_type,
            locator=spl,
            telescope=telescope,
            curate=(self.curator_type, target_list),
            manifest_destination_dir=self.manifest_destination_dir,
            file_list=file_list,
            product_group=product_group,
            input_group_locator=input_group_locator,
        ).build()

    def launch(self) -> (int, bool):
        """
        Prepare for and run ingest

        :return: (return code from ingest, whether to run the indexer)
        """
        self.logger.info(f"RUNNING {self.sci_product_type.upper()} CURATION!")
        self.prepare_for_launch()

        self.logger.info("Running curator!")
        return (
            trigger_curator(
                self.parameters["useIngest"], self.manifest_destination_dir, self.parameters["script_location"]
            ),
            True,
        )
