#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.

""" Helper for building image ingestion manifest """
import logging
from pathlib import Path

from ingest_envoy.manifest_components import (
    PARAMETER_FILENAME,
    REIMAGING_FILENAME,
    SELFCAL_AUX_FILENAME,
    TARFILE_EXT,
    WEBLOG_FILENAME,
    AncillaryProduct,
    OutputScienceProduct,
)
from ingest_envoy.utilities import AncillaryProductType, ScienceProductType

FITS = "fits"
CSV = "csv"
TT0 = "tt0"
RMS = "rms"
PB = "pb"
MASK = "mask"
ALPHA = "alpha"
PBCOR = "pbcor"


# pylint: disable=R1721
class ImageIngestionProductsFinder:
    """Finds ancillary science products and other ancillary products needed for image ingestion"""

    def __init__(self, files_found: list[Path], sp_type: ScienceProductType):
        self.logger = logging.getLogger("ingest_envoy")
        self.sp_type = sp_type
        self.files_found = files_found
        self.output_science_products = self._find_output_science_products()
        self.ancillary_products = find_group_level_ancillary_products(
            self.files_found, self.sp_type in [ScienceProductType.VLASS_SECI, ScienceProductType.VLASS_COARSE]
        )

    def _find_output_science_products(self) -> list[OutputScienceProduct]:
        """
        Find the ancillary products belonging to the science product*
        in the staging dir.

        N.B. ingestion artifacts file will not be there until manifest has been created!

        :return: The output science products as a list
        """
        if self.sp_type == ScienceProductType.VLASS_SECI:
            sp_image_files = [
                file
                for file in self.files_found
                if file.name.endswith(FITS) and TT0 in file.name and RMS not in file.name
            ]
        else:
            # The science products will be named *pbcor.tt0.fits
            sp_image_files = [
                file
                for file in self.files_found
                if file.name.endswith(FITS) and PBCOR in file.name and TT0 in file.name
            ]
        self.logger.info(f"Science Product(s) to ingest: {sp_image_files}")
        image_files = [
            file
            for file in self.files_found
            if (file.name.endswith(FITS) or file.name.endswith(".png") or file.name.endswith(CSV))
            and file not in sp_image_files
        ]
        self.logger.info(f"Science Product ancillaries are: {image_files}")

        # Build a mapping of science products -> ancillaries
        sp_aps = {k: [] for k in sp_image_files}
        for image_file in image_files:
            sp_ap = self._build_ancillary_image_science_product(image_file)
            for sp in sp_aps.keys():
                if image_file.name.startswith(sp.name.split(PBCOR)[0]):
                    # Add this as an ancillary to a similarly named science product
                    sp_aps.get(sp).append(sp_ap)
                    break
                elif sp == next(reversed(sp_aps.keys())):
                    # No similar names found, add this as an ancillary to the first sp (covers VLASS_SECI too)
                    next(iter(sp_aps.values())).append(sp_ap)
        self.logger.info(f"Science Product -> Ancillary mapping: {sp_aps}")

        science_products = []
        for file in sp_image_files:
            # Add all science products and their ancillaries
            science_products.append(
                OutputScienceProduct(
                    type=AncillaryProductType.FITS,
                    filename=file.name,
                    ancillary_products=sp_aps.get(file),
                )
            )

        return science_products

    @staticmethod
    def _build_ancillary_image_science_product(file: Path):
        """
        Image science products will have ancillary products of their very own,
        distinct from other ancillary products that might be in the staging dir.

        :param file: a possible ancillary image product
        :return: the corresponding AncillaryProduct, if applicable
        """
        filename = file.name
        if "image" in filename:
            if filename.endswith(".png"):
                return AncillaryProduct(type=AncillaryProductType.THUMBNAIL_IMG, filename=filename)
            if filename.endswith(FITS) and RMS in filename:
                return AncillaryProduct(type=AncillaryProductType.RMS_IMAGE, filename=filename)
            return AncillaryProduct(type=AncillaryProductType.FITS, filename=filename)
        if filename.endswith(FITS):
            if PB in filename:
                return AncillaryProduct(type=AncillaryProductType.PB_FITS, filename=filename)
            if MASK in filename:
                return AncillaryProduct(type=AncillaryProductType.FITS_MASK, filename=filename)
            if ALPHA in filename:
                return AncillaryProduct(type=AncillaryProductType.ALPHA_FITS, filename=filename)
            else:
                return AncillaryProduct(type=AncillaryProductType.FITS, filename=filename)
        if filename.endswith(CSV):
            return AncillaryProduct(type=AncillaryProductType.COMPONENT_LIST, filename=filename)

        raise ValueError(f"Cannot determine what type of image '{filename}' is.")


def find_group_level_ancillary_products(
    files_found: list[Path], is_vlass_seci_ingestion: bool = False
) -> list[AncillaryProduct]:
    """
    Find the group-level ancillary products in the staging dir: there should be:
        * a weblog
        * a pipeline artifacts tar
        * (optional) a selfcal auxproducts tar.
    (The ingestion artifacts tar will be produced during the building of the manifest.)

    :param files_found: List of files to search for ancillary products
    :return:
    """

    try:
        pipeline_artifacts_tar = [
            file
            for file in files_found
            if file.name.startswith(AncillaryProductType.PIPELINE_ARTIFACTS.value) and file.name.endswith(TARFILE_EXT)
        ][0]
    except IndexError as err:
        raise FileNotFoundError(f"WARNING: No pipeline artifacts found in {files_found}") from err

    ancillary_products = [
        AncillaryProduct(type=AncillaryProductType.PIPELINE_WEBLOG, filename=WEBLOG_FILENAME),
        AncillaryProduct(type=AncillaryProductType.PIPELINE_ARTIFACTS, filename=pipeline_artifacts_tar.name),
    ]

    # SECI ingestion has additional group ancillaries to ingest:
    if is_vlass_seci_ingestion is True:
        ancillary_products.append(
            AncillaryProduct(type=AncillaryProductType.REIMAGING_RSC, filename=REIMAGING_FILENAME)
        )
        ancillary_products.append(
            AncillaryProduct(type=AncillaryProductType.PARAMETER_LIST, filename=PARAMETER_FILENAME)
        )

    # if selfcal completed successfully, ingest the resulting auxproducts tar
    selfcal_auxproducts_tar = [file for file in files_found if file.name.endswith(SELFCAL_AUX_FILENAME)]
    if selfcal_auxproducts_tar:
        ancillary_products.append(
            AncillaryProduct(type=AncillaryProductType.AUX_PRODUCTS, filename=selfcal_auxproducts_tar[0].name)
        )

    return ancillary_products
