#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.

""" Helper for building image ingestion manifest """
import logging
import re
from pathlib import Path

from ingest_envoy.manifest_components import (
    TARFILE_EXT,
    WEBLOG_FILENAME,
    AncillaryProduct,
    OutputScienceProduct,
)
from ingest_envoy.utilities import AncillaryProductType, ScienceProductType

FITS = "fits"
RMS = "rms"
SUBIM = "subim"
MEAN_MINMAX = "V_mean_minmaxrej"


class VlassCoarseCubeIngestionProductsFinder:
    """
    Finds the products--science products, product-level ancillary products,
    and group-level ancillary products--needed for VLASS coarse cube ingestion.

    Note: this assumes that there is only one cube being ingested.
    """

    def __init__(self, files_found: list[Path]):
        self.logger = logging.getLogger("ingest_envoy")
        self.files_found = files_found
        self.filename = self.get_filename()
        self.output_science_products = self.find_output_science_products()
        self.ancillary_products = self.find_group_level_ancillary_products()

    def get_filename(self) -> str:
        """
        :return: str. The filename to be put into the manifest.
                For example: VLASS2.1.cc.T10t02.J005000-023000.06.2048.v1.spw2.IQU.iter3.image.pbcor.tt0.subim.fits
                Get the first part (VLASS2.1.cc.T10t02.J005000-023000.06.2048.v1).
        """
        pattern = r"(VLASS.*\.v\d+)\.spw\d+\.IQU\.iter\d+\.image\.pbcor\.tt0\.subim\.fits"
        matches = None
        for file in self.files_found:
            matches = re.search(pattern, file.name)
            if matches:
                break
        return matches.group(1)

    def find_output_science_products(self) -> list[OutputScienceProduct]:
        """
        Find the ancillary products belonging to the science product*
        in the staging dir.

        N.B. ingestion artifacts file will not be there until manifest has been created!

        :return: The output science products as a list
        """
        primary_images = [
            file
            for file in self.files_found
            if file.name.endswith(FITS) and (MEAN_MINMAX in file.name or SUBIM in file.name and RMS not in file.name)
        ]
        self.logger.info("Science Product primaries are: %s", primary_images)

        ancillary_images = [file for file in self.files_found if file.name.endswith(FITS) and RMS in file.name]
        self.logger.info("Science Product ancillaries are: %s", ancillary_images)

        # Build a mapping of science products -> primary + ancillaries
        primary_products = [self.build_ancillary_image_science_product(file) for file in primary_images]
        self.logger.info("Primary cube images: %s", primary_products)

        ancillary_products = [self.build_ancillary_image_science_product(file) for file in ancillary_images]
        self.logger.info("Ancillary cube images: %s", ancillary_products)

        science_products = [
            OutputScienceProduct(
                type=ScienceProductType.VLASS_COARSE,
                filename=self.filename,
                primary_products=primary_products,
                ancillary_products=ancillary_products,
            )
        ]

        return science_products

    @staticmethod
    def build_ancillary_image_science_product(file: Path):
        """
        VLASS coarse cube science products will have ancillary products of their very own,
        distinct from other ancillary products that might be in the staging dir.

        :param file: a possible ancillary image product
        :return: the corresponding AncillaryProduct, if applicable
        """
        filename = file.name

        if filename.endswith(FITS):
            if RMS in filename:
                return AncillaryProduct(type=AncillaryProductType.RMS_IMAGE, filename=filename)
            if MEAN_MINMAX in filename:
                return AncillaryProduct(type=AncillaryProductType.FITS, filename=filename)
            if SUBIM in filename:
                return AncillaryProduct(type=AncillaryProductType.PB_FITS, filename=filename)

        raise ValueError(f"Cannot determine what type of image '{filename}' is.")

    def find_group_level_ancillary_products(self) -> list[AncillaryProduct]:
        """
        Find the group-level ancillary products in the staging directory. There should be:
            * a weblog
            * a pipeline artifacts tar
        (The ingestion artifacts tar will be produced during the building of the manifest.)

        :return:
        """
        parameter_list_regex = ".*parameter.*list"

        try:
            pipeline_artifacts_tar = [
                file
                for file in self.files_found
                if file.name.startswith(AncillaryProductType.PIPELINE_ARTIFACTS.value)
                and file.name.endswith(TARFILE_EXT)
            ][0]
        except IndexError as err:
            raise FileNotFoundError(f"WARNING: No pipeline artifacts found in {self.files_found}") from err

        ancillary_products = [
            AncillaryProduct(type=AncillaryProductType.PIPELINE_WEBLOG, filename=WEBLOG_FILENAME),
            AncillaryProduct(type=AncillaryProductType.PIPELINE_ARTIFACTS, filename=pipeline_artifacts_tar.name),
        ]

        for filename in self.files_found:
            if re.match(parameter_list_regex, filename.name):
                ancillary_products.append(
                    AncillaryProduct(type=AncillaryProductType.PARAMETER_LIST, filename=filename.name)
                )

        return ancillary_products
