#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
import re
import shutil
import tarfile
from pathlib import Path

import pytest
from ingest_envoy.product_collectors import ImageProductCollector
from pycapo import CapoConfig

IMAGE_DIR = "tmpimg"
SPECTRAL_CUBE_DIR = "tmpimgcube"
ARTIFACTS = "pipeline_artifacts_date_time"

SBIN = Path(CapoConfig().getstring("edu.nrao.workspaces.ProcessingSettings.scriptLocation"))
SPOOL = Path(CapoConfig().getstring("edu.nrao.workspaces.ProcessingSettings.rootDirectory"))
STAGING = Path(CapoConfig().getstring("edu.nrao.workspaces.IngestionSettings.stagingDirectory"))
STORAGE = Path(CapoConfig().getstring("edu.nrao.workspaces.IngestionSettings.storageDirectory"))

PRODUCTS = {
    "images": [
        "oussid._0542+498_3C147__bp.K_band.cont.regcal.I.alpha.error.fits",
        "oussid._0542+498_3C147__bp.K_band.cont.regcal.I.alpha.fits",
        "oussid._0542+498_3C147__bp.K_band.cont.regcal.I.pb.tt0.fits",
        "oussid._0542+498_3C147__bp.K_band.cont.regcal.I.tt0.fits",
        "oussid._0542+498_3C147__bp.K_band.cont.regcal.I.tt1.fits",
        "oussid.J0414+3418_ph.K_band.cont.regcal.I.alpha.error.fits",
        "oussid.J0414+3418_ph.K_band.cont.regcal.I.alpha.fits",
        "oussid.J0414+3418_ph.K_band.cont.regcal.I.pb.tt0.fits",
        "oussid.J0414+3418_ph.K_band.cont.regcal.I.tt0.fits",
        "oussid.J0414+3418_ph.K_band.cont.regcal.I.tt1.fits",
        "oussid.J0443+3441_ph.K_band.cont.regcal.I.alpha.error.fits",
        "oussid.J0443+3441_ph.K_band.cont.regcal.I.alpha.fits",
        "oussid.J0443+3441_ph.K_band.cont.regcal.I.pb.tt0.fits",
        "oussid.J0443+3441_ph.K_band.cont.regcal.I.tt0.fits",
        "oussid.J0443+3441_ph.K_band.cont.regcal.I.tt1.fits",
        "oussid.L1473maser_sci.K_band.cont.regcal.I.alpha.error.fits",
        "oussid.L1473maser_sci.K_band.cont.regcal.I.alpha.fits",
        "oussid.L1473maser_sci.K_band.cont.regcal.I.mask.fits",
        "oussid.L1473maser_sci.K_band.cont.regcal.I.pbcor.tt0.fits",
        "oussid.L1473maser_sci.K_band.cont.regcal.I.pb.tt0.fits",
        "oussid.L1473maser_sci.K_band.cont.regcal.I.tt0.fits",
        "oussid.L1473maser_sci.K_band.cont.regcal.I.tt1.fits",
        "oussid.L1482maser_sci.K_band.cont.regcal.I.alpha.error.fits",
        "oussid.L1482maser_sci.K_band.cont.regcal.I.alpha.fits",
        "oussid.L1482maser_sci.K_band.cont.regcal.I.mask.fits",
        "oussid.L1482maser_sci.K_band.cont.regcal.I.pbcor.tt0.fits",
        "oussid.L1482maser_sci.K_band.cont.regcal.I.pb.tt0.fits",
        "oussid.L1482maser_sci.K_band.cont.regcal.I.tt0.fits",
        "oussid.L1482maser_sci.K_band.cont.regcal.I.tt1.fits",
        "oussid.L1483maser_sci.K_band.cont.regcal.I.alpha.error.fits",
        "oussid.L1483maser_sci.K_band.cont.regcal.I.alpha.fits",
        "oussid.L1483maser_sci.K_band.cont.regcal.I.mask.fits",
        "oussid.L1483maser_sci.K_band.cont.regcal.I.pbcor.tt0.fits",
        "oussid.L1483maser_sci.K_band.cont.regcal.I.pb.tt0.fits",
        "oussid.L1483maser_sci.K_band.cont.regcal.I.tt0.fits",
        "oussid.L1483maser_sci.K_band.cont.regcal.I.tt1.fits",
        "oussid.LkHA101maser_sci.K_band.cont.regcal.I.alpha.error.fits",
        "oussid.LkHA101maser_sci.K_band.cont.regcal.I.alpha.fits",
        "oussid.LkHA101maser_sci.K_band.cont.regcal.I.mask.fits",
        "oussid.LkHA101maser_sci.K_band.cont.regcal.I.pbcor.tt0.fits",
        "oussid.LkHA101maser_sci.K_band.cont.regcal.I.pb.tt0.fits",
        "oussid.LkHA101maser_sci.K_band.cont.regcal.I.tt0.fits",
        "oussid.LkHA101maser_sci.K_band.cont.regcal.I.tt1.fits",
    ],
    "cubes": [
        "oussid.L1473maser_sci.spw11.cube.regcal.I.mask.fits",
        "oussid.L1473maser_sci.spw11.cube.regcal.I.pbcor.fits",
        "oussid.L1473maser_sci.spw11.cube.regcal.I.pb.fits",
        "oussid.L1473maser_sci.spw12.cube.regcal.I.mask.fits",
        "oussid.L1473maser_sci.spw12.cube.regcal.I.pbcor.fits",
        "oussid.L1473maser_sci.spw12.cube.regcal.I.pb.fits",
        "oussid.L1473maser_sci.spw13.cube.regcal.I.mask.fits",
        "oussid.L1473maser_sci.spw13.cube.regcal.I.pbcor.fits",
        "oussid.L1473maser_sci.spw13.cube.regcal.I.pb.fits",
        "oussid.L1473maser_sci.spw15.cube.regcal.I.mask.fits",
        "oussid.L1473maser_sci.spw15.cube.regcal.I.pbcor.fits",
        "oussid.L1473maser_sci.spw15.cube.regcal.I.pb.fits",
        "oussid.L1473maser_sci.spw18.cube.regcal.I.mask.fits",
        "oussid.L1473maser_sci.spw18.cube.regcal.I.pbcor.fits",
        "oussid.L1473maser_sci.spw18.cube.regcal.I.pb.fits",
        "oussid.L1473maser_sci.spw5.cube.regcal.I.mask.fits",
        "oussid.L1473maser_sci.spw5.cube.regcal.I.pbcor.fits",
        "oussid.L1473maser_sci.spw5.cube.regcal.I.pb.fits",
        "oussid.L1482maser_sci.spw11.cube.regcal.I.mask.fits",
        "oussid.L1482maser_sci.spw11.cube.regcal.I.pbcor.fits",
        "oussid.L1482maser_sci.spw11.cube.regcal.I.pb.fits",
        "oussid.L1482maser_sci.spw12.cube.regcal.I.mask.fits",
        "oussid.L1482maser_sci.spw12.cube.regcal.I.pbcor.fits",
        "oussid.L1482maser_sci.spw12.cube.regcal.I.pb.fits",
        "oussid.L1482maser_sci.spw13.cube.regcal.I.mask.fits",
        "oussid.L1482maser_sci.spw13.cube.regcal.I.pbcor.fits",
        "oussid.L1482maser_sci.spw13.cube.regcal.I.pb.fits",
        "oussid.L1482maser_sci.spw15.cube.regcal.I.mask.fits",
        "oussid.L1482maser_sci.spw15.cube.regcal.I.pbcor.fits",
        "oussid.L1482maser_sci.spw15.cube.regcal.I.pb.fits",
        "oussid.L1482maser_sci.spw18.cube.regcal.I.mask.fits",
        "oussid.L1482maser_sci.spw18.cube.regcal.I.pbcor.fits",
        "oussid.L1482maser_sci.spw18.cube.regcal.I.pb.fits",
        "oussid.L1482maser_sci.spw5.cube.regcal.I.mask.fits",
        "oussid.L1482maser_sci.spw5.cube.regcal.I.pbcor.fits",
        "oussid.L1482maser_sci.spw5.cube.regcal.I.pb.fits",
        "oussid.L1483maser_sci.spw11.cube.regcal.I.mask.fits",
        "oussid.L1483maser_sci.spw11.cube.regcal.I.pbcor.fits",
        "oussid.L1483maser_sci.spw11.cube.regcal.I.pb.fits",
        "oussid.L1483maser_sci.spw12.cube.regcal.I.mask.fits",
        "oussid.L1483maser_sci.spw12.cube.regcal.I.pbcor.fits",
        "oussid.L1483maser_sci.spw12.cube.regcal.I.pb.fits",
        "oussid.L1483maser_sci.spw13.cube.regcal.I.mask.fits",
        "oussid.L1483maser_sci.spw13.cube.regcal.I.pbcor.fits",
        "oussid.L1483maser_sci.spw13.cube.regcal.I.pb.fits",
        "oussid.L1483maser_sci.spw15.cube.regcal.I.mask.fits",
        "oussid.L1483maser_sci.spw15.cube.regcal.I.pbcor.fits",
        "oussid.L1483maser_sci.spw15.cube.regcal.I.pb.fits",
        "oussid.L1483maser_sci.spw18.cube.regcal.I.mask.fits",
        "oussid.L1483maser_sci.spw18.cube.regcal.I.pbcor.fits",
        "oussid.L1483maser_sci.spw18.cube.regcal.I.pb.fits",
        "oussid.L1483maser_sci.spw5.cube.regcal.I.mask.fits",
        "oussid.L1483maser_sci.spw5.cube.regcal.I.pbcor.fits",
        "oussid.L1483maser_sci.spw5.cube.regcal.I.pb.fits",
        "oussid.LkHA101maser_sci.spw11.cube.regcal.I.mask.fits",
        "oussid.LkHA101maser_sci.spw11.cube.regcal.I.pbcor.fits",
        "oussid.LkHA101maser_sci.spw11.cube.regcal.I.pb.fits",
        "oussid.LkHA101maser_sci.spw12.cube.regcal.I.mask.fits",
        "oussid.LkHA101maser_sci.spw12.cube.regcal.I.pbcor.fits",
        "oussid.LkHA101maser_sci.spw12.cube.regcal.I.pb.fits",
        "oussid.LkHA101maser_sci.spw13.cube.regcal.I.mask.fits",
        "oussid.LkHA101maser_sci.spw13.cube.regcal.I.pbcor.fits",
        "oussid.LkHA101maser_sci.spw13.cube.regcal.I.pb.fits",
        "oussid.LkHA101maser_sci.spw15.cube.regcal.I.mask.fits",
        "oussid.LkHA101maser_sci.spw15.cube.regcal.I.pbcor.fits",
        "oussid.LkHA101maser_sci.spw15.cube.regcal.I.pb.fits",
        "oussid.LkHA101maser_sci.spw18.cube.regcal.I.mask.fits",
        "oussid.LkHA101maser_sci.spw18.cube.regcal.I.pbcor.fits",
        "oussid.LkHA101maser_sci.spw18.cube.regcal.I.pb.fits",
        "oussid.LkHA101maser_sci.spw5.cube.regcal.I.mask.fits",
        "oussid.LkHA101maser_sci.spw5.cube.regcal.I.pbcor.fits",
        "oussid.LkHA101maser_sci.spw5.cube.regcal.I.pb.fits",
    ],
    "other": [
        "17A-438.sb33693673.eb33715233.57879.73143714121.ms.calapply.txt",
        "17A-438.sb33693673.eb33715233.57879.73143714121.ms.flagversions.tgz",
        "casa_commands.log",
        "casa_piperestorescript.py",
        "casa_pipescript.py",
        "unknown.default.caltables.tgz",
        "unknown.hifv_calimage_cont_cube_selfcal.auxproducts.tgz",
        "unknown.hifv_calimage_cont_cube_selfcal.pipeline_manifest.xml",
        "weblog.tgz",
    ],
}


def stub_area(workflow_dir: str, products: list) -> Path:
    """
    Fixture that creates a temporary mock image products directory.

    :return: Path to the base of the products directory
    """
    # Make spool, staging, and storage directories
    for d in [SPOOL / workflow_dir, STAGING / workflow_dir, STORAGE / workflow_dir]:
        d.mkdir(parents=True, exist_ok=True)

    # Populate spool dir
    workflow_path = SPOOL / workflow_dir
    spool_dirs = ["products", "rawdata", "working"]
    spool_files = ["metadata.json", "PPR.xml"]
    for d in spool_dirs:
        (workflow_path / d).mkdir(parents=True)
    for f in spool_files:
        (workflow_path / f).touch()

    # Populate spool/products
    products_path = workflow_path / "products"
    for f in products:
        (products_path / f).touch()

    yield workflow_path

    shutil.rmtree(SPOOL / workflow_dir)
    shutil.rmtree(STAGING / workflow_dir)
    shutil.rmtree(STORAGE / workflow_dir)

    assert not (SPOOL / workflow_dir).exists()
    assert not (STAGING / workflow_dir).exists()
    assert not (STORAGE / workflow_dir).exists()


@pytest.fixture(scope="function")
def stub_image_area() -> Path:
    """
    Fixture that creates a temporary mimic of an image products directory.

    :return: Path to the base of the products directory
    """
    yield from stub_area(IMAGE_DIR, PRODUCTS["images"] + PRODUCTS["other"])


@pytest.fixture(scope="function")
def stub_spectral_cube_area() -> Path:
    """
    Fixture that creates a temporary mimic of an image products directory with cube images.

    :return: Path to the base of the products directory
    """
    yield from stub_area(SPECTRAL_CUBE_DIR, PRODUCTS["images"] + PRODUCTS["cubes"] + PRODUCTS["other"])


def test_image_product_collector_weblog(stub_image_area: Path):
    """
    Test that the image collector places the already-tarred weblog in staging.
    """
    collector = ImageProductCollector(IMAGE_DIR, str(STAGING), ARTIFACTS)
    collector()
    assert (STAGING / IMAGE_DIR / "weblog.tgz").exists()


def test_image_product_collector_weblog_starts_untarred(stub_image_area: Path):
    """
    Test that the image collector tars the weblog and places it in staging.
    """
    (stub_image_area / "products" / "weblog.tgz").unlink()
    (stub_image_area / "products" / "pipeline-20241108T165151" / "html").mkdir(parents=True, exist_ok=True)
    (stub_image_area / "products" / "pipeline-20241108T165151" / "html" / "index.html").write_text("I am a weblog")

    collector = ImageProductCollector(IMAGE_DIR, str(STAGING), ARTIFACTS)
    collector()
    assert (STAGING / IMAGE_DIR / "weblog.tgz").exists()
    # Check that the weblog isn't nested in a directory tree
    for name in tarfile.open(STAGING / IMAGE_DIR / "weblog.tgz").getnames():
        assert re.match("^pipeline-20241108T165151.*$", name)


def test_image_product_collector_selfcal(stub_image_area: Path):
    """
    Test that the image collector collects the selfcal auxproducts and places it in staging.
    """
    collector = ImageProductCollector(IMAGE_DIR, str(STAGING), ARTIFACTS)
    collector()
    assert (STAGING / IMAGE_DIR / "unknown.hifv_calimage_cont_cube_selfcal.auxproducts.tgz").exists()


def test_image_product_collector_artifacts(stub_image_area: Path):
    """
    Test that the image collector tars the artifacts and places it in staging.
    """
    collector = ImageProductCollector(IMAGE_DIR, str(STAGING), ARTIFACTS)
    collector()
    assert (STAGING / IMAGE_DIR / "pipeline_artifacts_date_time.tar").exists()
    # Check that an artifact isn't nested in a directory tree
    assert any(
        name == "casa_commands.log"
        for name in tarfile.open(STAGING / IMAGE_DIR / "pipeline_artifacts_date_time.tar").getnames()
    )


def test_image_product_collector_image_fits(stub_image_area: Path):
    """
    Test that the image collector collects the fits files and places them in staging.
    """
    collector = ImageProductCollector(IMAGE_DIR, str(STAGING), ARTIFACTS)
    collector()
    for f in PRODUCTS["images"]:
        assert (STAGING / IMAGE_DIR / f).exists()


def test_image_product_collector_cube_fits(stub_spectral_cube_area: Path):
    """
    Test that the image collector collects the cube fits files and places them in staging.
    """
    collector = ImageProductCollector(SPECTRAL_CUBE_DIR, str(STAGING), ARTIFACTS)
    collector()
    for f in PRODUCTS["cubes"]:
        assert (STAGING / SPECTRAL_CUBE_DIR / f).exists()
