#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
"""
Tests for ingest_envoy.launchers
"""
from unittest.mock import patch

import pytest
from ingest_envoy import launchers
from ingest_envoy.launchers import (
    IngestCalibrationLauncher,
    IngestImageLauncher,
    IngestVLASSLauncher,
)
from ingest_envoy.schema import AbstractTextFile
from ingest_envoy.utilities import IngestType, VLASSIngestType

parameters = {
    "sdmId": "16B-069_sb32814386_1_001.57685.66193635417",
    "telescope": "EVLA",
    "workflowName": "std_calibration",
    "spl": "uid://evla/execblock/48ba4c9d-d7c7-4a8f-9803-1115cd52459b",
    "processingStart": "2021_07_06T21_50_48",
    "destinationDir": "/lustre/aoc/cluster/pipeline/docker/workspaces/spool/tmp576dknjp",
    "workflowDir": "tmp576dknjp",
    "staging_area": "/lustre/aoc/cluster/pipeline/docker/workspaces/staging",
    "storage_area": "/lustre/aoc/cluster/pipeline/docker/workspaces/storage",
    "script_location": "/lustre/aoc/cluster/pipeline/docker/workspaces/sbin",
    "useIngest": False,
    "multiVersion": False,
}

image_parameters = {
    "sdmId": "brain_000.58099.67095825232",
    "telescope": "EVLA",
    "workflowName": "std_cms_imaging",
    "spl": "uid://evla/execblock/ec082e65-452d-4fec-ad88-f5b4af1f9e36",
    "processingStart": "2021_07_06T21_50_48",
    "destinationDir": "/lustre/aoc/cluster/pipeline/docker/workspaces/spool/tmprb1se376",
    "workflowDir": "tmprb1se376",
    "calSpl": "uid://evla/calibration/c78ccfd6-fe4e-43c6-a5c5-70e5bcfde16b",
    "staging_area": "/lustre/aoc/cluster/pipeline/docker/workspaces/staging",
    "storage_area": "/lustre/aoc/cluster/pipeline/docker/workspaces/storage",
    "script_location": "/lustre/aoc/cluster/pipeline/docker/workspaces/sbin",
    "useIngest": False,
}

coarse_parameters = {
    "telescope": "EVLA",
    "workflowName": "vlass_coarse",
    "project": "TEST000",
    "multiVersion": False,
    "calSpl": "uid://evla/calibration/ba6c6404-b196-4d5d-9ab8-8e2b5bce62c1",
    "workflowDir": "VLASS3.1_T10t10.J063800-023000_P106057v1_2023_03_07T22_58_33.300",
    "sourcePath": "/lustre/aoc/cluster/pipeline/vlass_dev/spool/quicklook/",
    "qa_status": "PASS",
    "staging_area": "/lustre/aoc/cluster/pipeline/docker/workspaces/staging",
    "storage_area": "/lustre/aoc/cluster/pipeline/docker/workspaces/storage",
    "script_location": "/lustre/aoc/cluster/pipeline/docker/workspaces/sbin",
}


@patch("subprocess.run")
def test_trigger_ingest(mock_run):
    launchers.trigger_ingest(parameters["useIngest"], parameters["destinationDir"], parameters["script_location"])
    assert mock_run.call_count == 0
    launchers.trigger_ingest(True, parameters["destinationDir"], parameters["script_location"])
    assert mock_run.call_count == 1


class TestIngestCalibrationLauncher:
    def test_launch_ingestion(self):
        with patch("ingest_envoy.launchers.IngestCalibrationLauncher.prepare_for_launch") as prepare:
            with patch("ingest_envoy.launchers.trigger_ingest") as ingest:
                IngestCalibrationLauncher(IngestType.CAL, parameters).launch()
                assert prepare.call_count == 1
                assert ingest.call_count == 1

    @patch("subprocess.run", return_value=0)
    def test_prepare_for_ingest(self, mock_run):
        with patch("ingest_envoy.launchers.IngestCalibrationLauncher.run_collection_script") as collector:
            with patch("ingest_envoy.launchers.IngestCalibrationLauncher.create_manifest") as manifest:
                IngestCalibrationLauncher(IngestType.CAL, parameters).prepare_for_launch()
                assert collector.call_count == 1
                assert manifest.call_count == 1

    @patch("subprocess.run")
    def test_run_collection_script(self, mock_run):
        mock_run.return_value.returncode = 0
        IngestCalibrationLauncher(IngestType.CAL, parameters).run_collection_script()
        assert mock_run.call_count == 1

    @pytest.mark.skip("Skip. Ignores manifest builder mock")
    def test_create_manifest(self):
        with patch("ingest_envoy.ingestion_manifest.IngestionManifestBuilder.build") as manifest:
            IngestCalibrationLauncher(IngestType.CAL, parameters).create_manifest()
            assert manifest.call_count == 1


class TestIngestImageLauncher:
    def test_launch_ingestion(self):
        with patch("ingest_envoy.launchers.IngestImageLauncher.prepare_for_launch") as prepare:
            with patch("ingest_envoy.launchers.trigger_ingest") as ingest:
                IngestImageLauncher(IngestType.IMG, image_parameters).launch()
                assert prepare.call_count == 1
                assert ingest.call_count == 1

    def test_prepare_for_ingest(self):
        with patch("ingest_envoy.launchers.IngestImageLauncher.run_collector") as collector:
            with patch("ingest_envoy.launchers.IngestImageLauncher.create_manifest") as manifest:
                IngestImageLauncher(IngestType.IMG, image_parameters).prepare_for_launch()
                assert collector.call_count == 1
                assert manifest.call_count == 1

    @patch("builtins.open")
    @patch("subprocess.run")
    def test_run_collector(self, mock_run, mock_open):
        with patch("ingest_envoy.collectors.ImageCollector.collect_products") as products:
            mock_run.return_value.returncode = 0
            aux_file = IngestImageLauncher(IngestType.IMG, image_parameters).run_collector()
            assert products.call_count == 1
            assert isinstance(aux_file, AbstractTextFile)

    @patch("builtins.open")
    @patch("subprocess.run")
    def test_run_coarse_collector(self, mock_run, mock_open):
        with patch("ingest_envoy.collectors.CoarseCollector.collect_products") as products:
            mock_run.return_value.returncode = 0
            aux_file = IngestVLASSLauncher(VLASSIngestType.COARSE, coarse_parameters).run_collector()
            assert products.call_count == 1
            assert isinstance(aux_file, AbstractTextFile)

    @pytest.mark.skip("Skip. Ignores manifest builder mock")
    def test_create_manifest(self):
        with patch("ingest_envoy.ingestion_manifest.IngestionManifestBuilder.build") as manifest:
            IngestImageLauncher(IngestType.IMG, image_parameters).create_manifest()
            assert manifest.call_count == 1
