# Copyright (C) 2023 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
#
import json
import tarfile
from pathlib import Path
from typing import Callable

from conftest import (
    AUX_IMAGE_METADATA_FILENAME,
    SPECTRAL_CUBE_ANCILLARIES,
    SPECTRAL_CUBE_CALIBRATION_SPL,
    SPECTRAL_CUBE_SCIENCE_PRODUCTS,
    find_example_manifest,
    find_ingestion_artifacts_tar,
)
from ingest_envoy.ingestion_manifest import (
    IngestionManifestBuilder,
    IngestionManifestSchema,
)
from ingest_envoy.manifest_components import (
    MANIFEST_FILENAME,
    AncillaryProduct,
    OutputScienceProduct,
)
from ingest_envoy.schema import AbstractTextFile
from ingest_envoy.utilities import AncillaryProductType, ScienceProductType, Telescope


def ordered(obj: dict | list) -> dict | list:
    """
    Ripped this off SO, recursively sorts the given object and all nested lists.

    :param obj: An object to sort
    :return: The given object sorted all the way down
    """
    if isinstance(obj, dict):
        return sorted((k, ordered(v)) for k, v in obj.items())
    if isinstance(obj, list):
        return sorted(ordered(x) for x in obj)
    else:
        return obj


def test_spectral_cube_manifest_creation(spectral_cube_path: Path):
    """Test manifest creation for a spectral cube"""
    actual_manifest, actual_manifest_file = IngestionManifestBuilder(
        manifest_source_dir=spectral_cube_path,
        additional_metadata=AbstractTextFile(filename=AUX_IMAGE_METADATA_FILENAME, content=""),
        telescope=Telescope.EVLA.value,
        sp_type=ScienceProductType.IMAGE.value,
        locator=SPECTRAL_CUBE_CALIBRATION_SPL,
    ).build()

    params = actual_manifest.parameters
    assert params.ngas_ingest is False
    assert params.telescope == Telescope.EVLA
    assert params.staging_source_dir == spectral_cube_path
    assert actual_manifest.reingest is None

    # Check Ancillaries more in the JSON-checking section since the ingest artifacts' filename depends on the current time
    output_group = actual_manifest.output_group
    assert output_group
    aps = output_group.ancillary_products
    assert aps
    assert len(aps) == len(SPECTRAL_CUBE_ANCILLARIES)

    # Check science products here in great detail because we can
    get_filename: Callable[[OutputScienceProduct | AncillaryProduct], str] = lambda p: p.filename
    actual_sps = output_group.science_products
    assert actual_sps
    assert len(actual_sps) == len(SPECTRAL_CUBE_SCIENCE_PRODUCTS)
    for actual_sp, expected_sp in zip(
        sorted(actual_sps, key=get_filename), sorted(SPECTRAL_CUBE_SCIENCE_PRODUCTS, key=get_filename)
    ):
        # Check the type, name, and ancillaries
        assert actual_sp.ancillary_products and expected_sp.ancillary_products
        actual_sp.ancillary_products.sort(key=get_filename)
        expected_sp.ancillary_products.sort(key=get_filename)
        assert actual_sp == expected_sp

    # Check actual file against the schema
    with open(actual_manifest_file) as f:
        actual_manifest_json = json.load(f)
    actual_manifest_schema = IngestionManifestSchema()
    # These are sorted because ancillary products are ordered differently between the two dumps for some asinine reason
    assert ordered(actual_manifest_schema.dump(actual_manifest)) == ordered(actual_manifest_json)

    # Check actual file vs. expected IngestionManifest object
    expected_manifest_file = find_example_manifest("spectral_cube_manifest")
    with open(expected_manifest_file) as f:
        expected_manifest = json.load(f)

    # ingestion_path depends on the ingest_path fixture, so ignore it
    expected_params = expected_manifest["parameters"]
    actual_params = actual_manifest_json["parameters"]
    assert actual_params["ngas_ingest"] == expected_params["ngas_ingest"]
    assert "reingest" not in actual_manifest_json and "reingest" not in expected_manifest
    assert actual_manifest_json.get("input_group") == expected_manifest.get("input_group")

    # Don't need to check ScienceProducts here in much detail since we did that with the Python representation
    expected_outgroup = expected_manifest["output_group"]
    expected_osp = expected_outgroup["science_products"]
    actual_outgroup = actual_manifest_json["output_group"]
    actual_osp = actual_outgroup["science_products"]
    assert len(actual_osp) == len(expected_osp)

    # Check Ancillaries here in great detail
    expected_aps = expected_outgroup["ancillary_products"]
    actual_aps = actual_outgroup["ancillary_products"]
    assert len(expected_aps) == len(actual_aps)
    for expected_ap, actual_ap in zip(
        sorted(expected_aps, key=lambda ap: ap["type"]), sorted(actual_aps, key=lambda ap: ap["type"])
    ):
        assert expected_ap["type"] == actual_ap["type"]
        # the ingestion_artifacts' filename depends on current time, so ignore it
        if expected_ap["type"] != AncillaryProductType.INGESTION_ARTIFACTS.value:
            assert expected_ap == actual_ap

    # Check the ingestion_artifacts tar
    #     It should contain ONLY the ingestion manifest and the additional metadata file
    addl_md = actual_manifest.parameters.additional_metadata
    assert addl_md
    ingestion_artifacts_tar = find_ingestion_artifacts_tar(spectral_cube_path)
    with tarfile.open(ingestion_artifacts_tar, "r") as tar:
        members = tar.getmembers()
        assert len(members) == 2
        for member in members:
            assert member.name.endswith(MANIFEST_FILENAME) or member.name.endswith(addl_md.filename)
