#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
"""
Contains the tests for the VLASS Coarse Cube product collector.
"""
import os
import shutil
import tarfile
import typing
from pathlib import Path

import pytest
from conftest import get_vlass_coarse_cube_cache_products_dict
from ingest_envoy.product_collectors import VlassCoarseProductCollector
from pycapo import CapoConfig

CAPO_CONFIG = CapoConfig()

CACHE = Path(CAPO_CONFIG.getstring("edu.nrao.workspaces.VlassDeliverySettings.cache"))
CACHE_DIR = CACHE / "tmp_coarse_cube_cache"


@pytest.fixture(name="cache_dir")
def make_cache_dir(
    vlass_coarse_cache_products: list[str], vlass_coarse_cube_staging_dir: Path
) -> typing.Generator[Path, None, None]:
    """
    Fixture that creates a temporary mock cube directory.

    :param vlass_coarse_cache_products: list. A list of the products that should be in the temporary cache directory.
    :param vlass_coarse_cube_staging_dir: Path. The path to the temporary VLASS coarse cube staging directory.
    :return: typing.Generator[Path]. A Path to the base of the cube directory.
    """
    # Make cache directory
    CACHE_DIR.mkdir(parents=True)

    # Populate cache directory
    for f in vlass_coarse_cache_products:
        (CACHE_DIR / f).touch()

    yield CACHE_DIR

    shutil.rmtree(CACHE_DIR)
    shutil.rmtree(vlass_coarse_cube_staging_dir)
    assert not CACHE_DIR.exists()
    assert not vlass_coarse_cube_staging_dir.exists()


@pytest.fixture(name="collector")
def create_collector(cache_dir: Path, vlass_coarse_cube_staging_dir: Path) -> VlassCoarseProductCollector:
    """
    Creates and a VLASS coarse cube product collector.
    :param cache_dir: Path. Allows us to get the cache dir as a path.
    :param vlass_coarse_cube_staging_dir: Path. The path to the temporary VLASS coarse cube staging directory.
    :return VlassCoarseProductCollector. Instantiated using the cache dir and a given pipeline artifacts tar name.
    """
    return VlassCoarseProductCollector(cache_dir, str(vlass_coarse_cube_staging_dir), "pipeline_artifacts")


def test_vlass_coarse_product_collector_prepare(collector, vlass_coarse_cube_staging_dir: Path):
    """
    Calls and tests the result of the VlassCoarseProductCollector.prepare method.
    :param collector: VlassCoarseProductCollector. A Pytest fixture.
    :param vlass_coarse_cube_staging_dir: Path. The path to the temporary VLASS coarse cube staging directory.
    """
    collector.prepare()
    assert vlass_coarse_cube_staging_dir.exists()
    staging_dir_permissions = str(oct(VlassCoarseProductCollector.STAGING_PERMISSIONS))
    assert oct(os.stat(vlass_coarse_cube_staging_dir).st_mode & 0o777) == staging_dir_permissions


def test_vlass_coarse_product_collector(
    collector, vlass_coarse_staging_products: list[str], vlass_coarse_cube_staging_dir: Path
):
    """
    Calls and tests the result of the VlassCoarseProductCollector.collect method.
    :param collector: VlassCoarseProductCollector. A Pytest fixture.
    :param vlass_coarse_staging_products: list[str]. The list of products that should be in the staging directory.
    :param vlass_coarse_cube_staging_dir: Path. The path to the temporary VLASS coarse cube staging directory.
    """
    # Arrange
    collector.prepare()

    # Act
    collector.collect()

    # Assert
    # Check to see that all the files that
    # should be in the staging directory are there
    for file in vlass_coarse_staging_products:
        assert (vlass_coarse_cube_staging_dir / file).exists()

    # Check if .subim.com.fits files were not copied to staging
    for file in get_vlass_coarse_cube_cache_products_dict()["other_products"]:
        assert not (vlass_coarse_cube_staging_dir / file).exists()

    # Check to see that only what we expect to be in the staging directory is there
    assert sorted(os.listdir(vlass_coarse_cube_staging_dir)) == sorted(vlass_coarse_staging_products)


def test_coarse_cube_cache_artifacts(collector, vlass_coarse_cube_staging_dir: Path):
    """
    Tests the collector tarred artifacts properly.

    :param collector: A Pytest fixture.
    :param vlass_coarse_cube_staging_dir: Path. The path to the temporary VLASS coarse cube staging directory.
    """
    collector.prepare()
    collector.collect()

    assert (vlass_coarse_cube_staging_dir / "pipeline_artifacts.tar").exists()
    # Check that an artifact isn't nested in a directory tree
    assert any(
        name == "casa_commands.log"
        for name in tarfile.open(vlass_coarse_cube_staging_dir / "pipeline_artifacts.tar").getnames()
    )
