#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
"""
Misc. REST Calls and HTCondor calls
"""
import json
import logging
import os
import subprocess
from typing import Optional

import requests

logger = logging.getLogger("mediator")

# Workflow Service REST Calls


def get_request(url: str, wf_id: int) -> json:
    """
    Lookup relevant workflow requests

    :param url: system workflow url
    :param wf_id: relevant workflow request id
    :return:
    """
    response = requests.get(f"{url}/workflows/na/requests/{wf_id}")
    return response.json()


def get_capability_request(url: str, cr_id: int) -> json:
    """
    Lookup relevant workflow requests

    :param url: system workflow url
    :param cr_id: relevant capability request id
    :return:
    """
    response = requests.get(f"{url}/capability/request/{cr_id}")
    return response.json()


def get_condor_id(url: str, wf_id: int) -> int:
    response = requests.get(f"{url}/workflows/na/requests/{wf_id}/htcondor_id")
    return response.json()["htcondor_job_id"]


# HTCondor action calls


# shamelessly stolen from Nathan's wf_inspector code
def get_container_id_by_substr(substr: str) -> Optional[bytes]:
    """
    Get ID of running container searched on a substring of its name

    :param substr: Substring to search on
    :return: Container ID, if found; else None
    """
    result = subprocess.run(["docker", "ps", "-q", "-f", f"name={substr}"], stdout=subprocess.PIPE)
    if result.returncode == 0:
        if len(result.stdout) > 0:
            # Container was found matching substr
            return result.stdout.strip()
        else:
            return None
    else:
        logger.error("Docker command not accessible. Docker is either not installed or you lack permissions to use it.")
        logger.info("Are you running this from within a Docker container? Docker is not installed in containers.")
        return None


def is_docker():
    host = os.environ.get("HOSTNAME")

    if any(char.isdigit() for char in host):
        return True

    return False
