#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
"""
Functions for performing intercessions
Function name must match intervention type
"""
import logging
import subprocess

import requests

from . import calls

logger = logging.getLogger("mediator")


def force_fail(url: str, wf_ids: list[int]):
    """
    Force a specified workflow request to fail

    :return:
    """
    for wf in wf_ids:
        # is HTCondor running this request?
        condor_id = calls.get_condor_id(url, wf)
        condor_id = None if condor_id == "None" else condor_id

        if condor_id is not None:
            # evict job from condor, which will also fail the request
            logger.info(f"Evicting request #{wf} from HTCondor...")
            if calls.is_docker():
                response = subprocess.run(["condor_rm", f"{condor_id}"], capture_output=True)
            else:
                workflow_container = calls.get_container_id_by_substr("workflow")
                response = subprocess.run(
                    ["docker", "exec", "-it", workflow_container, "condor_rm", f"{condor_id}"], capture_output=True
                )

            outerr = response.stderr.decode()
            logger.info(f"{outerr}")
            if "condor_rm:0:There are no jobs in the queue" in outerr:
                # send manual fail
                logger.info(f"No condor job was found for request #{wf}. Failing via REST call...")
                workflow_url = f"{url}/workflows/na/requests/{wf}/fail"
                logger.info(f"Sending 'force fail' REST call to workflow service for request #{wf}...")
                requests.post(workflow_url)
        else:
            # manually send fail message
            workflow_url = f"{url}/workflows/na/requests/{wf}/fail"
            logger.info(f"Sending 'force fail' REST call to workflow service for request #{wf}...")
            requests.post(workflow_url)


def submit(url: str, wf_ids: list[int]):
    """
    Manually submit an already existing workflow

    :return:
    """
    for wf in wf_ids:
        wf_name = calls.get_request(url, wf)["workflow_name"]
        workflow_url = f"{url}/workflows/{wf_name}/requests/{wf}/submit"
        logger.info(f"Sending 'submit' REST call to workflow service for request #{wf}...")
        requests.post(workflow_url)


def cancel(url, cr_ids: list[str]):
    """
    Manually cancel one or more Capability Requests

    :param url:
    :param cr_ids:
    :return:
    """

    logger.info(f"Cancelling the following Capability Requests: {str(cr_ids)}")

    for cr in cr_ids:
        latest_version = calls.get_capability_request(url, int(cr))["versions"][-1]
        version_num = latest_version["version_number"]
        cancel_url = f"{url}/capability/request/{cr}/version/{version_num}/cancel"
        logger.info(f"Cancelling Capability Request #{cr}, Version #{version_num}...")
        requests.post(cancel_url)
