#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
"""
Functions for resolving arbitration
Function name must match intervention type
"""
import logging
import time

from . import calls

logger = logging.getLogger("mediator")


def log_confirmations_to_user(resolution_type: str, failed: list[int], success: list[int], user_list_size: int):
    logger.info(f"Resolution Summary for {resolution_type}:")
    if user_list_size > len(failed) > 0 and user_list_size > len(success) > 0:
        logger.info(
            f"Success ({len(success)} of {user_list_size}): {str(success)}\n"
            f"Unable to confirm ({len(failed)} of {user_list_size}): {str(failed)} "
            f"\nPlease check the database."
        )
    elif len(failed) == user_list_size:
        logger.info(
            f"Unable to confirm ({len(failed)} of {user_list_size}): {str(failed)}. \nPlease check the database."
        )
    elif len(success) == user_list_size:
        logger.info(f"Success ({len(success)} of {user_list_size}): Resolution is complete!")
    else:
        logger.info(f"Unknown resolution state for all requests. \nPlease check the database.")


def force_fail(url: str, wf_ids: list[int]):
    """
    Report whether force fail intervention succeeded or failed

    :return:
    """
    success = []
    failed = []

    time.sleep(1)
    for wf in wf_ids:
        check_status = calls.get_request(url, wf)["state"]
        if check_status in ["Failed", "Error"]:
            success.append(wf)
        else:
            failed.append(wf)

    log_confirmations_to_user("forced fail", failed, success, len(wf_ids))


def submit(url: str, wf_ids: list[int]):
    """
    Report whether submit intervention succeeded or failed

    :return:
    """
    success = []
    failed = []

    time.sleep(1)
    for wf in wf_ids:
        check_status = calls.get_request(url, wf)["state"]
        if check_status == "Created":
            failed.append(wf)
        else:
            success.append(wf)

    log_confirmations_to_user("submission", failed, success, len(wf_ids))


def cancel(url: str, cr_ids: list[int]):
    """
    Report whether cancel intervention succeeded or failed

    :param url: The base url
    :param cr_ids: the capability requests needing intervention
    :return:
    """
    success = []
    failed = []

    time.sleep(1)
    for cr in cr_ids:
        check_status = calls.get_capability_request(url, cr)["state"]
        if check_status != "Cancelled":
            failed.append(cr)
        else:
            success.append(cr)

    log_confirmations_to_user("cancellation", failed, success, len(cr_ids))
