#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
""" Tests for the product fetcher """

# pylint: disable=C0103, E0401, E0402, E1136, R0201, R0903

from pathlib import Path
from typing import List, Union
from unittest.mock import patch

import pytest
from productfetcher.fetchers import DryRunFakeFileFetcher
from productfetcher.interfaces import (
    FetcherFactory,
    FetchPlan,
    FileFetcher,
    ForceMode,
    LocatedFile,
)
from productfetcher.locations import FileLocator, ServiceLocator
from productfetcher.product_fetcher import (
    CLIParam,
    CommandLineFetchContext,
    FetchAncillaryOnly,
    FetchBothScienceAndAncillary,
    FetchContext,
    FetchScienceAndImageAncillaries,
    FetchScienceOnly,
    SdmOnly,
)


class FakePlan(FetchPlan):
    """A non-fetching fetch plan"""

    def __init__(self, fetchers, concurrency):
        self.fetchers = fetchers
        self.concurrency = concurrency

    def fetch(self, context: FetchContext):
        # Let's not actually fetch anything today
        pass


class FakeFactory(FetcherFactory):
    """Creates non-fetching fetchers"""

    def fetch_ngas(self, file: LocatedFile) -> FileFetcher:
        return DryRunFakeFileFetcher(file)

    def fetch_oracle_xml(self, file: LocatedFile) -> FileFetcher:
        return DryRunFakeFileFetcher(file)

    def fetch_plan(self, fetchers: List[Union[FileFetcher, FetchPlan]], concurrency: int = 1) -> FetchPlan:
        return FakePlan(fetchers, concurrency)


def test_plan_generation_science_only(resource_path_root: Path):
    """
    Can we generate a test plan?

    :param resource_path_root:
    :return:
    """
    report = FileLocator(resource_path_root / "location_files" / "13B-014.json").locate()
    fake_factory = FakeFactory()
    fetchers = report.fetchers(FetchScienceOnly(), fake_factory)

    # we should absolutely have the same number of fetchers as we have files
    assert len(fetchers) == len(report.files)

    # let's make a plan
    plan = CommandLineFetchContext.calculate_plan(fake_factory, fetchers)

    # we should have the same number of fetchers in the plan as in the original report
    assert all(isinstance(subplan, FetchPlan) for subplan in plan.fetchers)
    assert sum(len(subplan.fetchers) for subplan in plan.fetchers) == len(report.files)


def test_plan_generation_ancillaries_only(resource_path_root: Path):
    """
    Can we generate a test plan with all the science and ancillary products?

    :param resource_path_root:
    :return:
    """
    fake_factory = FakeFactory()
    report = FileLocator(resource_path_root / "location_files" / "CALIBRATION.json").locate()
    assert report.files
    assert report.ancillaries

    fetchers = report.fetchers(FetchAncillaryOnly(), fake_factory)

    ancillary_files = [
        ancillary_file for ancillary_prod in report.ancillaries.values() for ancillary_file in ancillary_prod.files
    ]
    assert ancillary_files
    # we should absolutely have the same number of fetchers as we have ancillaries
    assert len(fetchers) == len(ancillary_files)

    # let's make a plan
    plan = CommandLineFetchContext.calculate_plan(fake_factory, fetchers)

    # we should have the same number of fetchers in the plan as in the original report
    assert all(isinstance(subplan, FetchPlan) for subplan in plan.fetchers)
    assert sum(len(subplan.fetchers) for subplan in plan.fetchers) == len(ancillary_files)


def test_plan_generation_science_and_ancillaries(resource_path_root: Path):
    """
    Can we generate a test plan with all the science and ancillary products?

    :param resource_path_root:
    :return:
    """
    fake_factory = FakeFactory()
    report = FileLocator(resource_path_root / "location_files" / "CALIBRATION.json").locate()
    assert report.files
    assert report.ancillaries

    fetchers = report.fetchers(FetchBothScienceAndAncillary(), fake_factory)

    ancillary_files = [
        ancillary_file for ancillary_prod in report.ancillaries.values() for ancillary_file in ancillary_prod.files
    ]
    # we should absolutely have the same number of fetchers as we have files + ancillaries
    assert len(fetchers) == len(report.files) + len(ancillary_files)

    # let's make a plan
    plan = CommandLineFetchContext.calculate_plan(fake_factory, fetchers)

    # we should have the same number of fetchers in the plan as in the original report
    assert all(isinstance(subplan, FetchPlan) for subplan in plan.fetchers)
    assert sum(len(subplan.fetchers) for subplan in plan.fetchers) == len(report.files) + len(ancillary_files)


def test_plan_generation_science_and_only_image_ancillaries(resource_path_root: Path):
    """
    Can we generate a test plan with science products and just the image ancillaries?

    :param resource_path_root:
    :return:
    """
    fake_factory = FakeFactory()
    report = FileLocator(resource_path_root / "location_files" / "VLASS_COARSE_CUBE.json").locate()
    assert report.files
    assert report.ancillaries

    fetchers = report.fetchers(FetchScienceAndImageAncillaries(), fake_factory)

    image_ancillary_files = [
        ancillary_file
        for ancillary_prod in report.ancillaries.values()
        for ancillary_file in ancillary_prod.files
        if ancillary_file.is_image_ancillary
    ]
    assert image_ancillary_files  # If we don't have image ancillaries, what's the point of this test?
    # we should absolutely have the same number of fetchers as we have files + image ancillaries
    assert len(fetchers) == len(report.files) + len(image_ancillary_files)

    # let's make a plan
    plan = CommandLineFetchContext.calculate_plan(fake_factory, fetchers)

    # we should have the same number of fetchers in the plan as in the original report
    assert all(isinstance(subplan, FetchPlan) for subplan in plan.fetchers)
    assert sum(len(subplan.fetchers) for subplan in plan.fetchers) == len(report.files) + len(image_ancillary_files)


def test_plan_generation_sdm_only(resource_path_root: Path):
    """
    Can we generate a test plan with all the science and ancillary products?

    :param resource_path_root:
    :return:
    """
    fake_factory = FakeFactory()
    report = FileLocator(resource_path_root / "location_files" / "13B-014.json").locate()
    assert report.files

    fetchers = report.fetchers(SdmOnly(FetchScienceOnly()), fake_factory)

    sdm_files = [sp_file for sp_file in report.files if sp_file.is_sdm]
    # we should absolutely have the same number of fetchers as we have files + ancillaries
    assert len(fetchers) == len(sdm_files)

    # let's make a plan
    plan = CommandLineFetchContext.calculate_plan(fake_factory, fetchers)

    # we should have the same number of fetchers in the plan as in the original report
    assert all(isinstance(subplan, FetchPlan) for subplan in plan.fetchers)
    assert sum(len(subplan.fetchers) for subplan in plan.fetchers) == len(sdm_files)


def test_multiple_locator_fetching(capsys, resource_path_root):
    """
    Can we handle multiple locators?

    :param capsys:
    :return:
    """
    with patch("productfetcher.product_fetcher.CapoConfig") as mocked_capo:
        mocked_capo.return_value.settings.return_value.defaultThreadsPerHost = 1
        img = resource_path_root / "location_files" / "EVLA_IMG_QUICKLOOK.json"
        cal = resource_path_root / "location_files" / "CALIBRATION.json"

        # parse the command line with these two
        fc = CommandLineFetchContext.parse_commandline([CLIParam.FILE.value, str(img), CLIParam.FILE.value, str(cal)])
        assert len(fc.locators) == 2

        # let's make the plan and ensure we have all the stuff we expect from both
        plan = fc.generate_plan()

        # we'll need to open these files ourselves to figure out what fetchers we expect
        for locator_file in [img, cal]:
            for file in FileLocator(locator_file).locate().files:
                seen = False

                # there may be a more "test friendly" way of doing this, such as by asking the plan
                # if it is fetching a certain file, but it seems like a lot of refactoring for this
                # one test, so I'm going to leave it alone for now
                for hostgroup_fetcher in plan.fetchers:
                    seen = seen or file in [fetcher.file for fetcher in hostgroup_fetcher.fetchers]
                assert seen


def test_argument_parsing(capsys):
    """
    Can we parse the command-line arguments passed in?

    :param capsys:
    :return:
    """
    with patch("productfetcher.product_fetcher.CapoConfig") as mocked_capo:
        mocked_capo.return_value.settings.return_value.defaultThreadsPerHost = 1
        # ensure that various combinations do not work
        with pytest.raises(SystemExit):
            # must have an SPL or a file
            CommandLineFetchContext.parse_commandline([])

        # check the dry run value
        fc = CommandLineFetchContext.parse_commandline([CLIParam.DRY.value, CLIParam.FILE.value, "foo"])
        assert fc.dry_run
        fc = CommandLineFetchContext.parse_commandline([CLIParam.FILE.value, "foo"])
        assert not fc.dry_run

        # check the force value
        fc = CommandLineFetchContext.parse_commandline([CLIParam.FORCE.value, CLIParam.FILE.value, "foo"])
        assert fc.force == ForceMode.FORCE
        fc = CommandLineFetchContext.parse_commandline([CLIParam.FILE.value, "foo"])
        assert fc.force == ForceMode.NORMAL

        fc = CommandLineFetchContext.parse_commandline([CLIParam.FILE.value, "foo"])
        assert isinstance(fc.locators[0], FileLocator)
        assert fc.locators[0].file == Path("foo")

        fc = CommandLineFetchContext.parse_commandline([CLIParam.SPL.value, "uid://this/is/a/fakesy"])
        assert isinstance(fc.locators[0], ServiceLocator)
        assert fc.locators[0].spl == "uid://this/is/a/fakesy"

        fc = CommandLineFetchContext.parse_commandline([CLIParam.FILE.value, "foo", CLIParam.CONCURRENCY.value, "732"])
        assert fc.concurrency == 732

        capsys.readouterr()
