#!/usr/bin/env bash
#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.

#
#  A script from the archive repo adapted for Workspaces
#
#  Old archive description:
#       A replacement for the old qaarchive script used by the
#       data analysts.  This script will tar & zip the pipeline
#       weblog into WEBLOG.tar.gz and then collect everything but
#       FITS files in the products directory from a CIPL run into
#       a single tar file (naming convention TBD) that is created
#       in a storage directory for ingestion.
#
#  Arguments:
#       1: Working directory to be worked upon
#       2: Project code of observation
#       3: Start time of running this script
#       4: CAPO profile
#

# Basics: Path modification
set -o errexit -o nounset -o xtrace

#
# Command-line arguments
#

# Dir that identifies particular workflow
WORKFLOW_DIR=$1;shift
# Metadata for the output tar's naming scheme
SDM_ID=$1;shift
CAL_PROCESSING_DATETIME=$1;shift

# version 1 location, if applicable (optional argument)
INITIAL_DIR=${1:-default}

# Assuming pycapo is in the same directory as this script
CUR_DIR=$(dirname "$(realpath "$0")")

# Get the spool, staging and storage paths from CAPO
SPOOL_DIR=$(${CUR_DIR}/pycapo -P "${CAPO_PROFILE}" -q edu.nrao.workspaces.ProcessingSettings.rootDirectory)
STAGE_DIR=$(${CUR_DIR}/pycapo -P "${CAPO_PROFILE}" -q edu.nrao.workspaces.IngestionSettings.stagingDirectory)
STORAGE_DIR=$(${CUR_DIR}/pycapo -P "${CAPO_PROFILE}" -q edu.nrao.workspaces.IngestionSettings.storageDirectory)

SOURCE_DIR=${SPOOL_DIR}/${WORKFLOW_DIR}/products

# File name for output tar of calibration products
#
# Naming scheme: <sdmId>.<calProcessingDatetime>.caltables.tar
#         a.k.a. <project_code>.<sb>.<eb>.<obs_time_in_mjd>.<cal_processing_timestamp>.caltables.tar
FILENAME="${SDM_ID}"."${CAL_PROCESSING_DATETIME}".caltables

#
# For the ability to reproduce results, we'll want the PPR.xml file.  Ensure it is
# included in the products/ directory:
#
# Check for spool/<workflow_dir>/products/*.pprequest.xml. If it exists, do nothing. If not, move base_dir/PPR.xml to products
if test -n "$(find "${SOURCE_DIR}" -maxdepth 1 -name '*.pprequest.xml' -print -quit)"
then
    cp "${SPOOL_DIR}"/"${WORKFLOW_DIR}"/PPR.xml "${SOURCE_DIR}"/PPR.xml
fi

#
# Both the pipeline-YYYMMDDTHHMMSS directory and weblog.tgz should exist.  We prefer the
# directory (in case of updates/edits), but fall back on the tgz file.
#
# Check that they're both home, as we expect
WEBLOG_DIR=$(ls -t "${SOURCE_DIR}" | grep pipeline- | head -1)
if [ -n "$WEBLOG_DIR" ]
then
    # if weblog.tgz exists, we want to remove it
    if [ -e "${SOURCE_DIR}"/weblog.tgz ]
    then
        rm -f "${SOURCE_DIR}"/weblog.tgz
    fi

    # Tar & Zip the weblog
    tar -C "${SOURCE_DIR}" -czf "${SOURCE_DIR}"/weblog.tgz "${WEBLOG_DIR}"

    if [ $? -ne 0 ]
    then
        echo "Creation of weblog.tgz failed, exiting"
        exit 1
    fi
else
    # no weblog directory.  If there's no weblog.tgz file, there's an issue: Issue a warning
    if [ ! -e "${SOURCE_DIR}"/weblog.tgz ]
    then
        echo "Neither weblog.tgz or the weblog directory exist, continuing"
    fi
fi

# If initial version information is provided, package up version 1 weblog for ingestion and
# move to current version products directory for ingestion
if [ "$INITIAL_DIR" != "default" ]; then
  INITIAL_WEBLOG_DIR=$(ls -t "${SPOOL_DIR}"/"${INITIAL_DIR}"/products | grep pipeline- | head -1)
  VALID_INITIAL=true
  if [ -n "$INITIAL_WEBLOG_DIR" ]
  then
      # if weblog.tgz exists, we want to remove it
      if [ -e "${SPOOL_DIR}"/"${INITIAL_DIR}"/products/weblog.tgz ]
      then
          rm -f "${SPOOL_DIR}"/"${INITIAL_DIR}"/products/weblog.tgz
      fi

      # Tar & Zip the weblog
      tar -C "${SPOOL_DIR}"/"${INITIAL_DIR}"/products -czf "${SPOOL_DIR}"/"${INITIAL_DIR}"/products/initial_weblog.tgz "${INITIAL_WEBLOG_DIR}"

      if [ $? -ne 0 ]
      then
          echo "Creation of initial_weblog.tgz failed, exiting"
          exit 1
      fi
      mv "${SPOOL_DIR}"/"${INITIAL_DIR}"/products/initial_weblog.tgz "${SOURCE_DIR}"

  else
      # no weblog directory.  If there's no weblog.tgz file, there's an issue: Issue a warning
      if [ ! -e "${SPOOL_DIR}"/"${INITIAL_DIR}"/products/weblog.tgz ]
      then
        echo "Did not find weblog.tgz or the weblog directory in the initial version's products, checking for working directory copies..."
        WORKING_WEBLOG_DIR=$(ls -t "${SPOOL_DIR}"/"${INITIAL_DIR}"/working | grep pipeline- | head -1)
        if [ -n "$WORKING_WEBLOG_DIR" ];
        then
            # Tar & Zip the weblog
            tar -C "${SPOOL_DIR}"/"${INITIAL_DIR}"/working -czf "${SPOOL_DIR}"/"${INITIAL_DIR}"/working/initial_weblog.tgz "${WORKING_WEBLOG_DIR}"

            if [ $? -ne 0 ]
            then
                VALID_INITIAL=false
                echo "Creation of initial_weblog.tgz failed, continuing"
            fi
          mv "${SPOOL_DIR}"/"${INITIAL_DIR}"/working/initial_weblog.tgz "${SOURCE_DIR}"
        else
          VALID_INITIAL=false
          echo "Neither weblog.tgz or the weblog directory exist in the initial version's products or working directories, continuing"
        fi
      else
        mv "${SPOOL_DIR}"/"${INITIAL_DIR}"/products/weblog.tgz "${SPOOL_DIR}"/"${INITIAL_DIR}"/products/initial_weblog.tgz
        mv "${SPOOL_DIR}"/"${INITIAL_DIR}"/products/initial_weblog.tgz "${SOURCE_DIR}"
      fi
  fi
fi

#
# Check if the staging directory exists, and delete it if it does.
# Otherwise, ingestion might fail.
#
if [[ -d "${STAGE_DIR}/${WORKFLOW_DIR}" ]]
then
  /bin/rm -rf "${STAGE_DIR}/${WORKFLOW_DIR}"
fi

#
# Create a staging subdirectory for this cal. This needs to be done because we just deleted it if it exists.
#
mkdir -p "${STAGE_DIR}"/"${WORKFLOW_DIR}"

#
# tar all non-fits and non-weblog-related files into a tar archive in the storage path
# SSA-6115:  Don't exclude the weblog.tgz: Users and DAs prefer it bundled in.
#
tar --exclude=\*.fits --exclude=pipeline-\* -C "${SPOOL_DIR}"/"${WORKFLOW_DIR}" -cvf "${STAGE_DIR}"/"${WORKFLOW_DIR}"/"${FILENAME}".tar products
if [ $? -ne 0 ]
then
    echo "Creation of main tar file failed, exiting"
    exit 1
fi

#
# Copy the weblog(s) over, for ingestion as an ancillary file
#
/bin/cp -f "${SOURCE_DIR}"/weblog.tgz "${STAGE_DIR}"/"${WORKFLOW_DIR}"

# If not initial version, copy initial weblog
if [ "$INITIAL_DIR" != "default" ]; then
  if [ "$VALID_INITIAL" = true ]; then
        /bin/cp -f "${SOURCE_DIR}"/initial_weblog.tgz "${STAGE_DIR}"/"${WORKFLOW_DIR}"
  fi
fi

if [ $? -ne 0 ]
then
    echo "Copy of the weblog to staging location failed.  Exiting."
    exit 1
fi

#
# To stay consistent with current working methods: Copy from STAGE_DIR to STORAGE_DIR
#
mkdir -p "${STORAGE_DIR}"/"${WORKFLOW_DIR}"
cp "${STAGE_DIR}"/"${WORKFLOW_DIR}"/"${FILENAME}".tar "${STORAGE_DIR}"/"${WORKFLOW_DIR}"/"${FILENAME}".tar
if [ $? -ne 0 ]
then
    # If something goes wrong, make some noise, but continue on.
    echo "Failed to copy the calibration to ${STORAGE_DIR}, continuing."
    $? = 0
fi
