import json
import logging
import pytest

from datetime import datetime
from unittest.mock import Mock
from aat_wrest.metadata_wrester import WrestWorkflowMetadata


# TODO: Finish adding the unit tests for the WrestWorkflowMetadata class
class TestWrestWorkflowMetadata:

    def validate_date_string(self, date_string):
        # Testing utility method to ensure that strings match the expected date format (eg 2024-10-30T09:57:23)
        assert(bool(datetime.strptime(date_string, "%Y-%m-%dT%H:%M:%S")))

    def create_mock_connection_object(self, return_value):
        # Testing utility method to create a reusable mock object to handle the functions of the MDDBConnector class
        mock_db_connection = Mock()
        mock_cursor = Mock()

        mock_db_connection.cursor.return_value = mock_cursor
        mock_cursor.fetchall.return_value = return_value
        return mock_db_connection

    def test_wrest_standard_cal_info(self):
        """
        Tests the wrest_standard_cal_info method when there is valid data in the database for the given spl.
        """
        # Set up the mocks and their returns
        db_return_tuple = ("sdmId_result", "projectCode_result", "title_result", "startTime_result", "observer_result",
                           "telescope_result", "observing_types_result")
        connection = self.create_mock_connection_object([db_return_tuple])

        # Create the test object and run it
        metadata_wrester = WrestWorkflowMetadata(connection, "spl", "sdm_id", "mous_id")
        wrest_cal_output = metadata_wrester.wrest_standard_cal_info()

        # Ensure the output data has the fields we expect with the values we expect
        wrest_cal_output = json.loads(wrest_cal_output)
        assert(len(wrest_cal_output.keys()) == 8)
        assert(wrest_cal_output["sdmId"] == "sdmId_result")
        assert(wrest_cal_output["projectCode"] == "projectCode_result")
        assert(wrest_cal_output["title"] == "title_result")
        assert(wrest_cal_output["startTime"] == "startTime_result")
        assert(wrest_cal_output["observer"] == "observer_result")
        assert(wrest_cal_output["telescope"] == "telescope_result")
        assert(wrest_cal_output["observingTypes"] == "observing_types_result")
        self.validate_date_string(wrest_cal_output["created_at"])

    @pytest.mark.usefixtures("caplog")
    def test_wrest_standard_cal_info_no_data(self, caplog):
        """
        Tests the wrest_standard_cal_info method when there is no data in the database for the given spl.
        """
        # Set up the mocks and their returns
        connection = self.create_mock_connection_object([])

        # Create the test object and run it
        metadata_wrester = WrestWorkflowMetadata(connection, "bad_spl", "sdm_id", "mous_id")
        wrest_cal_output = metadata_wrester.wrest_standard_cal_info()

        # Ensure the output data is an empty dict and that the module logged an error level message with the correct spl
        assert(wrest_cal_output == {})

        caplog.set_level(logging.ERROR)
        error_log_string = "ERROR: aat-wrest query returned no results! The database appears to be missing information for SPL bad_spl!"
        assert error_log_string in caplog.text
