#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
"""

Core Sampler

This program extracts a "core sample" from a database. You supply a database name, a table name, and a primary key.
The core sampler outputs an SQL file you can use to load the core sample into a copy of the database somewhere else.

"""

import argparse
from typing import Literal

import psycopg2 as pg
from psycopg2 import extras
from pycapo import CapoConfig

from .database import PGTable
from .interfaces import RowSet, Table, RowWriter
from .row_writer import (
    PostgresCopyRowWriter,
    TopologicallySortingRowWriter,
    UniquifyingRowWriter,
    DictionaryKeeper,
)

# pylint: disable=C0103, E0402, R0201, R0903


# stolen shamelessly from aat_wrest
class MDDBConnector:
    """Use this connection to interrogate this science product locator"""

    def __init__(self):
        self.connection = self._connect_to_mddb()

    def _connect_to_mddb(self):
        """
        Establish a DB connection

        :return:
        """
        settings = CapoConfig().settings("metadataDatabase")

        host_slash_db = settings["jdbcUrl"][len("jdbc:postgresql://") :]
        host, dbname = host_slash_db.split("/")
        port = 5432
        if ":" in host:
            host, port = host.split(":")
        try:
            conn = pg.connect(
                host=host,
                port=port,
                database=dbname,
                user=settings.jdbcUsername,
                password=settings.jdbcPassword,
            )
            return conn
        except Exception as exc:
            print(f"Unable to connect to database: {exc}")
            raise exc

    def cursor(self):
        return self.connection.cursor(cursor_factory=extras.RealDictCursor)

    def commit(self):
        self.connection.commit()

    def close(self):
        self.connection.close()


class CoreSampler:
    """
    A device for retrieving core samples from a database. When used from main, it is called as
    ``CoreSampler(connection, format="sql")`` and the result is PostgreSQL ``COPY`` statements
    written to standard output. To use it for obtaining a dictionary of differences, do this:

    .. code-block:: python

        sampler = CoreSampler(conn, format="dict")
        list_of_tables_and_rows = sampler.sample_project("uid://evla/execblock/...").items

    The returned items will look like this in YAML format:

    .. code-block:: yaml

        table_name:
          - column1: value1
            column2: value2
        ...

    Note that the same table name may appear multiple times in the output. This will occur
    if there are dependencies between rows that necessitate it, such as self-references that
    have references to values in other tables. It should always be possible to insert directly
    into a new database from these rows.
    """

    def __init__(self, connection, *, format: Literal["sql", "dict"] = "sql"):
        self.connection = connection

        # if there is no supplied writer, we generate one carefully
        if format == "sql":
            writer = TopologicallySortingRowWriter(UniquifyingRowWriter(PostgresCopyRowWriter()))
        else:
            writer = TopologicallySortingRowWriter(UniquifyingRowWriter(DictionaryKeeper()))

        self.writer = writer
        self.exclusions = set()
        self.visited = set()

    def set_exclusions(self, table_list: list):
        """
        Set any specified tables to ignore

        :param table_list: the list of user specified tables
        :return: a set object
        """
        self.exclusions.update(table_list)

    def sample_science_product(self, product_locator: str) -> RowWriter:
        """
        Get metadata from the archive database for a certain science product locator

        :param product_locator: SPL of interest
        :return:
        """

        # the first time through, we select from the science_products table and get that row
        science_products = self.table("science_products")
        requested = science_products.fetch({"science_product_locator": product_locator})
        self.save(requested)
        self.writer.close()
        return self.writer.bottom()

    def sample_product_group(self, product_group_id: str) -> RowWriter:
        """
        Get metadata from the archive database for a certain product group, including individual product information

        :param product_group_id: group id of interest
        :return:
        """
        # get the top level product group information
        product_groups = self.table("product_groups")
        product_group_rows = product_groups.fetch({"product_group_id": product_group_id})
        self.save(product_group_rows)

        # the product group table doesn't directly connect to the science products table via a key,
        # so the sampler can't find them by default; grab them manually
        products = self.table("science_products")
        finder = ProductFinder(self.connection)
        science_product_locators = finder.find_product_locators(product_group_id)
        science_product_rows = products.fetch_from_many({"science_product_locator": science_product_locators})
        self.save(science_product_rows)

        # ancillary products can be linked either by group id or a spl, but never both; the previous code grabs
        # ancillaries if it is associated with a science product, so grab the ones connected by the group id
        ancillary_products = self.table("ancillary_products")
        finder = AncillaryProductFinder(self.connection)
        ancillary_product_locators = finder.find_ancillary_product_locators(
            [product_group_id], science_product_locators
        )
        ancillary_rows = ancillary_products.fetch_from_many({"ancillary_product_locator": ancillary_product_locators})

        # ancillaries can be accessed via science products, clear that table so entries will save
        self.visited.clear()
        self.save(ancillary_rows)

        # the science products table doesn't directly connect to the projects table via a key,
        # so the sampler can't find them by default; grab the project code entries manually
        products_projects = self.table("science_products_projects")
        finder = ProductsProjectsFinder(self.connection)
        project_codes = finder.find_project_codes_for_product(science_product_locators)
        products_projects_rows = products_projects.fetch_from_many(
            {"project_code": project_codes, "science_product_locator": science_product_locators}
        )
        self.save(products_projects_rows)

        # finally, grab the files associated with the entries
        filegroup_ids = [row["filegroup_id"] for row in science_product_rows] + [
            row["filegroup_id"] for row in ancillary_rows
        ]
        filegroups = self.table("filegroups")
        filegroup_rows = filegroups.fetch_from_many({"filegroup_id": filegroup_ids})
        self.save(filegroup_rows)

        self.writer.close()
        return_contents = self.writer.bottom()
        return return_contents

    def remove_product_group(self, product_group_id: str):
        """
        Remove metadata from the archive database for a certain product group, including individual product information
        comprising the group

        :param product_group_id: group id of interest
        :return:
        """
        # get the top level product group information
        projects = self.table("product_groups")
        requested_product_group_rows = projects.fetch({"product_group_id": product_group_id})

        # the product group table doesn't directly connect to the science products table via a key,
        # so the sampler can't find them by default; grab them manually
        products = self.table("science_products")
        finder = ProductFinder(self.connection)
        science_product_locators = finder.find_product_locators(product_group_id)
        requested_product_rows = products.fetch_from_many({"science_product_locator": science_product_locators})

        # ancillary products can be linked either by group id or a spl, but never both; the previous code grabs
        # ancillaries if it is associated with a science product, so override that with all ancillaries
        ancillary_products = self.table("ancillary_products")
        finder = AncillaryProductFinder(self.connection)
        ancillary_product_locators = finder.find_ancillary_product_locators(
            [product_group_id], science_product_locators
        )
        requested_ancillary_rows = ancillary_products.fetch_from_many(
            {"ancillary_product_locator": ancillary_product_locators}
        )

        # the science products table doesn't directly connect to the projects table via a key,
        # so the sampler can't find them by default; grab the project code entries manually
        products_projects = self.table("science_products_projects")
        finder = ProductsProjectsFinder(self.connection)
        project_codes = finder.find_project_codes_for_product(science_product_locators)
        requested_project_product_rows = products_projects.fetch_from_many(
            {"project_code": project_codes, "science_product_locator": science_product_locators}
        )

        # Create one dictionary pairing table names to row sets
        rows_dict = self.create_row_dict(requested_product_group_rows, {})
        rows_dict = self.create_row_dict(requested_product_rows, rows_dict)
        rows_dict = self.create_row_dict(requested_ancillary_rows, rows_dict)
        rows_dict = self.create_row_dict(requested_project_product_rows, rows_dict)

        # the correct table deletion order is not the same as the reverse of the topographic map, set the order
        deletion_order = [
            "ancillary_products",
            "science_products_projects",
            "science_products_product_groups",
            "product_groups",
            "images",
            "rh_locators",
            "science_products",
        ]

        self.delete_from_product_group(deletion_order, rows_dict)
        self.writer.close()

    def sample_project(self, project_code: str) -> RowWriter:
        """
        Get project metadata from the archive database.

        :param project_code: project code of interest
        :return:
        """

        # the first time through, we select from the projects table and get that row
        projects = self.table("projects")
        requested = projects.fetch({"project_code": project_code})
        self.visited.clear()
        self.save(requested)

        products = self.table("science_products")
        finder = ProductFinder(self.connection)
        science_product_locators = finder.find_product_locators_by_project_code(project_code)
        science_product_rows = products.fetch_from_many({"science_product_locator": science_product_locators})
        self.save(science_product_rows)

        filegroup_ids = [row["filegroup_id"] for row in science_product_rows]
        filegroups = self.table("filegroups")

        groups = self.table("product_groups")
        product_groups = finder.find_product_groups_by_locator(science_product_locators)
        product_groups_rows = groups.fetch_from_many({"product_group_id": product_groups})
        self.save(product_groups_rows)

        ancillary_products = self.table("ancillary_products")
        finder = AncillaryProductFinder(self.connection)
        ancillary_product_locators = finder.find_ancillary_product_locators(product_groups, science_product_locators)
        ancillary_rows = ancillary_products.fetch_from_many({"ancillary_product_locator": ancillary_product_locators})
        self.save(ancillary_rows)

        filegroup_ids = filegroup_ids + [row["filegroup_id"] for row in ancillary_rows]
        filegroup_rows = filegroups.fetch_from_many({"filegroup_id": filegroup_ids})
        self.save(filegroup_rows)

        self.writer.close()
        return self.writer.bottom()

    def sample_eb(self, sdm_name: str) -> RowWriter:
        """
        Pull execution block metadata from the archive database.

        :param sdm_name: SDM of interest
        :return:
        """
        ebs = self.table("execution_blocks")

        finder = ExecBlockFinder(self.connection)
        eb_id = finder.find_eb_id(sdm_name)
        requested = ebs.fetch({"execution_block_id": eb_id})
        self.save(requested)
        self.writer.close()
        return self.writer.bottom()

    def save(self, rows: RowSet):
        """
        Save some rows, and then go and fetch their related rows and save them too, recursively.

        :param rows:  the seed rows to start from
        """
        # bail out if we've already seen this table, or it's to be excluded

        if rows.table.name in self.visited.union(self.exclusions):
            return

        self.visited.add(rows.table.name)
        self.write(rows)
        for relation in rows.relations():
            more = relation.fetch_related_to(rows)

            # recur, if we have some rows
            if len(more) > 0:
                self.save(more)

    def delete_from_product_group(self, deletion_order, rows_dict: dict):
        """
        Fetch the related rows and delete starting from the provided deletion order.

        :param deletion_order:  list of strings representing the table order in which entries should be deleted
        :param rows_dict:  dictionary mapping table name to row set containing all relevant entries for the table
        """
        # Delete rows for each table in the deletion order, starting from the top of the list
        for table_name in deletion_order:
            if table_name in rows_dict.keys():
                rows_dict[table_name].delete()
            else:
                print(f"{table_name} not present in dictionary for this product group, ignoring...")

        # only commit the deletes if everything was removed without error
        self.connection.commit()

    def create_row_dict(self, starting_rowset, rowset_dict):
        """
        Build out a dictionary pairing table names to row set objects, accumulating by table all
        rows related to the given seed rows.

        :param starting_rowset:  the seed rows to start from
        :param rowset_dict:  the dictionary to add to
        """
        # when traversing from the bottom, multiple tables can connect to the same table; capture all rows
        if starting_rowset.table.name in rowset_dict.keys():
            rowset_dict[starting_rowset.table.name].combine_rows(starting_rowset)
        else:
            rowset_dict[starting_rowset.table.name] = starting_rowset

        # traverse the other tables related to this one, if necessary
        for relation in starting_rowset.relations():
            more = relation.fetch_related_to(starting_rowset)
            if len(more) > 0:
                rowset_dict = self.create_row_dict(more, rowset_dict)

        return rowset_dict

    def table(self, name: str) -> Table:
        """
        Return a Table with the given name.

        :param name: name of the table we're talking about
        :return: a Table with this name
        """
        return PGTable(self.connection.cursor(), name)

    def write(self, rows: RowSet):
        """
        Record the rows we've found.

        :param rows:  the rows to record
        """
        rows.write_to(self.writer)


class ProductsProjectsFinder:
    """Looks up project codes for a list of spls"""

    def __init__(self, connection: MDDBConnector):
        self.connection = connection

    def find_project_codes_for_product(self, spls: [str]) -> [str]:
        cursor = self.connection.cursor()
        sql = """
            SELECT DISTINCT project_code FROM science_products_projects
            WHERE science_product_locator IN %(spls)s
        """
        cursor.execute(sql, {"spls": tuple(spls)})
        data = cursor.fetchall()

        if len(data) != 1:
            raise Exception("Either too many project codes or no project codes were found for product group.")

        # All spls will have the same project code, but return a list with a length matching the spls list
        return [data[0]["project_code"]] * len(spls)


class ProductFinder:
    """Looks up science product locators for a product group id"""

    def __init__(self, connection: MDDBConnector):
        self.connection = connection

    def find_product_locators(self, product_group_id: str) -> [str]:
        cursor = self.connection.cursor()
        sql = """
            SELECT science_product_locator FROM science_products_product_groups
            WHERE product_group_id=%(product_group_id)s
        """
        cursor.execute(sql, {"product_group_id": product_group_id})
        data = cursor.fetchall()
        return [item["science_product_locator"] for item in data]

    def find_product_locators_by_project_code(self, project_code: str) -> [str]:
        cursor = self.connection.cursor()
        sql = """
            SELECT science_product_locator FROM science_products_projects
            WHERE project_code=%(project_code)s
        """
        cursor.execute(sql, {"project_code": project_code})
        data = cursor.fetchall()
        return [item["science_product_locator"] for item in data]

    def find_product_groups_by_locator(self, locators: list) -> []:
        cursor = self.connection.cursor()
        sql = """
        SELECT product_group_id FROM science_products_product_groups
        WHERE science_product_locator in %(locator_list)s
        """
        cursor.execute(sql, {"locator_list": tuple(locators)})
        data = cursor.fetchall()
        return [item["product_group_id"] for item in data]


class AncillaryProductFinder:
    """Looks up ancillary product locators for a product group id and list of spls"""

    def __init__(self, connection: MDDBConnector):
        self.connection = connection

    def find_ancillary_product_locators(self, product_group_id: [str], spls: [str]) -> [str]:
        cursor = self.connection.cursor()
        pg_where = (
            "product_group_id IN %(product_group_ids)s"
            if len(product_group_id) > 1
            else "product_group_id=%(product_group_ids)s"
        )
        sql = f"""
            SELECT ancillary_product_locator FROM ancillary_products
            WHERE {pg_where}
            OR science_product_locator IN %(spls)s
        """
        group_ids = (
            None
            if len(product_group_id) == 0
            else (tuple(product_group_id) if len(product_group_id) > 1 else product_group_id[0])
        )

        cursor.execute(
            sql,
            {
                "product_group_ids": group_ids,
                "spls": tuple(spls),
            },
        )
        data = cursor.fetchall()
        return [item["ancillary_product_locator"] for item in data]


class ExecBlockFinder:
    """Looks up execution block ID for an SDM"""

    def __init__(self, connection: MDDBConnector):
        self.connection = connection

    def find_eb_id(self, sdm_name: str) -> int:
        """
        Find an execution block id for an sdm

        :param sdm_name:
        :return:
        """
        cursor = self.connection.cursor()
        sql = """
SELECT execution_block_id FROM execution_blocks
WHERE ngas_fileset_id=%(sdm_name)s
        """
        cursor.execute(sql, {"sdm_name": sdm_name})
        data = cursor.fetchone()
        return data["execution_block_id"]


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-T",
        "--exclude-table",
        type=str,
        help="tables to exclude from return",
        action="append",
    )
    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument(
        "-p",
        "--project_code",
        type=str,
        nargs=1,
        help="Project code from which to start core sampling",
        action="store",
    )
    group.add_argument(
        "-s",
        "--science_product_locator",
        type=str,
        nargs=1,
        help="Product locator id from which to start core sampling",
        action="store",
    )
    group.add_argument(
        "-g",
        "--product_group_id",
        type=str,
        nargs=1,
        help="Product group id from which to start core sampling",
        action="store",
    )
    group.add_argument(
        "-e",
        "--sdm_name",
        type=str,
        nargs=1,
        help="name of SDM for which to pull a core sample, e.g., 21A-409.sb39530397.eb39561636.59309.07888592593",
        action="store",
    )
    ns = parser.parse_args()
    conn = MDDBConnector()
    sampler = CoreSampler(conn)

    try:
        if ns.exclude_table:
            sampler.set_exclusions(ns.exclude_table)
        if ns.project_code:
            sampler.sample_project(ns.project_code[0])
        elif ns.science_product_locator:
            sampler.sample_science_product(ns.science_product_locator[0])
        elif ns.product_group_id:
            sampler.sample_product_group(ns.product_group_id[0])
        elif ns.sdm_name:
            sampler.sample_eb(ns.sdm_name[0])
    finally:
        conn.close()

    return 0


if __name__ == "__main__":
    main()
