/*
 * Copyright (C) 2023 Associated Universities, Inc. Washington DC, USA.
 *
 * This file is part of NRAO Workspaces.
 *
 * Workspaces is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Workspaces is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
 *
 */
import { Component, Input, OnInit } from "@angular/core";
import { CapabilityRequest } from "../../../../model/capability-request";
import { Capability } from "../../../../model/capability";
import { CapabilityVersion } from "../../../../model/capability-version";
import { WorkflowService } from "../../../../services/workflow.service";
import { CapabilityRequestService } from "../../../../services/capability-request.service";
import {ProductMetadata} from "../../../../model/archive-models";

@Component({
  selector: "app-capability-data-access",
  templateUrl: "./capability-data-access.component.html",
  styleUrls: ["./capability-data-access.component.scss"],
})
export class CapabilityDataAccessComponent implements OnInit {
  @Input() capabilityRequest: CapabilityRequest;
  @Input() capability: Capability;
  @Input() currentVersion: CapabilityVersion;
  @Input() productMetadata: ProductMetadata | null = null

  public qaNotes: string;

  // States and names where the version should be downloadable
  private download_names: string[] = ["download", "restore_cms", "std_restore_imaging"];
  private download_states = ["Complete"];

  // States and names where the version should have a weblog
  private weblog_names = [
    "std_calibration",
    "std_cms_imaging",
    "std_calibration_and_imaging",
    "restore_cms",
    "std_restore_imaging",
  ];
  private weblog_states = [
    "Awaiting QA",
    "Stage 2 Review",
    "Complete",
    "Error",
    "Failed",
    "QA Closed",
  ];

  constructor(
    private workflowService: WorkflowService,
    private capabilityRequestService: CapabilityRequestService,
  ) {}

  ngOnInit(): void {}

  // Observer for capability objects
  private qaObserver = {
    next: (notes) => {
      this.qaNotes = notes["resp"];
    },
    error: (error) => console.error("Error when retrieving capability:" + error),
  };

  private saveQaNotesObserver = {
    next: () => {},
    error: (error) => console.error("Error when saving QA notes edit:", error),
  };

  // Get qa_notes.html contents for editing
  getQaNotesURL(): void {
    const curCapabilityName = this.capability.name;
    const curCapabilityRequestId = this.currentVersion.current_execution
      .current_workflow_request_id;
    this.workflowService
      .retrieveQaNotes(curCapabilityName, curCapabilityRequestId)
      .subscribe(this.qaObserver);
  }

  emitEditEventToParent(edits: string): void {
    const curCapabilityName = this.capability.name;
    const curCapabilityRequestId = this.currentVersion.current_execution
      .current_workflow_request_id;
    this.workflowService
      .saveQaNotes(curCapabilityName, curCapabilityRequestId, edits)
      .subscribe(this.saveQaNotesObserver);
  }

  /**
   *.
   * @return Nothing. Opens the URL for the download page in a new tab.
   */
  openDownloadPage(url: string): void {
    window.open(url, "_blank");
  }

  /**
   *.
   * @return Boolean value. Whether processing is complete.
   */
  processingComplete(): boolean {
    return this.shouldHaveWeblog() || this.canHaveDownload();
  }

  /**
   *.
   * @return Boolean value. Whether the user should be able to edit QA notes.
   */
  canEditQANotes(): boolean {
    return (
      this.shouldHaveWeblog() &&
      !this.capabilityRequest.ingested &&
      this.capability.requires_qa &&
      this.currentVersion.state !== "Failed"
    );
  }

  /**
   * Do we expect current version of capability request to have a weblog?
   *
   @return true if current version's capability and state imply it should have a weblog
   */
  shouldHaveWeblog(): boolean {
    if (
      !this.capabilityRequest ||
      !this.currentVersion.current_execution ||
      !this.currentVersion.current_execution.state_name
    ) {
      return false;
    }

    const request = this.capabilityRequest;
    const version = this.currentVersion;

    return (
      this.weblog_states.includes(version.current_execution.state_name) &&
      this.weblog_names.includes(request.capability_name)
    );
  }

  /**
   * Do we expect current version of capability request to have downloadable products?
   *
   @return true if current version's capability and state imply it should be downloadable, otherwise false
   */
  shouldHaveDownload(): boolean {
    if (
      !this.capabilityRequest ||
      !this.currentVersion.current_execution ||
      !this.download_states.includes(this.currentVersion.current_execution.state_name)
    ) {
      return false;
    } else {
      return this.canHaveDownload();
    }
  }

  /**
   *
   @return true if current version's capability imply it can be downloadable at some point
   */
  canHaveDownload(): boolean {
    return this.download_names.includes(this.capabilityRequest.capability_name);
  }
}
