/*
 * Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
 *
 * This file is part of NRAO Workspaces.
 *
 * Workspaces is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Workspaces is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
 */
import { Component, EventEmitter, Input, OnInit, Output } from "@angular/core";
import { CapabilityRequest } from "../../../../model/capability-request";
import { Capability } from "../../../../model/capability";
import { CapabilityLauncherService } from "../../../../services/capability-launcher.service";
import { WorkflowLauncherService } from "../../../../services/workflow-launcher.service";
import { WorkflowService } from "../../../../services/workflow.service";
import { CapabilityExecution } from "../../../../model/capability-execution";
import { CapabilityRequestService } from "../../../../services/capability-request.service";
import { CapabilityVersion } from "../../../../model/capability-version";
import { CapabilitiesService } from "../../../../services/capabilities.service";

@Component({
  selector: "app-request-operations",
  templateUrl: "./request-operations.component.html",
  styleUrls: ["./request-operations.component.scss"],
})
export class RequestOperationsComponent implements OnInit {
  constructor(
    public capabilityLauncherService: CapabilityLauncherService,
    public workflowLauncherService: WorkflowLauncherService,
    public workflowService: WorkflowService,
    private capabilitiesService: CapabilitiesService,
    private capabilityRequestService: CapabilityRequestService,
  ) {}
  @Input() capabilityRequest: CapabilityRequest;
  @Input() latestVersion: CapabilityVersion;
  @Input() selectedVersion: CapabilityVersion;
  @Input() capability: Capability;
  public capabilityExecution: CapabilityExecution;
  @Output() cartaLaunched: EventEmitter<any> = new EventEmitter<any>();
  public cartaResponse: any;
  public hasBeenClicked: boolean = false;
  public cartaEmail: string;
  public defaultCC: string;
  public analystList: string;
  public emailParams: any;

  // Observer for submitting capability request objects (returns a capability execution)
  public submitRequestObserver = {
    next: (submitResponse: CapabilityExecution) => {
      this.capabilityExecution = submitResponse;
      window.location.reload();
    },
    error: (error) => {
      console.log(error);
    },
  };

  public launchCartaObserver = {
    next: (resp: any) => {
      this.cartaResponse = resp;
      this.hasBeenClicked = true;
      // this.cartaLaunched.emit({ type: "carta", email: this.cartaEmail });
    },
    error: (error) => {
      console.log(error);

      this.cartaLaunched.emit({ type: "error" });
    },
  };

  public followonRequestObserver = {
    next: (followonResponse: CapabilityRequest) => {
      if (followonResponse.id) {
        // Capability request created; ID found
        // Redirect to request status page
        this.capabilityRequestService.redirectToRequestStatusPage(followonResponse.id);
      }
    },
    error: (error) => {
      console.log(error);
    },
  };

  public cancelVersionObserver = {
    next: () => {
      window.location.reload();
    },
    error: (error) => {
      console.log(error);
    },
  };



  /**
   * Used to determine if the user can create a follow-on image request.
   *
   * @return Whether or not the request has been partial-passed. Currently, this
   *         is only set for 'std_calibration_and_imaging' capability versions.
   */
  private partialPassed(): boolean {
    return this.selectedVersion.parameters["is_partial_pass"] === true;
  }

  /**
   * Used to determine if the user can
   * create a follow-on image request.
   *
   * @return Whether or not we should
   *         allow a follow-on image request.
   */
  public allowFollowOnImageRequest(): boolean {
    return (
      this.capabilityRequest.state === "Complete" &&
      (this.capabilityRequest.capability_name === "std_calibration" ||
        this.capabilityRequest.capability_name === "restore_cms" ||
        (this.capabilityRequest.capability_name === "std_calibration_and_imaging" &&
          this.partialPassed())) &&
      this.selectedVersion.current_execution != null
    );
  }

  /**
   * Refresh the window which will show that no more versions can be made on this request after it's closed.
   */
  public closeRequestObserver = {
    next: () => {
      window.location.reload();
    },
    error: (error) => {
      console.log(error);
    },
  };

  launchCarta($event) {
    // In this case we do actually want the user_email from the parent version, it should (currently) be the analyst list
    if (this.selectedVersion.parameters["user_email"] !== null &&
      this.selectedVersion.parameters["user_email"] === this.analystList ) {
      this.cartaEmail = this.analystList;
    } else {
      this.cartaEmail = null;
    }

    let email = this.cartaEmail !== null ? this.cartaEmail : "n/a"
    let params = {
      need_data: false,
      notify_ready: true,
      product_locator: null,
      user_email: email,
      parent_wf_request_id: this.selectedVersion.current_execution.current_workflow_request_id,
    };
    this.workflowLauncherService
      .createAndSubmitWorkflow("carta", params)
      .subscribe(this.launchCartaObserver);
  }

  cancelVersion() {
    const requestId = this.selectedVersion.current_execution["current_workflow_request_id"];
    if (requestId == null) {
      console.log("This execution, #" + this.selectedVersion.current_execution.id + ", has no workflow request.");
    } else {
      this.workflowService.abortWorkflow(requestId).subscribe(this.cancelVersionObserver);
    }
  }

  isVersionCancellable(): boolean {
    return (
        this.selectedVersion &&
        this.selectedVersion.state === 'Running' &&
        this.selectedVersion.current_execution.state_name !== 'Stage 2 Review' &&
        this.selectedVersion.current_execution.state_name !== 'Awaiting QA' &&
        this.selectedVersion.current_execution.state_name !== 'Ingesting'
      )
  }

  /**
   * Send a POST request to close (seal) this capability request.
   */
  closeRequest() {
    let request_id = this.capabilityRequest.id;
    this.capabilityRequestService.sendCloseRequest(request_id).subscribe(this.closeRequestObserver);
  }

  /**
   * The request is closable if it's failed or in an error state.
   */
  isRequestCloseable(): boolean {
    let closableVersions = this.capabilityRequest.versions.filter(v => (v.state === "Failed" || v.state === "Error"))

    return (closableVersions.length === this.capabilityRequest.versions.length) && !this.capabilityRequest.sealed
  }

  public loadDefaultCC() {
    const getAnalystEmailObserver = {
      next: (response) => {
        if (response.resp) {
          if (response.resp.includes("test")){
            this.defaultCC = response.resp
          } else {
            this.defaultCC = "schedsoc@nrao.edu," + response.resp;
          }
          // keep around on it's own for general use
          this.analystList = response.resp
        }
      },
      error: (error) => {
        console.error("Failed to load analyst email:", error);
        this.defaultCC = "schedsoc@nrao.edu";
      }
    };

    return this.capabilitiesService.getAnalystEmail().subscribe(getAnalystEmailObserver);
  }

  public loadEmailParams() {
    const getEmailObserver = {
      next: (response) => {
        if (response.resp) {
          this.emailParams = {
            "destination_email": response.resp,
            "version": this.selectedVersion,
            "workflow_metadata": this.selectedVersion.workflow_metadata
          };
        } else {
          this.emailParams = {
            "destination_email": null,
            "version": this.selectedVersion,
            "workflow_metadata": this.selectedVersion.workflow_metadata
          };
        }
      },
      error: (error) => {
        console.error("Failed to load destination email:", error);
        this.emailParams = {
          "destination_email": null,
          "version": this.selectedVersion,
          "workflow_metadata": this.selectedVersion.workflow_metadata
        };
      }
    };

    return this.capabilityRequestService
      .getUserEmail(this.capabilityRequest.id, this.selectedVersion.version_number)
      .subscribe(getEmailObserver);
  }

  public getEmailParams() {
    return this.emailParams;
  }

  ngOnInit(): void {
    // these only need to populate for standard processing capabilities. Let's avoid errors
    // TODO: figure out why this.capability isn't available to this component here as expected.
    //  currently 'undefined' when accessed here
    if (this.capabilityRequest.capability_name.startsWith('std')) {
      this.loadDefaultCC();
      this.loadEmailParams();
    } else {
      this.defaultCC = null
      this.emailParams = {}
    }
  }
}
