#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
""" Tests for ArchiveService """

# pylint: disable=C0103, E0401, R0201, W0621
import http
import json
from typing import Dict
from unittest.mock import MagicMock, patch, PropertyMock, call

import pytest
from mock_alchemy.mocking import UnifiedAlchemyMagicMock

from ..conftest import DUMMY_EMAIL
from workspaces.system.schema import StandardProcessingSettings, AutomaticSubmitStatus
from workspaces.system.services.archive_service import (
    ArchiveService,
    PauseState,
    RunState,
    StandardProcessingManager,
    StopState,
)

SPL = "uid://evla/execblock/8fbfb54b-d141-42fe-b079-609339a69cba"
ROUTER_PATCH = "workspaces.system.services.archive_service.Router"

standard_settings_json = {
    "type": "StandardProcessingSettings",
    "default_capability": "std_calibration_and_imaging",
    "auto_submit_status": "RUN",
}
standard_settings_json2 = {
    "type": "StandardProcessingSettings",
    "default_capability": "std_calibration",
    "auto_submit_status": "PAUSE",
}


@pytest.fixture(scope="module")
def fake_metadata() -> json:
    return json.dumps(
        {
            "spl": SPL,
            "bands": "Ka",
            "array_config": "D",
            "obs_start_time": "1234567",
            "obs_end_time": "123598898",
        }
    )


class TestArchiveService:
    """Test class for ArchiveService"""

    def test_on_ingestion_complete(self, fake_metadata: Dict[str, str]):
        ingestion_message = {
            "user": "test_user",
            "hostname": "hostname01",
            "timestamp": {
                "date": {"year": 2021, "month": 6, "day": 17},
                "time": {"hour": 18, "minute": 24, "second": 52, "nano": 18355000},
            },
            "application": "unknown",
            "request": "snek asks nice-like",
            "version": 1.0,
            "message": "Ingestion complete",
            "logData": {
                "ingestion_type": "rawdata",
                "execblock_id": "29223944",
                "override_db_screen": "False",
                "fileset_id": "13B-014.sb29151475.eb29223944.56810.442529050924",
                "telescope": "EVLA",
                "schedblock_id": "29151475",
                "schedblock_type": "OBSERVER",
                "project_code": "13B-014",
                "file_count": "1",
                "profile": None,
            },
        }

        expected_spl = SPL
        with patch(
            "workspaces.system.services.archive_service.WrestObservationMetadata.wrest_observation_info",
        ) as mock_wrest_obs_metadata, patch(
            "workspaces.system.services.archive_service.MDDBConnector"
        ) as mock_mddbconnector, patch(
            "workspaces.system.services.archive_service.StandardProcessingManager"
        ) as mock_manager:
            mock_wrest_obs_metadata.return_value = fake_metadata
            mock_manager.get_settings.return_value = StandardProcessingSettings.from_json(standard_settings_json)
            with patch(ROUTER_PATCH):
                service = ArchiveService(session_factory=UnifiedAlchemyMagicMock())
                service.capability_manager = MagicMock()
                service.capability_manager.ingestion_complete = MagicMock()
                result = service.on_ingestion_complete(**ingestion_message)
                assert result is not None
                assert service.capability_manager.ingestion_complete.call_count == 1
            assert mock_wrest_obs_metadata.call_count == 1
            assert mock_mddbconnector.call_count == 1
        spl, execblock_id, sdm_id, project_code = result
        assert spl == expected_spl
        assert execblock_id == "29223944"
        assert sdm_id == "13B-014.sb29151475.eb29223944.56810.442529050924"
        assert project_code == "13B-014"

    def test_on_ingestion_complete_without_optional_parameters(self, fake_metadata: Dict[str, str]):
        ingestion_message = {
            "logData": {
                "fileset_id": "13B-014.sb29151475.eb29223944.56810.442529050924",
                "schedblock_type": "OBSERVER",
                "telescope": "EVLA",
            },
        }

        expected_spl = "uid://evla/execblock/8fbfb54b-d141-42fe-b079-609339a69cba"
        with patch(
            "workspaces.system.services.archive_service.WrestObservationMetadata.wrest_observation_info",
        ) as mock_wrest_obs_metadata, patch(
            "workspaces.system.services.archive_service.MDDBConnector"
        ) as mock_mddbconnector, patch(
            "workspaces.system.services.archive_service.StandardProcessingManager"
        ) as mock_manager:
            mock_wrest_obs_metadata.return_value = fake_metadata
            mock_manager.get_settings.return_value = StandardProcessingSettings.from_json(standard_settings_json)
            with patch(ROUTER_PATCH):
                service = ArchiveService(session_factory=UnifiedAlchemyMagicMock())
                service.capability_manager = MagicMock()
                service.capability_manager.ingestion_complete = MagicMock()
                result = service.on_ingestion_complete(**ingestion_message)
                assert result is not None
                assert service.capability_manager.ingestion_complete.call_count == 1
            assert mock_wrest_obs_metadata.call_count == 1
            assert mock_mddbconnector.call_count == 1
        spl, execblock_id, sdm_id, project_code = result
        assert spl == expected_spl
        assert execblock_id == ""
        assert sdm_id == "13B-014.sb29151475.eb29223944.56810.442529050924"
        assert project_code == ""

    def test_on_ingest_complete_returns_nothing_for_vlass(self, fake_metadata: Dict[str, str]):
        """
        Do not submit VLASS calibrations for ingestion.
        """
        ingestion_message = {
            "user": "test_user",
            "hostname": "hostname01",
            "timestamp": {
                "date": {"year": 2021, "month": 6, "day": 17},
                "time": {"hour": 18, "minute": 24, "second": 52, "nano": 18355000},
            },
            "application": "unknown",
            "request": "snek asks nice-like",
            "version": 1.0,
            "message": "Ingestion complete",
            "logData": {
                "ingestion_type": "rawdata",
                "execblock_id": "29223944",
                "override_db_screen": "False",
                "fileset_id": "VLASS1.1.sb29151475.eb29223944.56810.442529050924",
                "telescope": "EVLA",
                "schedblock_id": "29151475",
                "schedblock_type": "SCIENCE",
                "project_code": "VLASS1.1",
                "file_count": "1",
                "profile": None,
            },
        }
        with patch(
            "workspaces.system.services.archive_service.WrestObservationMetadata.wrest_observation_info",
        ) as mock_wrest_obs_metadata, patch(
            "workspaces.system.services.archive_service.StandardProcessingManager"
        ) as mock_manager:
            mock_wrest_obs_metadata.return_value = fake_metadata
            mock_manager.get_settings.return_value = StandardProcessingSettings.from_json(standard_settings_json)
            with patch(ROUTER_PATCH):
                service = ArchiveService(session_factory=UnifiedAlchemyMagicMock())
                service.capability_manager = MagicMock()

                with pytest.raises(TypeError) as t_err:
                    service.capability_manager.ingestion_complete = MagicMock()
                    () = service.on_ingestion_complete(**ingestion_message)
                    result = service.on_ingestion_complete(**ingestion_message)
                    assert str(t_err) == "TypeError: cannot unpack non-iterable NoneType object"
                    assert result is None
                    assert service.capability_manager.ingestion_complete.call_count == 1

    def test_on_ingest_complete_returns_nothing_for_tsky(self, fake_metadata: Dict[str, str]):
        """
        Do not submit TSKY calibrations for ingestion.
        """
        ingestion_message = {
            "user": "test_user",
            "hostname": "hostname01",
            "timestamp": {
                "date": {"year": 2021, "month": 6, "day": 17},
                "time": {"hour": 18, "minute": 24, "second": 52, "nano": 18355000},
            },
            "application": "unknown",
            "request": "snek asks nice-like",
            "version": 1.0,
            "message": "Ingestion complete",
            "logData": {
                "ingestion_type": "rawdata",
                "execblock_id": "29223944",
                "override_db_screen": "False",
                "fileset_id": "TSKY0001.sb29151475.eb29223944.56810.442529050924",
                "telescope": "EVLA",
                "schedblock_id": "29151475",
                "schedblock_type": "SCIENCE",
                "project_code": "TSKY0001",
                "file_count": "1",
                "profile": None,
            },
        }
        with patch(
            "workspaces.system.services.archive_service.WrestObservationMetadata.wrest_observation_info",
        ) as mock_wrest_obs_metadata, patch(
            "workspaces.system.services.archive_service.StandardProcessingManager"
        ) as mock_manager:
            mock_wrest_obs_metadata.return_value = fake_metadata
            mock_manager.get_settings.return_value = StandardProcessingSettings.from_json(standard_settings_json)
            with patch(ROUTER_PATCH):
                service = ArchiveService(session_factory=UnifiedAlchemyMagicMock())
                service.capability_manager = MagicMock()

                with pytest.raises(TypeError) as t_err:
                    service.capability_manager.ingestion_complete = MagicMock()
                    () = service.on_ingestion_complete(**ingestion_message)
                    assert service.capability_manager.ingestion_complete.call_count == 0
                    result = service.on_ingestion_complete(**ingestion_message)
                    assert str(t_err) == "TypeError: cannot unpack non-iterable NoneType object"
                    assert result is None
                    assert service.capability_manager.ingestion_complete.call_count == 1

    @pytest.mark.parametrize("generate_default", ["std_calibration", "std_calibration_and_imaging"])
    def test_create_request(self, fake_metadata: Dict[str, str], generate_default):
        spl = SPL
        email = DUMMY_EMAIL
        params = {
            "parameters": {
                "product_locator": spl,
                "user_email": email,
                "metadata": fake_metadata,
            }
        }

        with patch(ROUTER_PATCH):
            with patch(
                "workspaces.system.services.archive_service.requests.post",
            ) as mock_post, patch(
                "workspaces.system.services.archive_service.StandardProcessingManager.get_settings",
                return_value=StandardProcessingSettings.from_json(standard_settings_json),
            ) as mock_settings, patch("workspaces.system.services.archive_service.CapoConfig") as mock_capo_config:
                mock_post.return_value = MagicMock()
                mock_post.return_value.status_code = http.HTTPStatus.OK
                mock_capo_config.return_value.settings.return_value.serviceUrl = "http://capability:3457"
                mock_capo_config.return_value.settings.return_value.analystEmail = DUMMY_EMAIL
                with patch(
                    "workspaces.system.services.archive_service.StandardProcessingManager.default_processing_capability",
                    new_callable=PropertyMock,
                ) as mock_default:
                    mock_default.return_value = generate_default
                    spm = StandardProcessingManager(session_factory=UnifiedAlchemyMagicMock())
                    spm.create_std_request(spl, fake_metadata)
                    mock_post.assert_called_with(
                        f"http://capability:3457/capability/{generate_default}/request/create", json=params
                    )


class TestStandardProcessingState:
    def test_dispatch(self):
        with patch(
            "workspaces.system.services.archive_service.StandardProcessingManager.state_name", AutomaticSubmitStatus.RUN
        ):
            with patch("workspaces.system.services.archive_service.RunState.ingest_complete") as ic:
                std_cal_and_img = StandardProcessingManager(session_factory=UnifiedAlchemyMagicMock())
                std_cal_and_img.ingestion_complete("fakespl", {})
                assert ic.call_count == 1

        with patch(
            "workspaces.system.services.archive_service.StandardProcessingManager.state_name",
            AutomaticSubmitStatus.PAUSE,
        ):
            with patch("workspaces.system.services.archive_service.PauseState.ingest_complete") as ic:
                std_cal_and_img = StandardProcessingManager(session_factory=UnifiedAlchemyMagicMock())
                std_cal_and_img.ingestion_complete("fakespl", {})
                assert ic.call_count == 1

        with patch(
            "workspaces.system.services.archive_service.StandardProcessingManager.state_name",
            AutomaticSubmitStatus.STOP,
        ):
            with patch("workspaces.system.services.archive_service.StopState.ingest_complete") as ic:
                std_cal_and_img = StandardProcessingManager(session_factory=UnifiedAlchemyMagicMock())
                std_cal_and_img.ingestion_complete("fakespl", {})
                assert ic.call_count == 1

    @staticmethod
    def generate_fake_std_calibration_and_imaging():
        fake_std_calibration_and_imaging = MagicMock()
        fake_std_calibration_and_imaging.create_std_request = MagicMock()
        fake_std_calibration_and_imaging.create_std_request.return_value = -2
        fake_std_calibration_and_imaging.submit_std_request = MagicMock()
        return fake_std_calibration_and_imaging

    def test_run(self):
        fake_std_calibration_and_imaging = self.generate_fake_std_calibration_and_imaging()
        run = RunState(fake_std_calibration_and_imaging)
        run.ingest_complete("fake", {})
        assert fake_std_calibration_and_imaging.create_std_request.call_count == 1
        assert fake_std_calibration_and_imaging.submit_std_request.call_count == 1

    def test_pause(self):
        fake_std_calibration_and_imaging = self.generate_fake_std_calibration_and_imaging()
        run = PauseState(fake_std_calibration_and_imaging)
        run.ingest_complete("fake", {})
        assert fake_std_calibration_and_imaging.create_std_request.call_count == 1
        assert fake_std_calibration_and_imaging.submit_std_request.call_count == 0

    def test_stop(self):
        fake_std_calibration_and_imaging = self.generate_fake_std_calibration_and_imaging()
        run = StopState(fake_std_calibration_and_imaging)
        run.ingest_complete("fake", {})
        assert fake_std_calibration_and_imaging.create_std_request.call_count == 0
        assert fake_std_calibration_and_imaging.submit_std_request.call_count == 0


class TestStandardProcessingManager:
    @staticmethod
    def session1() -> UnifiedAlchemyMagicMock:
        return UnifiedAlchemyMagicMock(
            data=[
                (
                    [call.query(StandardProcessingSettings)],
                    [
                        StandardProcessingSettings(
                            default_capability=standard_settings_json["default_capability"],
                            auto_submit_status=AutomaticSubmitStatus[standard_settings_json["auto_submit_status"]],
                        )
                    ],
                )
            ]
        )

    @staticmethod
    def session2() -> UnifiedAlchemyMagicMock:
        return UnifiedAlchemyMagicMock(
            data=[
                (
                    [call.query(StandardProcessingSettings)],
                    [
                        StandardProcessingSettings(
                            default_capability=standard_settings_json2["default_capability"],
                            auto_submit_status=AutomaticSubmitStatus[standard_settings_json2["auto_submit_status"]],
                        )
                    ],
                )
            ]
        )

    def test_get_settings(self):
        session1 = self.session1()
        session2 = self.session2()

        spm = StandardProcessingManager(session_factory=UnifiedAlchemyMagicMock())

        # test with given session
        settings_1 = spm.get_settings(session1)
        assert isinstance(settings_1, StandardProcessingSettings)
        assert settings_1.auto_submit_status == AutomaticSubmitStatus[standard_settings_json["auto_submit_status"]]
        assert settings_1.default_capability == standard_settings_json["default_capability"]
        assert session1.close.call_count == 0

        # test with generated session
        spm.session_factory.return_value = session2
        settings_2 = spm.get_settings()
        assert isinstance(settings_2, StandardProcessingSettings)
        assert settings_2.auto_submit_status == AutomaticSubmitStatus[standard_settings_json2["auto_submit_status"]]
        assert settings_2.default_capability == standard_settings_json2["default_capability"]
        assert session2.close.call_count == 1

    def test_update_settings(self):
        session_1 = self.session1()
        spm = StandardProcessingManager(session_factory=UnifiedAlchemyMagicMock())
        spm.default_settings = StandardProcessingSettings.from_json(standard_settings_json)
        spm.session_factory.return_value = session_1

        update_dict_1 = {"auto_submit_status": "STOP"}
        updated_setting = spm.update_settings(update_dict_1)
        assert updated_setting.auto_submit_status == AutomaticSubmitStatus.STOP
        # verify default capability didn't change
        assert updated_setting.default_capability == standard_settings_json["default_capability"]
        # verify default still same after refresh
        assert spm.default_processing_capability == updated_setting.default_capability

        update_dict_2 = {"default_capability": standard_settings_json2["default_capability"]}
        updated_settings_2 = spm.update_settings(update_dict_2)
        # verify auto status hasn't changed since last update
        assert updated_settings_2.auto_submit_status == AutomaticSubmitStatus.STOP
        # verify default capability has updated as expected
        assert updated_settings_2.default_capability == standard_settings_json2["default_capability"]
        # verify default updated after refresh
        assert spm.default_processing_capability == updated_setting.default_capability

        # Verify bad data handling
        update_dict_3 = {"auto_submit_status": "WRONG"}
        no_update_expected = spm.update_settings(update_dict_3)
        # verify no change to updated field
        assert no_update_expected.auto_submit_status == AutomaticSubmitStatus.STOP

    def test_refresh_settings(self):
        spm = StandardProcessingManager(session_factory=UnifiedAlchemyMagicMock())
        session = self.session2()
        # set initial settings to be different from current 'db' state
        spm.default_settings = StandardProcessingSettings.from_json(standard_settings_json)
        spm.session_factory.return_value = session

        # verify expected initial state
        assert (
            spm.default_settings.auto_submit_status
            == AutomaticSubmitStatus[standard_settings_json["auto_submit_status"]]
        )
        assert spm.default_processing_capability == standard_settings_json["default_capability"]
        spm.refresh_defaults()
        # verify refresh updates defaults
        assert spm.default_processing_capability == standard_settings_json2["default_capability"]
        assert (
            spm.default_settings.auto_submit_status
            == AutomaticSubmitStatus[standard_settings_json2["auto_submit_status"]]
        )

        # demonstrate initial and final states are different
        assert (
            AutomaticSubmitStatus[standard_settings_json["auto_submit_status"]]
            != AutomaticSubmitStatus[standard_settings_json2["auto_submit_status"]]
        )
