#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
""" Tests for CASA Matrix Service """

import subprocess
from io import StringIO
from pathlib import Path
from unittest.mock import MagicMock, patch

import pytest
from pycapo import CapoConfig
from sqlalchemy import create_engine
from sqlalchemy.orm import Session

from workspaces.capability.schema import CasaMatrixCasaVersion
from workspaces.capability.schema import mapper_registry as capability_mapper_registry
from workspaces.system.services import casa_matrix_service
from workspaces.system.services.casa_matrix_service import (
    CasaMatrixService,
    natural_sort,
)

settings = CapoConfig().settings("edu.nrao.archive.workflow.config.CasaVersions")

DEFAULT_CASA = settings["homeForDownloads"]
VLA_DEFAULT_CASA = settings["homeForVlaRestore"]
ALMA_DEFAULT_CASA = settings["homeForAlmaRestore"]
VUDI_DEFAULT_CASA = settings["homeForVlaUserDrivenImaging"]
AUDI_DEFAULT_CASA = settings["homeForAlmaUserDrivenImaging"]


@pytest.fixture(scope="module")
def fake_version() -> CasaMatrixCasaVersion:
    return CasaMatrixCasaVersion.from_json(
        CasaMatrixCasaVersion,
        {
            "casa_version": "1.2.3-4|default",
            "capabilities": ["restore_cms", "std_calibration"],
            "is_cluster_compatible": True,
        },
    )


@pytest.fixture(scope="module")
def session() -> Session:
    """
    Pytest fixture providing a mock CasaMatrixService object; mocked object can be used
    like a normal CasaMatrixService instance.

    :return: Mocked CasaMatrixService
    """

    engine = create_engine("sqlite:///:memory:")
    capability_mapper_registry.metadata.create_all(engine)
    session = Session(engine)
    this_dir = Path(__file__).parent
    with open(this_dir / "insert_casa_version_matrix.sql", "r") as f:
        insert_casa_version_matrix = f.readlines()
    for insert_statement in insert_casa_version_matrix:
        session.execute(insert_statement)
    session.commit()
    return session


@pytest.fixture(scope="function")
def fake_casa_root(tmp_path: Path) -> Path:
    this_dir = Path(__file__).parent
    out_path = tmp_path / "out"
    err_path = tmp_path / "err"
    with open(out_path, "w") as out:
        with open(err_path, "w") as err:
            a = subprocess.Popen(args=[this_dir / "setup_casa_env_in_cwd.sh", tmp_path], stdout=out, stderr=err)
        exit_code = a.wait()
    if exit_code != 0:
        with open(out_path, "r") as out:
            with open(err_path, "r") as err:
                raise RuntimeError(
                    f"Error encountered when setting up fake CASA root.\nstdout='{out.read()}'\nstderr='{err.read()}'"
                )
    return tmp_path / "home" / "ssa" / "casa"


@pytest.fixture
def mock_capo_config(fake_casa_root: Path) -> MagicMock:
    mock = MagicMock()
    mocked_capo_get = {
        "edu.nrao.archive.archiveIface.casa_root": str(fake_casa_root),
        # Might need to be updated with other capabilities' default_casa_capo_key's as tests expand
        "edu.nrao.archive.workflow.config.CasaVersions": {
            "homeForReprocessing": "/home/casa/packages/pipeline/current"
        },
        "edu.nrao.archive.archiveIface.allowDevCasaVersions": "true",
    }
    mock.return_value.profile = "docker"  # Matches the profile that setup_casa_env_in_cwd.sh uses
    mock.return_value.get.side_effect = lambda arg: mocked_capo_get[arg]
    return mock


class TestCasaMatrixService:
    """Test class for CasaMatrixService"""

    @pytest.mark.skip(reason="Directory structure in the container isn't properly set up")
    def test_make_links(self, mock_matrix_service: CasaMatrixService):
        ret_links = mock_matrix_service.make_links()
        assert len(ret_links) > 0

    @pytest.mark.skip(reason="Default lookup is getting KeyErrors")
    @pytest.mark.parametrize(
        "fake_request",
        [
            {"casa_version": None, "capability": None, "telescope": None},
            {"casa_version": None, "capability": "std_calibration", "telescope": None},
        ],
    )
    def test_default_version(self, mock_matrix_service: CasaMatrixService, fake_request: dict[str, str, str]):
        expected_version = DEFAULT_CASA
        ret_version = mock_matrix_service.get_version(None, fake_request["capability"])
        assert ret_version["version"] == expected_version

    @pytest.mark.skip(reason="Default lookup is getting KeyErrors")
    @pytest.mark.parametrize(
        "fake_request",
        [
            {"casa_version": "5.6.1-8|default", "capability": "restore_cms", "telescope": "VLA"},
            {"casa_version": "5.6.1-8|default", "capability": "restore_basic_cms", "telescope": "VLA"},
        ],
    )
    def test_vla_default_version(self, mock_matrix_service: CasaMatrixService, fake_request: dict[str, str, str]):
        expected_version = VLA_DEFAULT_CASA
        ret_version = mock_matrix_service.get_version(None, fake_request["capability"], fake_request["telescope"])
        assert ret_version["version"] == expected_version

    @pytest.mark.skip(reason="Default lookup is getting KeyErrors")
    @pytest.mark.parametrize(
        "fake_request",
        [
            {"casa_version": "5.6.1-8|default", "capability": "restore_cms", "telescope": "ALMA"},
            {"casa_version": "5.6.1-8|default", "capability": "restore_basic_cms", "telescope": "ALMA"},
        ],
    )
    def test_alma_default_version(self, mock_matrix_service: CasaMatrixService, fake_request: dict[str, str, str]):
        expected_version = ALMA_DEFAULT_CASA
        ret_version = mock_matrix_service.get_version(None, fake_request["capability"], fake_request["telescope"])
        assert ret_version["version"] == expected_version

    @pytest.mark.skip(reason="Default lookup is getting KeyErrors")
    @pytest.mark.parametrize(
        "fake_request",
        [
            {"casa_version": "5.6.1-8|default", "capability": None, "telescope": None},
            {"casa_version": "6.1.1-15|2020.1.0.40", "capability": None, "telescope": None},
        ],
    )
    def test_valid_version(self, mock_matrix_service: CasaMatrixService, fake_request: dict[str, str, str]):
        expected_version = fake_request["casa_version"]
        ret_version = mock_matrix_service.get_version(fake_request["casa_version"])
        assert ret_version["version"] == expected_version

    @pytest.mark.skip(reason="Default lookup is getting KeyErrors")
    @pytest.mark.parametrize(
        "fake_request",
        [
            {"casa_version": "0.0.0-0|default", "capability": None, "telescope": None},
            {"casa_version": "x.x.x-x|default", "capability": None, "telescope": None},
        ],
    )
    def test_invalid_version(self, mock_matrix_service: CasaMatrixService, fake_request: dict[str, str, str]):
        expected_version = DEFAULT_CASA
        ret_version = mock_matrix_service.get_version(fake_request["casa_version"])
        assert ret_version["version"] == expected_version

    @pytest.mark.skip(reason="Default lookup is getting KeyErrors")
    @pytest.mark.parametrize(
        "fake_request",
        [
            {"casa_version": "0.0.0-0|default", "capability": "restore_cms", "telescope": "VLA"},
            {"casa_version": "x.x.x-x|default", "capability": "restore_basic_cms", "telescope": "VLA"},
        ],
    )
    def test_vla_invalid_version(self, mock_matrix_service: CasaMatrixService, fake_request: dict[str, str, str]):
        expected_version = VLA_DEFAULT_CASA
        ret_version = mock_matrix_service.get_version(
            fake_request["casa_version"], fake_request["capability"], fake_request["telescope"]
        )
        assert ret_version["version"] == expected_version

    @pytest.mark.skip(reason="Default lookup is getting KeyErrors")
    @pytest.mark.parametrize(
        "fake_request",
        [
            {"casa_version": "0.0.0-0|default", "capability": "restore_cms", "telescope": "ALMA"},
            {"casa_version": "x.x.x-x|default", "capability": "restore_basic_cms", "telescope": "ALMA"},
        ],
    )
    def test_alma_invalid_version(self, casa_matrix_service: CasaMatrixService, fake_request: dict[str, str, str]):
        expected_version = ALMA_DEFAULT_CASA
        ret_version = casa_matrix_service.get_version(
            fake_request["casa_version"], fake_request["capability"], fake_request["telescope"]
        )
        assert ret_version["version"] == expected_version

    @pytest.mark.skip(reason="type of get_recipe return changed")
    @pytest.mark.parametrize("fake_capability", ["std_calibration"])
    def test_std_calibration_recipe(self, mock_matrix_service: CasaMatrixService, fake_capability: str):
        expected_recipe = "procedure_hifv.xml"
        ret_recipe = mock_matrix_service.get_recipe(fake_capability)
        assert ret_recipe == expected_recipe

    @pytest.mark.skip(reason="type of get_recipe return changed")
    @pytest.mark.parametrize("fake_capability", ["restore_cms", "restore_basic_ms"])
    def test_restore_recipe(self, mock_matrix_service: CasaMatrixService, fake_capability: str):
        expected_recipe = "template_hifv_deliver_ms.xml|template_hifa_deliver_ms.xml"
        ret_recipe = mock_matrix_service.get_recipe(fake_capability)
        assert ret_recipe == expected_recipe

    @pytest.mark.skip(reason="type of get_recipe return changed")
    @pytest.mark.parametrize("fake_capability", ["std_cms_imaging", "std_restore_imaging"])
    def test_imaging_recipe(self, mock_matrix_service: CasaMatrixService, fake_capability: str):
        expected_recipe = "procedure_hifv_contimage_selfcal.xml"
        ret_recipe = mock_matrix_service.get_recipe(fake_capability)
        assert ret_recipe == expected_recipe

    def test_get_versions(self, session: Session, mock_capo_config: MagicMock):
        mocked_capo_settings = {
            "edu.nrao.archive.workflow.config.CasaVersions": {
                "homeForReprocessing": "/home/casa/packages/pipeline/current"
            },
        }
        with patch("workspaces.system.services.casa_matrix_service.CapoConfig", new=mock_capo_config) as mcc:
            mcc.return_value.settings.side_effect = lambda arg: mocked_capo_settings[arg]
            cms = CasaMatrixService(session)
            ret_versions = cms.get_versions(capability="std_calibration")
        assert len(ret_versions) > 0

    @pytest.mark.parametrize(
        "fake_request",
        [
            # versions need to be in the matrix for the given capability
            # capabilities also need their default_casa_capo_key values in the mocked_capo_settings below
            {"version": "6.4.1-12|2022.2.0.64", "capability": "std_calibration", "telescope": None},
            {"version": "6.5.4-9|2023.1.0.124", "capability": "restore_cms", "telescope": None},
        ],
    )
    def test_get_versions_recommended(
        self, session: Session, mock_capo_config: MagicMock, fake_request: dict[str, str]
    ):
        expected_recommended_version = fake_request["version"]
        mocked_capo_settings = {
            "edu.nrao.archive.workflow.config.CasaVersions": {
                "homeForReprocessing": "/home/casa/packages/pipeline/current",
                "homeForVlaRestore": "/home/ssa/casa/RHEL6/docker/casa-pipeline-release-5.6.1-8",
            },
        }
        with patch("workspaces.system.services.casa_matrix_service.CapoConfig", new=mock_capo_config) as mcc:
            mcc.return_value.settings.side_effect = lambda arg: mocked_capo_settings[arg]
            cms = CasaMatrixService(session)
            ret_versions = cms.get_versions(**fake_request)
        assert len(ret_versions) > 0
        assert ret_versions[0]["version"] == expected_recommended_version
        # Ensure there are no versions older than ("before" in natural_sort) the recommended one
        assert expected_recommended_version == natural_sort([v["version"] for v in ret_versions])[0]

    def test_add_version(self, session: Session, fake_version: CasaMatrixCasaVersion):
        casa_matrix_service = CasaMatrixService(session)
        ret = casa_matrix_service.add_version(
            fake_version.casa_version, fake_version.capabilities, fake_version.is_cluster_compatible
        )
        casa_matrix_service.delete_version(fake_version.casa_version)
        assert ret

    def test_update_version(self, session: Session, fake_version: CasaMatrixCasaVersion):
        casa_matrix_service = CasaMatrixService(session)
        casa_matrix_service.add_version(fake_version.casa_version)
        ret = casa_matrix_service.update_version(
            fake_version.casa_version, fake_version.capabilities, fake_version.is_cluster_compatible
        )
        casa_matrix_service.delete_version(fake_version.casa_version)
        assert ret

    def test_delete_version(self, session: Session, fake_version: CasaMatrixCasaVersion):
        casa_matrix_service = CasaMatrixService(session)
        casa_matrix_service.add_version(fake_version.casa_version)
        ret = casa_matrix_service.delete_version(fake_version.casa_version)
        assert ret
