#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.

""" Builds AMQP messages needed for capabilities """

# pylint: disable=E0401

import logging

from immutable_views import DictView

from workspaces.shared_interfaces import MessageArchitectIF

logger = logging.getLogger(__name__)

"""
Class which creates an immutable dictionary of amqp message templates for easy access
"""

# DictView(Dict[msg_type, msg_content]
service_templates = DictView(
    {
        "capability": {
            "service": "capability",
            "routing_key": "capability",
            "subject": None,
            "type": None,
        },
        "archive": {
            "service": "archive",
            "routing_key": None,
            "subject": None,
        },
    }
)

# archive messages that require additional information
archive_msg_templates = DictView(
    {
        "restore_cms_complete": {
            "application": "capability",
            "message": "restore_cms complete",
            "delivery_url": None,
            "delivery_path": None,
            "token": None,
        },
        "restore_cms_failed": {
            "application": "capability",
            "message": "restore_cms failed",
        },
        "restore_cms_cancelled": {
            "application": "capability",
            "message": "restore_cms cancelled",
        },
        "update": {
            "application": "capability",
            "message": "capability updated",
            "stage": None,
            "start": None,
            "stop": None,
        },
    }
)
# messages that require additional information, currently unneeded
# capability_msg_templates = DictView(
#     {
#         "cancel": {
#             "type": "cancel",
#         },
#     }
# )


class CapabilityMessageArchitect(MessageArchitectIF):
    """Creates messages needed used by capabilities"""

    def __init__(
        self,
        execution: "CapabilityExecution" = None,
        request: "CapabilityRequest" = None,
        version: "CapabilityVersion" = None,
    ):
        self.subject = execution or request or version

    @staticmethod
    def get_message_template(msg_type: str) -> DictView:
        # If messages need additional information outside of the basic template,
        # uncomment this and add to relevant DictView
        # if msg_type in capability_msg_templates:
        #     return {**service_templates["capability"], **capability_msg_templates[msg_type]}
        # else:
        return DictView(service_templates["capability"])

    def compose_message(self, msg_type: str) -> DictView:
        template = self.get_message_template(msg_type)
        template = template.copy().dict

        template["subject"] = self.subject
        template["type"] = msg_type.replace("_", "-")
        return DictView(template)


class ArchiveMessageArchitect(MessageArchitectIF):
    """Builds archive system AMQP messages"""

    def __init__(
        self,
        request: "CapabilityRequest",
        routing_key: str = None,
        delivery_info: dict = None,
        stage_info: dict = None,
    ):
        """
        Initialize Architect

        :param routing_key: AMQP routing key to use
        :param request: The capability request this message concerns
        """
        self.routing_key = routing_key if routing_key is not None else self._make_archive_routing_key(request)
        self.request = request
        self.delivery = delivery_info
        self.stage_info = stage_info

    @staticmethod
    def _make_archive_routing_key(cr: "CapabilityRequest") -> str:
        return f"{cr.controller.routing_key_prefix}{cr.capability_name}.{cr.id}"

    @staticmethod
    def get_message_template(msg_type: str) -> DictView:
        """
        Construct the message template based on type

        :param msg_type: type of message to construct
        :return: full message template dictionary for rendering
        """

        if "update" in msg_type:
            return DictView({**service_templates["archive"], **archive_msg_templates["update"]})
        else:
            return DictView({**service_templates["archive"], **archive_msg_templates[msg_type]})

    def compose_message(self, msg_type: str) -> DictView:
        """
        Render message template for requested type

        :param msg_type: the type of message to send
        :return: the rendered message ready for sending
        """
        template = self.get_message_template(msg_type)
        template = template.copy().dict
        template["subject"] = f"WS capability request #{self.request.id}"
        template["routing_key"] = self.routing_key

        if self.delivery is not None:
            template["download_url"] = self.delivery["url"]
            template["delivery_path"] = self.delivery["delivered_to"]
            template["token"] = self.delivery["token"]

        if self.stage_info is not None and "update" in msg_type:
            template["stage"] = self.stage_info["stage"]
            template["start"] = self.stage_info["start"]
            template["stop"] = self.stage_info["stop"]

        return DictView(template)
