#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
from __future__ import annotations

import json
from enum import Enum
from pathlib import Path

import sqlalchemy as sa
from sqlalchemy import Enum as SaEnum

from pycapo import CapoConfig
from sqlalchemy.orm import registry, sessionmaker

mapper_registry = registry()


class JSONSerializable:
    def __json__(self, request=None) -> dict:
        """
        Allows this object to be converted to JSON

        :param request: this parameter is the active Pyramid request, if applicable (None otherwise)
        :return: a dictionary which can be converted to JSON using json.dump
        """
        pass

    @classmethod
    def from_json(cls, json: dict) -> any:
        pass


class AbstractFile(JSONSerializable):
    """
    Abstract file is exactly that, an abstract concept of what a file is, to be
    returned from various non-filesystem places.
    """

    def __init__(self, filename: str, content: bytes):
        self.filename, self.content = filename, content

    def write_to(self, directory: Path):
        (directory / self.filename).write_bytes(self.content)

    @classmethod
    def from_path(cls, path: Path) -> AbstractFile:
        return cls(path.name, path.read_bytes())


class AbstractRequest(JSONSerializable):
    """
    Abstract File is an abstract concept that is used to create requests
    """

    name: str
    argument: json

    def __init__(self, name: str, argument: json):
        self.name, self.argument = name, argument


def get_engine():
    """
    Generate the SQL Alchemy engine for us, using Capo

    :return: SQLAlchemy engine from database URL
    """

    capo = CapoConfig().settings("metadataDatabase")
    url = capo.jdbcUrl.replace("jdbc:", "").replace("://", f"://{capo.jdbcUsername}:{capo.jdbcPassword}@")
    return sa.create_engine(url)


def get_session_factory(engine):
    """
    Generate the SQL Alchemy session factory for us, using the supplied engine

    :param engine:
    :return: SQLAlchemy engine
    """

    factory = sessionmaker(expire_on_commit=False)
    factory.configure(bind=engine)
    return factory


@mapper_registry.mapped
class Controller(JSONSerializable):
    """
    A control system is a system capable of making direct workflow or capability service requests,
    i.e. WS Capability Service, VLASS Manager
    """

    control_system: str
    message_handler: str
    routing_key_prefix: str

    __tablename__ = "processing_controllers"
    control_system = sa.Column(
        "control_system",
        sa.String,
        primary_key=True,
        comment="A system capable of creating and submitting Workspaces Workflow or Capability Requests",
    )
    message_handler = sa.Column(
        "message_handler",
        sa.String,
        comment="which message formatter to use?",
    )
    routing_key_prefix = sa.Column(
        "routing_key_prefix",
        sa.String,
        comment="the prefix of the AMQP routing key for the associated handler",
    )

    # Pyramid support method: must accept a "request" argument that is unused by us
    def __json__(self, request=None) -> dict:
        return {
            "type": self.__class__.__name__,
            "control_system": self.control_system,
            "message_system": self.message_handler,
            "routing_key_prefix": self.routing_key_prefix,
        }


class AutomaticSubmitStatus(Enum):
    RUN = "RUN"
    PAUSE = "PAUSE"
    STOP = "STOP"


AutomaticSubmitStatusType = SaEnum(
    AutomaticSubmitStatus,
    name="auto_process_type",
    values_callable=lambda allowed_list: [status.value for status in allowed_list],
)


@mapper_registry.mapped
class StandardProcessingSettings(JSONSerializable):
    """
    A setting for control of standard processing. Includes Run State and Default Capability
    """

    default_capability: str
    auto_submit_status: AutomaticSubmitStatus

    __tablename__ = "processing_settings"
    default_capability = sa.Column(
        "default_capability",
        sa.String,
        primary_key=True,
        comment="The current default capability to respond to ingestion complete events",
    )
    auto_submit_status = sa.Column(
        "auto_submit_status",
        AutomaticSubmitStatusType,
        comment="What is the current status of automatic processing? RUN, PAUSE, or STOP",
    )

    def update_default_capability(self, new_default: str):
        self.default_capability = new_default

    def update_auto_submit_status(self, new_status: str):
        self.auto_submit_status = AutomaticSubmitStatus[new_status.upper()]

    # Pyramid support method: must accept a "request" argument that is unused by us
    def __json__(self, request=None) -> dict:
        return {
            "type": self.__class__.__name__,
            "default_capability": self.default_capability,
            "auto_submit_status": self.auto_submit_status.value,
        }

    @classmethod
    def from_json(cls, json: dict) -> StandardProcessingSettings:
        """
        Method that creates and instance of StandardProcessingSettings given JSON

        :param json: the description of these standard setting in a dictionary
        :return: StandardProcessingSettings instance
        """
        copy = dict(json)
        del copy["type"]
        copy["auto_submit_status"] = AutomaticSubmitStatus[copy["auto_submit_status"]]
        return StandardProcessingSettings(**copy)
