#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.

"""
Standard calibrations are generated as ingestions complete. There is a Capo property
that controls whether the calibration requests are generated, generated and submitted,
or just ignored. This is based on the earlier CIPL workflow in the archive. During
array configuration changes, DAs suspend the executions but allow requests to be made,
because move configurations are not well-behaved with the standard recipes, so they
require some manual intervention. Additionally, during various failure modes, the DAs
will completely ignore messages coming from ingest.

Since the Workspaces system will be deployed in parallel with the legacy CIPL
workflow, we need a lever to control which one is handling standard calibrations
(or for testing, perhaps both). For this reason we may want to completely ignore
ingestion-complete messages in one system or the other.

CIPL already implemented this functionality with a simple state in Capo. The state can
either be RUN, PAUSE or STOP. RUN (in essence) both creates and submits workflows,
whereas PAUSE creates but does not submit workflows, and STOP ignores the ingestion-
complete event entirely. So we duplicate this functionality here under a different key
so that both systems are configured the same way, to simplify life on the Capo editor.

What follows is an example of the Gang of Four Design Pattern "State." This is just
the state of the standard calibration listener, not to be confused with capability
or capability request states. The RunStates in this module shouldn't ever escape;
they're just an implementation detail in this module. By using this pattern we can
simplify the handling of the event, since we simply look up the state we're in and
ask it to deal with the event somehow.
"""

# pylint: disable=E0401, R0903, W0237, W1203

import abc
import json
import logging
from typing import Dict, Callable

import requests
from sqlalchemy.orm import Session

from aat_wrest.observation_wrester import WrestObservationMetadata
from aat_wrest.utilities import MDDBConnector
from messaging.router import Router, on_message
from pycapo import CapoConfig
from workspaces.system.schema import StandardProcessingSettings, AutomaticSubmitStatus

from workspaces.system.services.interfaces import ArchiveServiceIF

logger = logging.getLogger(__name__)


class StandardProcessingRunState(abc.ABC):
    """
    Represents a run state that the archive service uses to determine
    what to do in response to an ingest-complete message arriving.
    """

    @abc.abstractmethod
    def ingest_complete(self, locator: str, metadata: Dict[str, str]):
        """
        Handle an ingestion-complete message from the archive
        :param locator:  the science product locator of this data
        :param metadata: some metadata about this request
        :return:
        """
        pass


class RunState(StandardProcessingRunState):
    """
    The RUN state: create the request, and also start it executing.
    """

    def __init__(self, manager: "StandardProcessingManager"):
        self.manager = manager

    def ingest_complete(self, spl: str, metadata: Dict[str, str]):
        logger.debug(f"(RUN) Creating and submitting {self.manager.default_processing_capability} request for {spl}")
        request_id = self.manager.create_std_request(spl, metadata)
        self.manager.submit_std_request(request_id)
        logger.info(f"(RUN) Submitted {self.manager.default_processing_capability} request #{request_id} for {spl}")


class PauseState(StandardProcessingRunState):
    """
    The PAUSE state: create the request, but don't immediately start it
    """

    def __init__(self, manager: "StandardProcessingManager"):
        self.manager = manager

    def ingest_complete(self, spl: str, metadata: Dict[str, str]):
        logger.debug(f"(PAUSE) Creating {self.manager.default_processing_capability} request for {spl}")
        request_id = self.manager.create_std_request(spl, metadata)
        logger.info(f"(PAUSE) Created {self.manager.default_processing_capability} request #{request_id} for {spl}")


class StopState(StandardProcessingRunState):
    """
    The STOP state: drop the ingestion-complete message on the floor and do nothing with it.
    """

    def __init__(self, manager: "StandardProcessingManager"):
        self.manager = manager

    def ingest_complete(self, spl: str, metadata: Dict[str, str]):
        # we're not doing anything in response to this event here
        logger.info(f"(STOP) Ignoring ingestion-complete message for {spl}")


class StandardProcessingManager:
    def __init__(self, session_factory: Callable[[], Session]):
        self.session_factory = session_factory
        self.STATES = {
            AutomaticSubmitStatus.RUN: RunState(self),
            AutomaticSubmitStatus.STOP: StopState(self),
            AutomaticSubmitStatus.PAUSE: PauseState(self),
        }
        self.default_settings = self.get_settings()  # set current state of processing settings on load
        logger.info("Instantiating Standard Processing Manager")

    @property
    def default_processing_capability(self):
        return self.default_settings.default_capability

    @property
    def state(self) -> StandardProcessingRunState:
        """
        Return the standard calibration state we're currently in
        :return:
        """
        return self.STATES[self.state_name]

    @property
    def state_name(self) -> AutomaticSubmitStatus:
        return AutomaticSubmitStatus(self.default_settings.auto_submit_status)

    def ingestion_complete(self, spl: str, metadata: Dict[str, str]):
        self.state.ingest_complete(spl, metadata)

    def get_settings(self, session: Session = None) -> StandardProcessingSettings:
        """
        GET current processing settings from database. Session dependent

        :param session: If a session object is provided use it for the query, else we'll make a fresh session (Optional)
        :return: StandardProcessingSettings instance
        """
        current_session = session if session else self.session_factory()
        results = current_session.query(StandardProcessingSettings).first()
        if not session:
            current_session.close()
        return results

    def update_settings(self, setting_dict: Dict[str, str]) -> StandardProcessingSettings:
        """
        Update the requested settings
        :param setting_dict: dictionary object of settings to update
        :return: updated StandardProcessingSettings instance
        """
        session = self.session_factory()
        available_settings = ["default_capability", "auto_submit_status"]
        # new session, pull a new settings object
        current_settings = self.get_settings(session)
        logger.debug(current_settings.__json__())

        for setting in setting_dict:
            if (
                setting == available_settings[0]
                and setting_dict[available_settings[0]] != current_settings.default_capability
            ):
                logger.debug(f"1 updating setting {available_settings[0]} to {setting_dict[setting]}")
                current_settings.update_default_capability(setting_dict[available_settings[0]])
            elif (
                setting == available_settings[1]
                and setting_dict[available_settings[1]] != current_settings.auto_submit_status
            ):
                logger.debug(f"2 updating setting {available_settings[1]} to {setting_dict[setting]}")
                # Let's reinforce the table's enum restriction and try to catch the errors in advance
                if setting_dict[setting] in AutomaticSubmitStatus.__members__:
                    current_settings.update_auto_submit_status(setting_dict[available_settings[1]])
                else:
                    logger.error(
                        f"Cannot update auto_submit_status to {setting_dict[setting]}. Only statuses {list(AutomaticSubmitStatus.__members__)} are allowed."
                    )
            else:
                logger.error(f"Setting {setting} is either unchanged or unknown. Skipping.")

        session.commit()
        self.refresh_defaults()
        logger.info("Updates for Standard Processing Settings have completed.")

        return current_settings

    def refresh_defaults(self):
        """
        Update the default settings objects after settings changes

        :return:
        """
        logger.debug(f"Before refresh: {self.default_processing_capability}")
        self.default_settings = self.get_settings()
        logger.debug(f"After refresh: {self.default_processing_capability}")

        logger.info("Refreshed standard processing default settings")

    def create_std_request(self, spl: str, metadata: Dict[str, str]) -> int:
        """
        Creates a new standard processing request with information given from an ingestion-completed event

        :param spl: Science product locator of observation
        :param metadata: Dict with the following information inside:
            - sdm_id: SDM ID (sometimes called fileset id)
            - bands: Radio bands for observation
            - array_config: Configuration of array telescopes
            - obs_start_time: Start time of observation
            - obs_end_time: End time of observation
        :return: HTTP Response for attempt to create standard processing request
        """
        capability_url = CapoConfig().settings("edu.nrao.workspaces.CapabilitySettings").serviceUrl
        analyst_email = CapoConfig().settings("edu.nrao.workspaces.NotificationSettings").analystEmail
        parameters = {
            "parameters": {
                "product_locator": spl,
                "user_email": analyst_email,
                "metadata": metadata,
            }
        }

        result = requests.post(
            f"{capability_url}/capability/{self.default_processing_capability}/request/create", json=parameters
        )
        result.raise_for_status()
        return result.json()["id"]

    @staticmethod
    def submit_std_request(request_id: int):
        capability_url = CapoConfig().settings("edu.nrao.workspaces.CapabilitySettings").serviceUrl
        result = requests.post(f"{capability_url}/capability/request/{request_id}/version/1/submit")
        result.raise_for_status()


class ArchiveService(ArchiveServiceIF):
    """
    The archive service: clients access this to interface with the archive services
    """

    def __init__(self, session_factory: Callable[[], Session]):
        self.session_factory = session_factory
        self.capability_manager = StandardProcessingManager(self.session_factory)
        self.router = Router("archive")
        self.router.register(self)

    @on_message(message="ingestion complete")
    def on_ingestion_complete(self, **message: Dict) -> tuple[str, str, str, str] | None:
        """
        Accept an Ingestion Complete message from the archive

        :param message: the Ingestion Complete message
        """

        log_data = message["logData"]

        sdm_id = log_data["fileset_id"]
        sb_type = log_data["schedblock_type"]
        telescope = log_data["telescope"]

        if sb_type != "OBSERVER" or telescope != "EVLA":
            # not a candidate for standard processing
            return None

        # log message
        logger.info(f"Archive Ingestion Complete Event: {log_data}")

        if sdm_id.startswith("VLASS") or sdm_id.startswith("TSKY"):
            # Do not submit VLASS or TSKY execution blocks for calibration
            logger.info(f"Caught Ingestion Event was for VLASS or TSKY - ignoring.")
            return None

        return self.create_and_submit_from_sdm_id(log_data)

    def create_and_submit_from_sdm_id(self, log_data: Dict) -> tuple[str, str, str, str]:
        sdm_id = log_data["fileset_id"]
        execblock_id = log_data.get("execblock_id", "")
        project_code = log_data.get("project_code", "")

        wrester = WrestObservationMetadata(MDDBConnector(), sdm_id=sdm_id)
        metadata = json.loads(wrester.wrest_observation_info())
        metadata["sdm_id"] = sdm_id
        spl = metadata["spl"]
        del metadata["spl"]

        self.capability_manager.ingestion_complete(spl, metadata)

        return spl, execblock_id, sdm_id, project_code
