# Copyright (C) 2023 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
#
import logging
import os
import pathlib
import re
import shutil
from pathlib import Path

from pycapo import CapoConfig
from sqlalchemy import insert
from sqlalchemy.exc import SQLAlchemyError
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import Session

from workspaces.capability.schema import (
    Capability,
    CasaMatrixCasaVersion,
    matrix_capabilities,
)
from workspaces.system import helpers
from workspaces.system.services.interfaces import CasaMatrixServiceIF

logger = logging.getLogger(__name__)

Base = declarative_base()

# Defined version number layouts for release:
RELEASE_CASA_REGEX = r"^casa-[0-9\.\-]*(pipeline-)?.*"
DEV_CASA_REGEX = r"^casa-[0-9\.\-]*(pipeline-)?[^tp]"
CASA_VERSION_REGEX = r"[a-zA-Z\-]+([\d]+(\.[\d]+){2}(-[\d]+){0,1}).*"
PIPELINE_VERSION_REGEX = r".*([\d]{4}(\.[\d]+){3}).*"
EL_SUFFIX_REGEX = r"[\.\-]el[0-9]+$"
IGNORE_REGEX = r"\.|\-"
DEFAULT_CV = "unknown"
DEFAULT_PV = "default"


def casa_version_from_path(path: str) -> str:
    """
    Provided a CASA installation standard naming, extract the CASA (and perhaps pipeline) version
    strings, and paste them together validation checking and display.

    :param path: A path to a CASA version
    :return: The version string for both software sets, with defaults
    """

    # Defaults
    casa_version, pipeline_version = DEFAULT_CV, DEFAULT_PV

    cv_search = re.search(CASA_VERSION_REGEX, path)
    if cv_search is None:
        # this is probably a default link, make sure we use the actual version path
        path2 = pathlib.Path(path)
        if path2.is_symlink():
            path = str(path2.readlink())
            cv_search = re.search(CASA_VERSION_REGEX, path)

    pv_search = re.search(PIPELINE_VERSION_REGEX, path)
    if cv_search:
        casa_version = cv_search.group(1)
    if pv_search:
        # Strip el* from the release name (used to be a thing)
        pipeline_version = re.sub(EL_SUFFIX_REGEX, "", pv_search.group(1))

    # Return the combined version information:
    return casa_version + "|" + pipeline_version


def natural_sort(in_list: list, reverse: bool = False) -> list:
    """
    Clever little function to perform natural sorting.
    Example: l = ['RHEL7', 'RHEL8', 'RHEL9', 'RHEL20', 'RHEL11', 'RHEL10', 'RHEL6']
             natural_sort(l) = ['RHEL20', 'RHEL11', 'RHEL10', 'RHEL9', 'RHEL8', 'RHEL7', 'RHEL6']

    :param in_list: List to sort
    :param reverse: True to reverse the output list
    :return: A naturally sorted list
    """
    convert = lambda text: int(text) if text.isdigit() else text.lower()
    alphanum_key = lambda key: [convert(c) for c in re.split("([0-9]+)", key)]
    return sorted(in_list, key=alphanum_key, reverse=reverse)


def versions_match(req_v: str, dir_v: str) -> bool:
    """
    Compares CASA+pipeline version formats stored in the request vs installed.

    :param req_v: Version coming from the request
    :param dir_v: Version coming from a directory path
    :return: True if the CASA+pipeline versions match
    """
    req_cv = req_v.split("|")[0]
    dir_cv = dir_v.split("|")[0]
    req_pv = req_v.split("|")[1] if "|" in req_v else DEFAULT_PV
    dir_pv = dir_v.split("|")[1] if "|" in dir_v else DEFAULT_PV

    # Ignore comparing '.' and '-' characters because they like to be used interchangeably
    req_cv = re.sub(IGNORE_REGEX, "", req_cv)
    dir_cv = re.sub(IGNORE_REGEX, "", dir_cv)

    # Ignore comparing pipeline versions if they don't exist in the installed version's name
    return req_cv == dir_cv and (dir_pv == DEFAULT_PV or req_pv == dir_pv)


class CasaMatrixService(CasaMatrixServiceIF):
    """
    Service that enables clients to access the CASA Version Matrix.
    """

    def __init__(self, session: Session):
        self.session = session

        # Used for linking and looking up linked CASA versions
        self.env_dir = str(CapoConfig().profile).replace("dsoc-", "")

        # A valid version will cross-check installed versions in the casa_root against allowed versions in the database
        self.installed_versions = []
        self.allowed_versions = []

    def find_installed_versions(self) -> list[dict[str, str]]:
        """
        Makes a list of all CASA versions found within the casa_root for the current profile.
        Each 'version' will be a dict containing the version (casa_version | pipeline_version) and path.
        The list is first organized by OS versions and then flattened to only use old OSes as fallbacks.

        :return: A flattened list of the installed CASA versions + paths
        """
        flat_versions = []

        # casa_root is the directory that contains all installed versions of CASA for all OSes
        casa_root = CapoConfig().get("edu.nrao.archive.archiveIface.casa_root")
        try:
            os.listdir(casa_root)
        except FileNotFoundError:
            raise Exception("Failed to locate CASA root")

        if CapoConfig().get("edu.nrao.archive.archiveIface.allowDevCasaVersions").lower() == "true":
            prefix_match = RELEASE_CASA_REGEX
        else:
            prefix_match = DEV_CASA_REGEX

        # Start by building a list of version/path dictionaries, extracting the version name from the directory
        # names in our CASA root. The assumption here is that all these CASA installs are also pipeline installs.
        # The sorting here will place the newest OS version at the front.
        os_dirs = natural_sort(os.listdir(casa_root), reverse=True)

        all_versions = {}
        for os_dir in os_dirs:
            # casa_root / RHEL*
            os_dir = Path(casa_root) / os_dir
            if os_dir.is_dir():
                # casa_root / RHEL* / env
                os_env_dir = os_dir / self.env_dir
                if os_env_dir.exists() and os_env_dir.is_dir():
                    possible_versions = sorted(os.listdir(os_env_dir), reverse=True)
                    # Make a list of dicts containing CASA versions + paths for all the installed versions in here
                    all_versions[os_dir] = [
                        dict(version=casa_version_from_path(v), path=str(os_env_dir / v))
                        for v in possible_versions
                        if ((os_env_dir / v).is_dir() or (os_env_dir / v).is_symlink()) and re.match(prefix_match, v)
                    ]

        # Flatten the versions such that versions for the most recent OS will be preferred but will fall back to other
        # OSes if necessary
        for os_versions in all_versions.values():
            if len(flat_versions) == 0:
                flat_versions = os_versions
            else:
                for v in os_versions:
                    # Check if this version has already been added for a more recent OS
                    if v not in flat_versions and not any(v["version"] in fv["version"] for fv in flat_versions):
                        flat_versions.append(v)

        if not flat_versions:
            logger.critical(f"Did not find any links to installed CASA versions in {casa_root}")

        # Return the version list sorted with newer versions first
        return sorted(flat_versions, key=lambda v: v["version"], reverse=True)

    def fetch_allowed_versions(self, capability: str) -> list[str]:
        """
        Makes a lists of all allowed CASA versions for the given capability based on the casa_matrix tables.

        :param capability: (Optional) The capability to filter CASA matrix versions
        :return: A list of CASA versions that can be used for processing
        """
        versions = []

        result = (
            self.session.query(CasaMatrixCasaVersion.casa_version)
            .filter_by(is_cluster_compatible=True)
            .join(matrix_capabilities)
            .filter_by(capability_name=capability)
            .all()
        )
        for v in result:
            versions.append(v[0])

        if not versions:
            logger.critical(f"Did not find allowable CASA versions in the matrix tables for {capability}")

        return versions

    def get_installed_version(self, version: str) -> dict[str, str]:
        """
        Gets the installed version string + path for a given version string.

        :param version: CASA + pipeline version as a string
        :return: Dict containing the version string and installation path
        """
        ret = {}

        for v in self.installed_versions:
            if v["version"] == version:
                ret = v

        return ret

    def get_version(
        self, version: str | None = None, capability: str | None = None, telescope: str | None = None
    ) -> dict[str, str]:
        """
        Returns the CASA version that is valid for processing.
        If a version is provided and is valid, it will simply be returned.
        If the given version is invalid or no version is given, the default version is pulled from a CAPO using the
            capability + telescope.

        :param version: (Optional) A potential CASA version to validate
        :param capability: (Optional) The capability to distinguish a fallback value from CAPO
        :param telescope: (Optional) The telescope to distinguish a fallback value from CAPO
        :return: A dict mapping a CASA version to its path
        """
        versions = self.get_versions(version, capability, telescope)
        return versions[0] if versions else {}

    def get_versions(
        self, version: str | None = None, capability: str | None = None, telescope: str | None = None
    ) -> list[dict[str, str]]:
        """
        Returns the CASA versions that are valid for processing as a dict.
        The recommended version (either the given version or the default for the given capability) is first in the list.

        :param version: (Optional) A potential CASA version to use as the default
        :param capability: (Optional) The capability to filter versions on and to determine the recommended version
        :param telescope: (Optional) The telescope to distinguish a recommended version from CAPO
        :return: A dict mapping valid CASA versions to their paths
        """
        versions = []

        # Installed versions are a sorted, flattened list of dicts containing version names and paths
        self.installed_versions = self.find_installed_versions()

        # Allowed versions are a dict of capabilities mapping to lists of version names
        self.allowed_versions = self.fetch_allowed_versions(capability)

        # Add any versions that are both allowed and installed to the list, keeping it sorted by version names
        for installed_version in self.installed_versions:
            # Strip el* from the release name (used to be a thing)
            if re.sub(EL_SUFFIX_REGEX, "", os.path.basename(installed_version["path"])) in self.allowed_versions:
                versions.append(installed_version)

        # Put the requested version as first in the list if it's valid
        # Also, limit the list to requested version and versions newer than it
        if version:
            for index_of_version, v in enumerate(versions):
                if versions_match(version, v["version"]):
                    versions = [versions[index_of_version]] + versions[:index_of_version]
                    break

        # If the requested version is invalid, put the default as first in the list
        # If the default is invalid, the newest version will be first in the list
        if versions and (not version or not versions_match(version, versions[0]["version"])):
            default = self.get_default_version(capability, telescope)
            if default:
                for v in versions:
                    if versions_match(default, v["version"]):
                        versions.insert(0, versions.pop(versions.index(v)))
                        break

        logger.info(f"Found CASA versions for {capability}: {versions}")

        return versions

    def get_default_version(self, capability: str | None, telescope: str | None) -> str | None:
        """
        Gets the default CASA version from CAPO for the given capability (the std_calibration capability is default).

        :param capability: (Optional) The capability to get the default
        :param telescope: (Optional) The telescope to get the default (used for some capabilities)
        :return: The full path to the default CASA version as a string
        """
        if not capability:
            return None

        result = self.session.query(Capability.default_casa_capo_key).filter_by(name=capability).first()
        keys = result[0].split("|") if result else None

        if not keys:
            return None

        # Some capabilities have different paths per telescope (looking at you ALMA)
        if len(keys) > 1 and telescope and telescope.lower() == "alma":
            key = keys[1]
        else:
            key = keys[0]

        # The CAPO setting is a full path to a CASA installation (this assumes it is installed)
        casa_settings = CapoConfig().settings("edu.nrao.archive.workflow.config.CasaVersions")
        path = casa_settings[key]

        return casa_version_from_path(path)

    def get_recipe(
        self,
        capability: str,
        version: str | None = None,
        telescope: str | None = None,
        version_path: str | None = None,
        params_to_add: dict = None,
    ) -> dict[str, str, str]:
        """
        Returns recipe defined for the given capability stored in the given (or default) CASA version.
        To override any version checking, pass an absolute path to a valid CASA version.

        :param capability: The capability to look up the recipe's file name
        :param version: (Optional) The desired CASA version to retrieve the recipe contents
        :param telescope: (Optional) The telescope to decide between multiple recipes
        :param version_path: (Optional) An absolute path to a CASA version which will override any version checking
        :param params_to_add: (Optional) Dictionary pairing Command names to a set of parameter names and values
        :return: Dict containing the CASA recipe contents as a string along with the CASA version + path
        """
        ret = {}
        recipe_name = None
        result = self.session.query(Capability.casa_recipe).filter_by(name=capability).first()
        if result:
            # ALMA recipe names are stored in the DB after a pipe character
            recipe_names = result[0].split("|")
            if len(recipe_names) > 1 and telescope and telescope.lower() == "alma":
                recipe_name = recipe_names[1]
            else:
                recipe_name = recipe_names[0]
        else:
            logger.critical(f"Did not find recipe file name in the matrix tables for {capability}")

        if recipe_name:
            # Only look for a version + path if we weren't given an absolute path
            if version_path:
                ret = dict(version=casa_version_from_path(version_path), path=version_path)
            else:
                ret = self.get_version(version, capability, telescope)

            # Full recipe is stored at <casa_path>/pipeline/pipeline/recipes/<recipe_name>
            recipe_path = Path(ret.get("path")) / "pipeline" / "pipeline" / "recipes" / recipe_name

            if recipe_path.is_file():
                ret["recipe"] = recipe_path.read_text()
            else:
                ret["recipe"] = ""
                logger.critical(f"Did not find recipe file in the CASA version at {recipe_path}")

        # If we have a recipe and there's command parameters that need to be added, add them
        if ret["recipe"] and params_to_add:
            ret["recipe"] = helpers.add_ppr_parameter_tags(ret["recipe"], params_to_add)

        return ret

    def save_entity(self, entity: Base):
        """
        Save entity to database.

        :param entity: Entity to persist
        """
        self.session.add(entity)
        self.session.flush()

    def save_capabilities(self, version: str, capabilities: list[str]):
        """
        Saves the given version + capabilities to the casa_matrix_capabilities table.

        :param version: CASA matrix version to look up
        :param capabilities: List of capability names to associate to the given version
        """
        matrix_version = self.session.query(CasaMatrixCasaVersion).filter_by(casa_version=version).first()
        mc = [{"matrix_id": matrix_version.matrix_id, "capability_name": c} for c in capabilities]

        self.session.execute(insert(matrix_capabilities), mc)
        self.session.flush()

    def add_version(self, version: str, capabilities: list[str] = None, is_cluster_compatible: bool = True) -> bool:
        """
        Adds a new CASA version to the matrix using the provided version string and capability name list.
        By default, the new version will be flagged as being cluster compatible.

        :return: True if added successfully
        """
        matrix_version = CasaMatrixCasaVersion(
            casa_version=version,
            is_cluster_compatible=is_cluster_compatible,
        )

        try:
            self.save_entity(matrix_version)
            if capabilities:
                # Link the given capabilities to the new matrix version, wasn't sure how to do this in one go
                self.save_capabilities(version, capabilities)
            return True
        except SQLAlchemyError as e:
            logger.critical(f"Unable to add CASA version {version}: %s", e)
            return False

    def update_version(self, version: str, capabilities: list[str] = None, is_cluster_compatible: bool = True) -> bool:
        """
        Updates a CASA version in the matrix.

        :return: True if updated successfully
        """
        try:
            self.session.query(CasaMatrixCasaVersion).filter_by(casa_version=version).update(
                {
                    "casa_version": version,
                    "is_cluster_compatible": is_cluster_compatible,
                }
            )
            self.session.flush()
            if capabilities:
                # Link the given capabilities to the new matrix version, wasn't sure how to do this in one go
                self.save_capabilities(version, capabilities)
            return True
        except SQLAlchemyError as e:
            logger.critical(f"Unable to update CASA version {version} %s", e)
            return False

    def delete_version(self, version: str) -> bool:
        """
        Deletes a CASA version and its capability links from the matrix.

        :return: True if deleted successfully
        """
        try:
            matrix_version = self.session.query(CasaMatrixCasaVersion).filter_by(casa_version=version).first()
            matrix_version.capabilities = []
            self.session.delete(matrix_version)
            self.session.flush()
            return True
        except SQLAlchemyError as e:
            logger.critical(f"Unable to delete CASA version {version} %s", e)
            return False

    def make_links(self, dry_run: bool = False) -> dict[str, list[str]]:
        """
        Refreshes the list of symlinks in the casa_root pointing to installed CASA + pipeline versions.
        These links exist to not step on the toes of the legacy archive and NAASC or DSOC ops.
        The dest structure follows: $casa_root/$OS/$PROFILE/$CASA (e.g. $casa_root/RHEL7/dsoc-dev/casa-release-5.1.2-4).
        The links are first created in a temporary directory which then replaces the existing one, just in case.
        A dict mapping OS versions to CASA versions is returned for logging/testing.

        :param dry_run: Setting to True will not create the new directory + symlinks nor remove the old one
        :return: Dict containing lists of found versions per OS
        """
        # For logging
        found_versions = {}

        # The directory that contains all installed versions of CASA for all OSes
        casa_root = Path(CapoConfig().get("edu.nrao.archive.archiveIface.casa_root"))
        # The directory where we're generating links to casa_root
        casa_delivery_dir = Path(CapoConfig().get("edu.nrao.archive.workflow.config.CasaVersions"))

        # We're going to add symlinks to a new dir and then replace the old one (in case something goes awry)
        temp_dir_name = self.env_dir + "-new"

        # Start with the RHEL* directories in the CASA delivery area
        src_os_dirs = sorted([x for x in casa_delivery_dir.iterdir() if "RHEL" in x.name], reverse=True)
        for src_os_dir in src_os_dirs:
            found_versions[src_os_dir.name] = []

            # casa_root / RHEL* / env
            dest_os_dir = casa_root / src_os_dir.name / temp_dir_name
            if not dry_run and not dest_os_dir.exists():
                dest_os_dir.mkdir(parents=True, exist_ok=True)

            # casa_delivery_dir / RHEL* / release
            src_release_dir = src_os_dir / "release"

            if src_release_dir.is_dir():
                for src_casa_path in src_release_dir.iterdir():
                    if src_casa_path.is_dir():
                        # Only link releases with a pipeline
                        if "pipeline" in [f.name for f in src_casa_path.iterdir()]:
                            found_versions[src_os_dir.name].append(src_casa_path.name)
                            if not dry_run:
                                dest_path = dest_os_dir / src_casa_path.name
                                dest_path.symlink_to(src_casa_path)

            logger.info(f"Found CASA versions: {found_versions[src_os_dir.name]} in {src_os_dir}")
            if not dry_run:
                # Move all new symlinks into the old directory, overwriting any that existed by the same name
                # This does not delete any pre-existing links in order to preserve the 'default' link
                final_dest = casa_root / src_os_dir.name / self.env_dir
                logger.info(f"Replacing CASA versions in {final_dest}")
                if not final_dest.exists():
                    final_dest.mkdir(parents=True, exist_ok=True)
                for link in dest_os_dir.iterdir():
                    link.replace(final_dest / link.name)
                # Delete the now empty temp dir
                shutil.rmtree(dest_os_dir)

        return found_versions
