#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
import logging

from immutable_views import DictView, ListView

from workspaces.workflow.schema import WorkflowRequest

logger = logging.getLogger(__name__)

"""
Class which creates immutable dictionary and list views containing capo properties required by remote executing workflows.
When a workflow is confirmed as able to run remotely, this should be updated if necessary
"""
wf_properties_list = DictView(
    {
        "general": [
            "edu.nrao.workspaces.WorkflowSettings.serviceUrl",
            "edu.nrao.workspaces.NotificationSettings.serviceUrl",
            "edu.nrao.workspaces.UISettings.serviceUrl",
        ],
        "delivery_requirements": [
            "edu.nrao.workspaces.DeliverySettings.downloadDirectory",
            "edu.nrao.workspaces.DeliverySettings.downloadUrl",
        ],
        "fetcher_requirements": [
            "edu.nrao.workspaces.ProductFetcherSettings.locatorServiceUrlPrefix",
            "edu.nrao.workspaces.ProductFetcherSettings.defaultThreadsPerHost",
            "edu.nrao.workspaces.ProductFetcherSettings.executionSite",
            "edu.nrao.workspaces.ProductFetcherSettings.metadataServiceUrlPrefix",
            "almaMetadataDatabase.jdbcUsername",
            "almaMetadataDatabase.jdbcPassword",
            "almaMetadataDatabase.jdbcUrl",
        ],
        "casa_requirements": [
            "edu.nrao.workspaces.ProcessingSettings.useCasa",
            "edu.nrao.workspaces.ProcessingSettings.CasaVersion.vlass",
            "edu.nrao.archive.workflow.config.CasaVersions.homeForReprocessing",
        ],
        "carta_requirements": [
            "edu.nrao.workspaces.carta.useCarta",
            "edu.nrao.carta.timeoutInMinutes",
            "edu.nrao.carta.pathToExecutable",
            "edu.nrao.carta.redisHost",
            "edu.nrao.carta.redisPort",
            "edu.nrao.carta.redisPassword",
            "edu.nrao.carta.reverseProxyHost",
        ],
        "processing_requirements": [
            "edu.nrao.workspaces.ProcessingSettings.rootDirectory",
            "edu.nrao.workspaces.ProcessingSettings.scriptLocation",
        ],
    }
)


class ViewWorkflowCapoParameters:
    def __init__(self, wf_request: WorkflowRequest, is_remote: bool):
        self.request = wf_request
        self.needs_casa = wf_request.workflow.uses_casa
        self.is_remote = is_remote

    @staticmethod
    def general_parameters() -> list:
        logger.info("Gathering general properties...")
        return wf_properties_list["general"].copy()

    @staticmethod
    def delivery_parameters():
        logger.info("Gathering ALMA Download parameters...")
        return wf_properties_list["delivery_requirements"].copy()

    @staticmethod
    def casa_parameters() -> list:
        logger.info("Gathering CASA properties...")
        return wf_properties_list["casa_requirements"].copy()

    @staticmethod
    def carta_parameters():
        logger.info("Gathering CARTA parameters...")
        return wf_properties_list["carta_requirements"].copy()

    @staticmethod
    def fetcher_parameters():
        logger.info("Gathering Product Fetcher parameters...")
        return wf_properties_list["fetcher_requirements"].copy()

    @staticmethod
    def processing_parameters():
        logger.info("Gathering Processing parameters...")
        return wf_properties_list["processing_requirements"].copy()

    def construct_view(self) -> ListView:
        """
        Builds a ListView of necessary parameters based on the workflow.
        :return: ListView of parameters
        """
        wf = self.request.workflow_name
        params = self.general_parameters()
        if self.needs_casa:
            params += self.casa_parameters()
        if self.is_remote:
            if "download" in wf:
                params += self.fetcher_parameters() + self.delivery_parameters()
            if "carta" in wf:
                params += self.fetcher_parameters() + self.carta_parameters()
            if "restore" in wf:
                params += self.fetcher_parameters() + self.delivery_parameters() + self.processing_parameters()

        return ListView(params)
