#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
""" Builds AMQP messages based on templates """

import logging

# pylint: disable=E0401

from immutable_views import DictView

from workspaces.shared_interfaces import MessageArchitectIF

# pylint: disable=E0401
from workspaces.workflow.schema import WorkflowRequest

logger = logging.getLogger(__name__)

"""
Class which create an immutable dictionary of amqp message templates for easy access
"""

# basic templates
service_templates = DictView(
    {
        "ws": {
            "service": None,
            "routing_key": None,
            "subject": None,
            "type": None,
        },
        "archive": {
            "service": "archive",
            "routing_key": None,
            "subject": None,
        },
        "vlass": {
            "service": "archive",
            "routing_key": None,
            "subject": None,
            "application": "workflow",
            "message": None,
            "status": None,
        },
    }
)

# messages that require additional information
workflow_msg_templates = DictView(
    {
        "delivery": {
            "delivery": None,
        },
        "carta_ready": {
            "carta_url": None,
        },
        "update_wf_metadata": {
            "payload": None,
        },
    }
)

# archive messages that require additional information
archive_msg_templates = DictView(
    {
        "carta_ready": {
            "application": "workflow",
            "carta_url": None,
            "message": "carta ready",
        },
        "carta_failed": {
            "application": "workflow",
            "carta_url": None,
            "message": "carta failed",
        },
        "do_not_calibrate": {
            "application": "qa-script",
            "logData": {
                "status": "FAIL",
                "type": "calibration",
                "fileset_id": None,
                "markedDirectory": None,
            },
        },
        "cal_ingestion_complete": {
            "application": "ingestor",
            "message": "Calibration Ingestion Complete",
            "request": "calibration",
            "logData": {
                "ingestion_type": "evla_cal",
                "telescope": "EVLA",
                "fileset_id": None,
                "project_code": None,
                "status": "complete",
                "override_db_screen": "false",
            },
        },
        "img_ingestion_complete": {
            "application": "ingestor",
            "message": "Image Ingestion Complete",
            "request": "images",
            "logData": {
                "sourceDirectory": None,
                "project_code": None,
                "ingestion_type": "vlass_quicklook",
                "status": "complete",
            },
        },
        "obs_ingestion_complete": {
            "application": "ingestor",
            "message": "Observation Ingestion Complete",
            "request": "observation",
            "logData": {
                "project_code": None,
                "telescope": None,
                "ingestion_type": None,
                "status": "complete",
            },
        },
        "download_complete": {
            "application": "workflow",
            "download_url": None,
            "delivery_path": None,
            "token": None,
            "message": "download complete",
        },
        "download_failed": {
            "application": "workflow",
            "message": "download failed",
        },
        "restore_cms_complete": {
            "application": "workflow",
            "message": "restore_cms complete",
            "delivery_url": None,
            "delivery_path": None,
            "token": None,
        },
        "restore_cms_failed": {
            "application": "workflow",
            "message": "restore_cms failed",
        },
        "update": {
            "application": "workflow",
            "message": "workflow updated",
            "stage": None,
            "start": None,
            "stop": None,
        },
    }
)


class WorkflowMessageArchitect(MessageArchitectIF):
    """Builds workflow AMQP messages"""

    def __init__(
        self,
        request: WorkflowRequest = None,
        previous_info=None,
        delivery_info=None,
        carta_url=None,
        ui_info=None,
        json_payload=None,
    ):
        """
        Initialize Architect

        :param request: The workflow request this message concerns
        :param previous_info: The preceding workflow request if a follow-on (Optional)
        :param delivery_info: The delivery location (Optional)
        :param carta_url: The CARTA URL (Optional)
        :param ui_info: Information received from the UI (Optional)
        :param json_payload: Information to be sent as JSON (Optional)
        """
        self.subject = request or previous_info
        self.delivery = delivery_info
        self.carta_url = carta_url
        self.ui_info = ui_info
        self.json_payload = json_payload

    @staticmethod
    def get_message_template(msg_type: str) -> dict:
        """
        Construct the message template based on type

        :param msg_type: type of message to construct
        :return: full message template dictionary for rendering
        """

        if msg_type in workflow_msg_templates:
            # message needs to send extra information
            return {**service_templates["ws"], **workflow_msg_templates[msg_type]}
        else:
            # default message is sufficient
            return service_templates["ws"]

    def compose_message(self, msg_type: str) -> DictView:
        """
        Render message template for requested type

        :param msg_type: the type of message to send
        :return: the rendered message ready for sending
        """
        if any(item in msg_type for item in ["qa_", "carta", "stage"]):
            # Send messages that the CapabilityService wants to handle to the capability topic
            service = "capability"
        else:
            service = "workflow"

        template = self.get_message_template(msg_type)
        template = template.copy()

        template["service"] = service
        template["routing_key"] = service
        template["subject"] = self.subject
        template["type"] = msg_type.replace("_", "-")

        if self.delivery is not None:
            template["delivery"] = self.delivery

        if self.carta_url is not None:
            template["carta_url"] = self.carta_url

        if self.ui_info is not None:
            template["capability_version"] = self.ui_info

        if self.json_payload is not None:
            template["payload"] = self.json_payload

        return DictView(template)


class ArchiveMessageArchitect(MessageArchitectIF):
    """Builds archive system AMQP messages"""

    def __init__(
        self,
        routing_key: str,
        request: WorkflowRequest = None,
        carta_url: str = None,
        sdm_id: str = None,
        project_code: str = None,
        delivery_info: dict = None,
        stage_info: dict = None,
        telescope: str = None,
    ):
        """
        Initialize Architect

        :param routing_key: AMQP routing key to use
        :param request: The workflow request this message concerns
        :param carta_url: The CARTA URL (Optional)
        :param sdm_id: The observation external name (Optional)
        :param project_code: The observation project code (Optional)
        """
        self.key = routing_key
        self.request = request
        self.carta_url = carta_url
        self.sdm_id = sdm_id
        self.project = project_code
        self.delivery = delivery_info
        self.stage_info = stage_info
        self.telescope = telescope

    @staticmethod
    def get_message_template(msg_type: str) -> dict:
        """
        Construct the message template based on type

        :param msg_type: type of message to construct
        :return: full message template dictionary for rendering
        """

        if "update" in msg_type:
            return {**service_templates["archive"], **archive_msg_templates["update"]}
        elif "vlass" in msg_type:
            # all vlass messages have the same format
            return service_templates["vlass"]
        else:
            return {**service_templates["archive"], **archive_msg_templates[msg_type]}

    def compose_message(self, msg_type: str) -> DictView:
        """
        Render message template for requested type

        :param msg_type: the type of message to send
        :return: the rendered message ready for sending
        """
        template = self.get_message_template(msg_type)
        template = template.copy()
        # not using full request for AAT/VLASS messages, increased size caused VLASS message handling issues
        # image indexing is based on project not request id, use different subject
        template["subject"] = (
            f"WS workflow triggering indexer for {self.project}"
            if self.key in ["ingestor.images", "ingestion-complete.vlba", "ingestion-complete.metadata"]
            else f"WS workflow request #{self.request.workflow_request_id}"
        )
        template["routing_key"] = self.key

        if "vlass" in msg_type and "update" not in msg_type:
            # determine message fields from msg_type
            template["message"] = msg_type.replace("_", " ").split(" ", 1)[1]
            template["status"] = msg_type.split("_")[-1]

        if self.carta_url is not None:
            template["carta_url"] = self.carta_url

        if self.sdm_id is not None:
            template["logData"]["fileset_id"] = self.sdm_id
            if msg_type == "do_not_calibrate":
                template["logData"]["markedDirectory"] = self.request.results_dir

        if self.project is not None:
            template["logData"]["project_code"] = self.project

        if self.delivery is not None:
            template["download_url"] = self.delivery["url"]
            template["delivery_path"] = self.delivery["delivered_to"]
            template["token"] = self.delivery["token"]

        if self.stage_info is not None and "update" in msg_type:
            template["stage"] = self.stage_info["stage"]
            template["start"] = self.stage_info["start"]
            template["stop"] = self.stage_info["stop"]

        if self.telescope is not None:
            template["logData"]["telescope"] = self.telescope
            template["logData"]["ingestion_type"] = self._determine_indexer_type(self.telescope)

        return DictView(template)

    @staticmethod
    def _determine_indexer_type(telescope: str) -> str:
        """
        What type of indexing do we need?

        :param telescope: the ingested/curated product's telescope
        :return: the indexer type to use
        """
        if telescope.upper() == "VLBA" or telescope.upper() == "GMVA":
            return "vlba"
        if telescope.upper() == "EVLA" or telescope.upper() == "ALMA":
            return f"{telescope.lower()}_sdm"
        else:
            logger.error(f"Error: Unknown telescope {telescope}, can't send index message to AAT.")
