#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
import datetime
import logging
import re
from typing import Dict, List, Union

from pycapo import CapoConfig

from workspaces.system.schema import Controller
from workspaces.workflow.enum import WorkflowRequestState
from workspaces.workflow.schema import (
    ProcessingSite,
    Workflow,
    WorkflowProgress,
    WorkflowRequest,
    WorkflowRequestFile,
    WorkflowTemplate,
)
from workspaces.workflow.services.interfaces import WorkflowInfoIF

logger = logging.getLogger(__name__)


class WorkflowInfo(WorkflowInfoIF):
    """
    Implements WorkflowInfo, backed by a relational database, using SQL Alchemy.
    """

    def __init__(self, session):
        self.session = session

    def lookup_workflow_definition(self, name: str) -> Workflow:
        return self.session.query(Workflow).get(name)

    def lookup_workflow_request(self, request_id: int) -> WorkflowRequest:
        """
        Queries the workflow_requests table for a specific request

        :param request_id: the request to find
        :return: WorkflowRequest object
        """
        return self.session.query(WorkflowRequest).filter_by(workflow_request_id=request_id).first()

    def lookup_workflow_templates_for(self, workflow_name: str) -> List[WorkflowTemplate]:
        """
        Queries the workflow_templates table for all templates belonging to a given workflow

        :param workflow_name: Name of workflow to lookup templates for
        :return: List of templates
        """
        return self.session.query(WorkflowTemplate).filter_by(workflow_name=workflow_name).all()

    def lookup_workflow_stage(self, request_id: int, stage: str) -> Union[WorkflowProgress, None]:
        """
        Queries the workflow_progress table for stage entries belonging to the specified request with the
        specified stage name

        :param request_id: the request of interest
        :param stage: the stage name to find
        :return:
        """
        return (
            self.session.query(WorkflowProgress)
            .filter_by(workflow_request_id=request_id)
            .filter_by(stage_name=stage)
            .one_or_none()
        )

    def lookup_stale_requests(self, keep_days: int) -> List[WorkflowRequest]:
        """
        Queries the workflow_requests table for all requests that have not been
        cleaned up, and are older that the number of days specified by keep_days

        :param keep_days: the number of days to keep directories
        :return: a list of strings representing processing directories
        """

        return (
            self.session.query(WorkflowRequest)
            .filter((WorkflowRequest.updated_at + datetime.timedelta(days=keep_days)) < datetime.datetime.now())
            .filter_by(htcondor_job_id=None)
            .filter_by(cleaned=False)
            .all()
        )

    def lookup_workflow_request_by_directory(self, results_dir: str) -> WorkflowRequest:
        """
        Queries the workflow_requests table for the request with the matching results directory. Note that this locks
        the request until the session is flushed and will wait for previous locks to flush before selecting the request.

        :param results_dir: the results directory to browse
        :return: WorkflowRequest
        """
        return self.session.query(WorkflowRequest).filter_by(results_dir=results_dir).with_for_update().first()

    def lookup_processing_site(self, processing_site: str) -> ProcessingSite:
        """
        Queries the processing_sites table for the site provided

        :param processing_site: the site identifier (NMT, NAASC)
        :return:
        """
        return self.session.query(ProcessingSite).get(processing_site)

    def all_processing_sites(self) -> List[ProcessingSite]:
        """
        Queries the processing_sites table for all available processing locations and their attributes

        :return: List of processing site objects
        """
        return self.session.query(ProcessingSite).all()

    def lookup_controller(self, controller: str) -> Controller:
        """
        Queries the processing_controllers table for the site provided

        :param controller: the controller identifier (VLASS, AAT)
        :return:
        """
        return self.session.query(Controller).get(controller)

    def all_controllers(self) -> List[Controller]:
        """
        Queries the processing_controllers table for all available control systems and their attributes

        :return: List of controller objects
        """
        return self.session.query(Controller).all()

    def all_external_controllers(self) -> List[Controller]:
        """
        Queries the processing_controllers table for all non-workspaces control systmes and their attributes

        :return: List of external controller objects
        """
        return self.session.query(Controller).filter(Controller.control_system != "WS").all()

    def all_workflows(self) -> List[Workflow]:
        return self.session.query(Workflow).all()

    def all_workflow_requests(self) -> List[WorkflowRequest]:
        """
        Query for the list of all workflow requests

        :return: List of workflow requests
        """
        return self.session.query(WorkflowRequest).all()

    def create_workflow_request(self, workflow: Union[str, Workflow], argument: Dict) -> WorkflowRequest:
        """
        Create new workflow request and save to database

        :param workflow: workflow to run
        :param argument: workflow arguments
        :return: new WorkflowRequest
        """

        # if it's an object with a workflow_name, then that's what we use,
        # otherwise, the argument must be already be a string
        workflow_name = workflow.workflow_name if hasattr(workflow, "workflow_name") else workflow

        need_to_set_args = any(
            item in workflow_name for item in ["null", "carta", "vlass", "qa", "split", "ingest", "import"]
        )

        if need_to_set_args:
            argument["need_project_metadata"] = False

            try:
                product_locators = argument["product_locator"].split(",")
            except (KeyError, AttributeError):
                # KeyError for it not being one of the passed arguments
                # AttributeError for it being passed as a parameter but being a non-string
                #     or None value. This can happen when DAs view something in CARTA before
                #     it's been ingested, as the product locator doesn't yet exist in that case.
                product_locators = []

            # ALFALFA cubes use this type as well as VLASS ones, so ignore the telescope field
            is_vlass_coarse_cube = len(product_locators) == 1 and re.match(
                "uid://.+/vlass_coarse_cube/.+", product_locators[0]
            )
            if re.match("carta", workflow_name):
                if "parent_wf_request_id" in argument:
                    parent = self.lookup_workflow_request(argument["parent_wf_request_id"])
                    argument["source_dir"] = parent.results_dir

                if argument["need_data"] is True and len(product_locators) == 1:
                    argument["need_project_metadata"] = True
                    if not is_vlass_coarse_cube:
                        # If it's a VLASS coarse cube and we this argument to True, it won't work correctly.
                        argument["single_image"] = True

            if re.match("ingest_(?!.*obs$)", workflow_name):
                # Workflow name starts with ingest, but doesn't end with obs
                argument["need_project_metadata"] = True
            if re.search("(vlass|pims)", workflow_name):
                # re.search in case this matches ingest_vlass or vlass_*
                profile = CapoConfig().profile
                logger.info(profile)
                if profile == "docker":
                    # running locally
                    argument["vlass_env"] = "docker"
                else:
                    # determine vlass location from current profile
                    system = profile.split("-")[1]
                    argument["vlass_env"] = system
                if workflow_name == "vlass_coarse":
                    argument["need_project_metadata"] = True
        else:
            argument["need_project_metadata"] = True

        request = WorkflowRequest(
            workflow_name=workflow_name,
            argument=argument,
            state=WorkflowRequestState.Created.name,
            results_dir="" if "data_location" not in argument else argument["data_location"],
            controller="WS" if "controller" not in argument else argument["controller"],
        )
        self.save_request(request)
        logger.info(f"Created Workflow Request {request}")
        return request

    def save_request(self, request: WorkflowRequest):
        """
        Save a given request and return an integer identifier for it

        :param request: request to save
        """
        self.session.add(request)
        self.session.flush()

    def save_file(self, request: WorkflowRequest, filename: str, content: bytes) -> WorkflowRequestFile:
        """
        Save a given file for the specified request to the database

        :param request: request to update with new file
        :param filename: filename of new file
        :param content: contents of new file in bytes
        :return: the request file
        """
        wrf = WorkflowRequestFile(
            workflow_request_id=request.workflow_request_id,
            filename=filename,
            content=content,
        )
        self.session.add(wrf)
        self.session.flush()
        return wrf

    def save_progress(self, request: WorkflowRequest, stage: str, start: str = None, end: str = None):
        """
        Create/Update and Save a progress update for the specified request

        :param request: request to update progress for
        :param stage: the stage to update
        :param start: the start time to update
        :param end: the end time to update
        :return:
        """
        stage_info = self.lookup_workflow_stage(request.workflow_request_id, stage)

        if start and not stage_info:
            progress = WorkflowProgress(
                workflow_request_id=request.workflow_request_id,
                stage_name=stage,
                start=datetime.datetime.strptime(start, "%Y-%m-%dT%H:%M:%S%z"),
                end=None,
            )
        else:
            progress = stage_info
            if start is not None:
                progress.start = datetime.datetime.strptime(start, "%Y-%m-%dT%H:%M:%S%z")
            if end is not None:
                progress.end = datetime.datetime.strptime(end, "%Y-%m-%dT%H:%M:%S%z")

        self.session.add(progress)
        self.session.flush()
        return progress

    def update_request_cleaned(self, request: WorkflowRequest, num_cleaned: int):
        """
        Update an existing workflow request when some of its results directories are annihilated.

        :param request: The workflow request to update
        :param num_cleaned: The number of areas cleaned to increment the total by
        :return:
        """
        request.update_num_cleaned(request.num_cleaned + num_cleaned)
        self.save_request(request)

    def refresh_request(self, request: WorkflowRequest) -> WorkflowRequest:
        """
        Ensure we are operating on an up-to-date object

        :return: updated WorkflowRequest object
        """

        request = self.session.merge(request)
        return request
